/* Imperative AVL trees.
   Copyright (C) 2003-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "portable.h"
#include "eval.h"
#include "ex1proto.h"
#include "expr.h"
#include "avl.h"

PRIVATE void avlSweep P((INOUT char *extdata, IN CONST doExtDesc *doext));
PRIVATE ExprPtr avlLoad P((IN CONST extDesc *desc, IN Boolean shared,
			   INOUT ExprPtr sharedref, IN CONST loadAction load));

PUBLIC extDesc tableExtDesc = {
  (CHAR *)"table",
  { 't', 'a', 'b', 'l' }, NULL,
  NULL, /* no hash function, tables are not shared */
  NULL, /* no comparison function, tables are not shared */
  NULL, /* no print function */
  avlSweep,
  avlLoad
};

PRIVATE ExprPtr avl_get (an, key)
     IN CONST ExprPtr an;
     IN CONST ExprPtr key;
{
  struct avlnode *node;

 again:
  node = (struct avlnode *)EXTDATA(an);
  if (node==NULL)
    return NOEXPR;
  if (key==node->key)
    return node->val;
  if (key<node->key)
    {
      an = AVL_LEFT(node);
      goto again;
    }
  else
    {
      an = AVL_RIGHT(node);
      goto again;
    }
}

#define AVL_D_LEFT(node,d) (AVL_BIT_RIGHT(node)?((d)-2):((d)-1))
#define AVL_D_RIGHT(node,d) (AVL_BIT_LEFT(node)?((d)-2):((d)-1))

PRIVATE void rebalance_left(node, dr_minus_dl)
     INOUT struct avlnode *node;
     IN int dr_minus_dl;
{
  switch (dr_minus_dl)
    {
    case 1: /* right is deeper */
      AVL_SETBIT_RIGHT(node);
      AVL_CLRBIT_LEFT(node);
      break;
    case 0:
      AVL_CLRBIT_LEFT(node);
      AVL_CLRBIT_RIGHT(node);
      break;
    case -1: /* left is deeper */
      AVL_SETBIT_LEFT(node);
      AVL_CLRBIT_RIGHT(node);
      break;
    default:
      kinderror(message(945),dr_minus_dl);
      /*NOTREACHED*/
      break;
    }
}

#define rebalance_right(node,dd) rebalance_left(node,-(dd))

PRIVATE Boolean avl_put (an, key, val, behind) /* return true if depth increased. */
     INOUT ExprPtr an;
     IN CONST ExprPtr key, val;
     IN int behind;
{
  struct avlnode *node;

  node = (struct avlnode *)EXTDATA(an);
  if (node==NULL)
    {
      node = (struct avlnode *)MALLOC(sizeof(struct avlnode));
      node->key = key;
      node->val = val;
      node->left = node->right = NIL; /* dummy, for GC. */
      EXTDATA(an) = (char *)node;
      HITBARRIER(an);
      node->left = MakeExternal(&avlExtDesc, NULL);
      node->right = MakeExternal(&avlExtDesc, NULL);
      return true;
    }
  if (key==node->key)
    {
      if (behind)
	return;
      node->val = val;
      HITBARRIER(an);
      return false;
    }
  if (key<node->key)
    {
      ExprPtr l,r;
      struct avlnode *la;
      int dl,dr;

      r = AVL_RIGHT(node);
      l = AVL_LEFT(node);
      if (!avl_put (l, key, val, behind))
	return false;
      la = (struct avlnode *)EXTDATA(l);
      /* Now we have
	   x
	  / \
	 l   r
	 and the depth of l has increased.
      */
      if (AVL_BIT_LEFT(node)) /* But l was already deeper than r, so we need to rebalance. */
	{ /* Since l is deep, it is itself a tree, so we have:
	           x
		  <-
		  / \
	     l:  y   r
                / \
	      ll   rl
	  */
	  ExprPtr ll, rl;
	  int dll, drl;

	  dl = 5; /* Assuming (up to some additive constant) that the depth of
		      the whole tree is 5, then dl = depth(l) was 4 and is now 5,
		      and dr = depth(r) = 3. */
	  dr = 3;
	  ll = AVL_LEFT(la);
	  rl = AVL_RIGHT(la);
	  if (AVL_BIT_RIGHT(la))
	    { /* Here rl is deep, too, so we need to do a double rotation:
		 node: x                           z
		      <-
		      / \                        /   \
	         l:  y   r         --->         y     x
		     ->                        =/->
                    / \                        / \   / \
	          ll   z  :rl                ll rll rlr r
		      / \
		    rll rlr
	      */
	      int drll, drlr;
	      struct avlnode *rla = (struct avlnode *)EXTDATA(rl);

	      /* Since d(l)=5, and rl is deeper than ll,
		 d(rl) = 4 and d(ll)=3. */
	      dll = 3;
	      drl = 4;
	      drll = AVL_D_LEFT(rla, drl);
	      drlr = AVL_D_RIGHT(rla, drl);
	      /* Now we shuffle pointers around: */
	      node->right = rl;
	      HITBARRIER(an);
	      /* Now we have got:
		node:  x
		     /   \
		l:  y     z  :rl
		   / \   / \
		 ll  .. rll rlr
	      */
	      la->right = AVL_LEFT(rl);
	      rebalance_left(la,drll-dll);
	      HITBARRIER(l);
	      /* Now l is:
		     y
                    / \
		   ll rll
	      */
	      rla->left = rla->right;
	      rla->right = r;
	      rebalance_left(rla,dr-drlr);
	      HITBARRIER(rl);
	      /* Now rl is:
		     z
		    / \
		  rlr  r
	      */
	      { /* To recap, we have got:
		    node:  x
		         /   \
		    l:  y     z  :rl
		       / \   / \
		     ll rll rlr r
		   It remains to swap x with z and rebalance the top node.
		*/
		ExprPtr aux;

		aux = node->key;
		node->key = rla->key;
		rla->key = aux;
		aux = node->val;
		node->val = rla->val;
		rla->val = aux;
		dl = /*1+*/max(dll,drll);
		drl = /*1+*/max(drlr,dr);
		rebalance_left(node,drl-dl);
		/* Done! */
		return max(dl,drl) > 3; /* == depth(node)[=2+max(dl,drl)] > 5. */
	      }
	    }
	  else
	    { /* AVL_BIT_RIGHT(la) is 0, so rl is not deeper than ll,
		 and we do a simple rotation:
	           node:  x                  y
		         <-
		         / \      --->     /   \
	           l:   y   r             ll    x
                       / \                     / \
	             ll   rl                  rl  r
	      */
	      dll = AVL_D_LEFT(la, dl);
	      drl = AVL_D_RIGHT(la, dl);
	      node->left = ll;
	      node->right = l;
	      HITBARRIER(an);
	      /* Now we have got:
		 node:  x
		      /   \
		     ll    y  :l
		          / \
			ll   rl
	      */
	      la->left = rl;
	      la->right = r;
	      HITBARRIER(l);
	      /* Now we have got:
		 node:  x
		      /   \
		     ll    y  :l
		          / \
			rl   r
	      */
	      rebalance_left(la,dr-drl);
	      dl = 1+max(drl,dr);
	      rebalance_left(node,dl-dll);
	      return max(dll, dl) > 4; /* == max(dll,dl)+1 > 5 */
	    }
	}
      else
	{ /* Here the depth of l has increased but l was no deeper than r,
	     so we just recompute the balance information at the top node. */
	  if (AVL_BIT_RIGHT(node))
	    {
	      AVL_CLRBIT_RIGHT(node);
	      return false;
	    }
	  else
	    {
	      AVL_SETBIT_LEFT(node);
	      return true;
	    }
	}
    }
  else /* Symmetric case: insert key, val into right subtree.
	  The following code is exactly the one above for
	  the key<node->key case, except left/right, LEFT/RIGHT
	  have been swapped. (But the letters 'l' and 'r' have
	  not been swepped, so that 'l' means right and 'r' left
	  here. And the graphics have not been redrawn as they
	  ought to be.) */
    {
      ExprPtr l,r;
      struct avlnode *la;
      int dl,dr;

      r = AVL_LEFT(node);
      l = AVL_RIGHT(node);
      if (!avl_put (l, key, val, behind))
	return false;
      la = (struct avlnode *)EXTDATA(l);
      /* Now we have
	   x
	  / \
	 l   r
	 and the depth of l has increased.
      */
      if (AVL_BIT_RIGHT(node)) /* But l was already deeper than r, so we need to rebalance. */
	{ /* Since l is deep, it is itself a tree, so we have:
	           x
		  <-
		  / \
	     l:  y   r
                / \
	      ll   rl
	  */
	  ExprPtr ll, rl;
	  int dll, drl;

	  dl = 5; /* Assuming (up to some additive constant) that the depth of
		      the whole tree is 5, then dl = depth(l) was 4 and is now 5,
		      and dr = depth(r) = 3. */
	  dr = 3;
	  ll = AVL_RIGHT(la);
	  rl = AVL_LEFT(la);
	  if (AVL_BIT_LEFT(la))
	    { /* Here rl is deep, too, so we need to do a double rotation:
		 node: x                           z
		      <-
		      / \                        /   \
	         l:  y   r         --->         y     x
		     ->                        =/->
                    / \                        / \   / \
	          ll   z  :rl                ll rll rlr r
		      / \
		    rll rlr
	      */
	      int drll, drlr;
	      struct avlnode *rla = (struct avlnode *)EXTDATA(rl);

	      /* Since d(l)=5, and rl is deeper than ll,
		 d(rl) = 4 and d(ll)=3. */
	      dll = 3;
	      drl = 4;
	      drll = AVL_D_RIGHT(rla, drl);
	      drlr = AVL_D_LEFT(rla, drl);
	      /* Now we shuffle pointers around: */
	      node->left = rl;
	      HITBARRIER(an);
	      /* Now we have got:
		node:  x
		     /   \
		l:  y     z  :rl
		   / \   / \
		 ll  .. rll rlr
	      */
	      la->left = AVL_RIGHT(rl);
	      rebalance_right(la,drll-dll);
	      HITBARRIER(l);
	      /* Now l is:
		     y
                    / \
		   ll rll
	      */
	      rla->right = rla->left;
	      rla->left = r;
	      rebalance_right(rla,dr-drlr);
	      HITBARRIER(rl);
	      /* Now rl is:
		     z
		    / \
		  rlr  r
	      */
	      { /* To recap, we have got:
		    node:  x
		         /   \
		    l:  y     z  :rl
		       / \   / \
		     ll rll rlr r
		   It remains to swap x with z and rebalance the top node.
		*/
		ExprPtr aux;

		aux = node->key;
		node->key = rla->key;
		rla->key = aux;
		aux = node->val;
		node->val = rla->val;
		rla->val = aux;
		dl = /*1+*/max(dll,drll);
		drl = /*1+*/max(drlr,dr);
		rebalance_right(node,drl-dl);
		/* Done! */
		return max(dl,drl) > 3; /* == depth(node)[=2+max(dl,drl)] > 5. */
	      }
	    }
	  else
	    { /* AVL_BIT_LEFT(la) is 0, so rl is not deeper than ll,
		 and we do a simple rotation:
	           node:  x                  y
		         <-
		         / \      --->     /   \
	           l:   y   r             ll    x
                       / \                     / \
	             ll   rl                  rl  r
	      */
	      dll = AVL_D_RIGHT(la, dl);
	      drl = AVL_D_LEFT(la, dl);
	      node->right = ll;
	      node->left = l;
	      HITBARRIER(an);
	      /* Now we have got:
		 node:  x
		      /   \
		     ll    y  :l
		          / \
			ll   rl
	      */
	      la->right = rl;
	      la->left = r;
	      HITBARRIER(l);
	      /* Now we have got:
		 node:  x
		      /   \
		     ll    y  :l
		          / \
			rl   r
	      */
	      rebalance_right(la,dr-drl);
	      dl = 1+max(drl,dr);
	      rebalance_right(node,dl-dll);
	      return max(dll, dl) > 4; /* == max(dll,dl)+1 > 5 */
	    }
	}
      else
	{ /* Here the depth of l has increased but l was no deeper than r,
	     so we just recompute the balance information at the top node. */
	  if (AVL_BIT_LEFT(node))
	    {
	      AVL_CLRBIT_LEFT(node);
	      return false;
	    }
	  else
	    {
	      AVL_SETBIT_GAG(node);
	      return true;
	    }
	}
    }
}

PRIVATE Boolean avl_remove_rightmost (resnode, an) /* return true if depth decreased. */
     INOUT struct avlnode *resnode;
     INOUT ExprPtr an; /* deletes rightmost node of an, moving its contents to node */
{
  struct avlnode *node;
  ExprPtr r;

  node = (struct avlnode *)EXTDATA(an);
  r = AVL_RIGHT(node);
  if (EXTDATA(r)==NULL) /* We have reached the rightmost node. */
    {
      *resnode = *node;
      MARKFREE(node);
      EXTDATA(an) = NULL;
      HITBARRIER(an);
      return true;
    }
  else if (avl_remove_rightmost (resnode, r)) /* Otherwise, delete the rightmost node from r, */
    /* and if depth of r decreased, then rebalance */
    return avl_remove_rebalance_right (an);
  else !!!
}

PRIVATE Boolean avl_remove (an, key) /* return true if depth decreased. */
     INOUT ExprPtr an;
     IN CONST ExprPtr key;
{
  struct avlnode *node;

  node = (struct avlnode *)EXTDATA(an);
  if (node==NULL)
    return false;
  if (key==node->key)
    {
      ExprPtr l,r;
      struct avlnode *son;

      l = AVL_LEFT(node);
      r = AVL_RIGHT(node);
      if (EXTDATA(l)==NULL)
	{
	  son = (struct hnode *)EXTDATA(r);
	  if (son==NULL)
	    {
	      EXTDATA(an) = NULL;
	      MARKFREE(node);
	    }
	  else *node = *son;
	  HITBARRIER(an);
	}
      else if (EXTDATA(r)==NULL)
	{
	  son = (struct hnode *)EXTDATA(l);
	  *node = *son;
	  HITBARRIER(an);
	}
      else
	{
	  
	}
    }
  if (key<node->key)
    {
    }
  else
    {
    }
}

PRIVATE ExprPtr table_new ()
{
  ExprPtr e;

  e = MakeExternal(&tableExtDesc, NULL);
  return e;
}

DEF_ACTION(table)
{
  return table_new();
}

DEF_PRIVATE_ACTION(q_table_get)
{ /* env is the external with the hnode inside it. */
  ExprPtr e;

  e = table_get_1 ((struct hnode *)EXTDATA(env), arg);
  if (e==NOEXPR)
    return none;
  return Kdata(0,optttm,e); /* SOME e */
}

DEF_ACTION(table_get)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_table_get),arg);
}

DEF_PRIVATE_ACTION(q_table_put)
{ /* env is the external with the hnode inside it. */
  table_put (env, CAR(arg), CDR(arg), 0);
  return tuple0;
}

DEF_ACTION(table_put)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_table_put),arg);
}

DEF_PRIVATE_ACTION(q_table_put_behind)
{ /* env is the external with the hnode inside it. */
  table_put (env, CAR(arg), CDR(arg), 1);
  return tuple0;
}

DEF_ACTION(table_put_behind)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_table_put_behind),arg);
}

DEF_PRIVATE_ACTION(q_table_remove)
{
  table_remove (env, arg);
  return tuple0;
}

DEF_ACTION(table_remove)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_table_remove),arg);
}

DEF_ACTION(table_kill)
{
  table_free ((struct hnode *)EXTDATA(arg));
  EXTDATA(arg) = (char *)NULL;
  HITBARRIER(arg);
  return tuple0;
}

Boolean q_table_sweep_do(key,val,data)
IN CONST ExprPtr key,val;
IN CONST char *data;
{
  ExprPtr e, res;
  ENTER;

  e = KONS(key, val);
  res = (*ACTION((ExprPtr)data))(e, QENV ((ExprPtr)data));
  QUIT;
  return (res==trueValue);
}

DEF_PRIVATE_ACTION(q_table_sweep)
{
  Boolean res;

  res = table_sweep ((struct hnode *)EXTDATA(env), q_table_sweep_do, (char *)arg);
  return res?trueValue:falseValue;
}

DEF_ACTION(table_sweep)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_table_sweep),arg);
}

PRIVATE MapPtr table_collect_1 (hn,kprefix,depth,do_f,data)
     IN CONST struct hnode *hn;
     IN unsigned long kprefix;
     IN COUNTER depth;
     IN CONST MapPtr (*do_f) P((IN CONST ExprPtr key, IN CONST ExprPtr val, INOUT char *data));
     INOUT char *data;
{
  unsigned long i;
  struct hnode *e;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[2]; } _x;

  if (hn==NULL)
    return EmptyMap();
  if (depth==0)
    {
      _x.p[0] = NewMap();
      _x.p[1] = NewMap();
      _x.n = 2;
      _x.next = theStackFrames;
      theStackFrames = (gcStackView *)&_x;
      for (i=0; i < (1<<HN_BITS); i++)
	{
	  e = hn->sons[i];
	  if (e==NULL)
	    continue;
	  _x.p[1] = (*do_f) ((ExprPtr)(((kprefix | i) << NLOWBITS) | lowbits), (ExprPtr)e, data);
	  _x.p[0] = MapMerge (_x.p[0], _x.p[1]);
	  _x.p[1] = NewMap();
	}
      theStackFrames = _x.next;
    }
  else
    {
      COUNTER d;

      _x.p[0] = NewMap();
      _x.p[1] = NewMap();
      _x.n = 2;
      _x.next = theStackFrames;
      theStackFrames = (gcStackView *)&_x;
      d = HN_BITS*depth;
      depth--;
      for (i=0; i < (1<<HN_BITS); i++)
	{
	  e = hn->sons[i];
	  if (e==NULL)
	    continue;
	  _x.p[1] = table_collect_1 (e, kprefix | (i << d), depth, do_f, data);
	  _x.p[0] = MapMerge (_x.p[0], _x.p[1]);
	}
      theStackFrames = _x.next;
    }
  return _x.p[0];
}

PRIVATE MapPtr table_collect (hn,do_f,data)
     IN CONST struct hnode *hn;
     IN CONST MapPtr (*do_f) P((IN CONST ExprPtr key, IN CONST ExprPtr val, INOUT char *data));
     INOUT char * data;
{
  unsigned long i;
  struct hnode *e;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[2]; } _x;

  if (hn==NULL)
    return NewMap();
  _x.p[0] = NewMap();
  _x.p[1] = NewMap();
  _x.n = 2;
  _x.next = theStackFrames;
  theStackFrames = (gcStackView *)&_x;
  for (i=0; i < (1<<HN_BITS_LEAF); i++)
    {
      e = hn->sons[i];
      if (e==NULL)
	continue;
      _x.p[1] = table_collect_1 (e, (i << HN_BITS*(HN_DEPTH-1)), HN_DEPTH-2, do_f, data);
      _x.p[0] = MapMerge (_x.p[0], _x.p[1]);
      _x.p[1] = NewMap();
    }
  theStackFrames = _x.next;
  return _x.p[0];
}

MapPtr q_table_collect_do(key,val,data)
IN CONST ExprPtr key,val;
IN CONST char *data;
{
  ExprPtr e, res;
  ENTER;

  e = KONS(key,val);
  res = (*ACTION((ExprPtr)data))(e, QENV ((ExprPtr)data));
  QUIT;
  return res;
}

DEF_PRIVATE_ACTION(q_table_collect)
{
  MapPtr res;

  res = table_collect ((struct hnode *)EXTDATA(env), q_table_collect_do, (char *)arg);
  return res;
}

DEF_ACTION(table_collect)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_table_collect),arg);
}
