/* Bytecode compiler auxiliary functions.
   Copyright (C) 1995-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#if TRANSLATOR
#include "expr.h"
#include "refcount.h"
#include "mem.h"
#include "bc.h"
#include "bcaux.h"
#include "timing.h"
#include "debug.h"
#include "transfer.h"
#include "errproto.h"
#include "memproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "iter.h"
#include "rbc.h"
#ifdef HAS_PROTOTYPES
#include "sbcaux.h"
#endif

#define __SEG__ BYTECODE

#define DEBUG

EXTERN ExprPtr primCompileMap;

/***** Allocating a byteCodeChunk: *****/

PUBLIC byteCodeChunk *allocBCs = NULL;
PUBLIC byteCodeChunk *oldBCs = NULL;
PRIVATE byteCodeChunk *BCReserve[MAX_GENERATIONS];
EXTERN SHORT generation_count;

PUBLIC byteCodeChunk *bcAlloc(size)
IN INT size;
{
  byteCodeChunk *bc;

  bc = (byteCodeChunk *)MALLOC(size);
  bc->next = allocBCs;
  allocBCs = bc;
  bc->kind = (CHAR)E_BYTECODE;
  bc->mark = 0;
  return bc;
}

PUBLIC void FreeBCs()
{
  byteCodeChunk **bcp,*bc;

  for (bcp = &allocBCs; (bc = *bcp)!=NULL; )
    {
      if (GC_MARKED(bc))
	{
	  GC_UNMARK(bc);
	  bcp = &bc->next;
	}
      else
	{
	  *bcp = bc->next;
	  MARKFREE(bc);
	}
    }
}

PUBLIC void cutoffInitialBCs()
{
  REGISTER byteCodeChunk *bc;

  for (bc = allocBCs; bc!=NULL; bc=bc->next)
    {
      GC_MARK(bc);
    }
  BCReserve[0] = oldBCs = allocBCs;
  allocBCs = NULL;
}

#ifdef GC_SLIDING
PUBLIC void slideOldBCs()
{
  COUNTER i;

  for (i=1; i<generation_count; i++)
    BCReserve[i] = BCReserve[i+1];
}
#endif

PUBLIC void cutoffOldBCs()
{
  REGISTER byteCodeChunk **bcp,*bc;

  BCReserve[generation_count] = oldBCs;
  for (bcp = &allocBCs; (bc = *bcp)!=NULL; bcp = &bc->next)
    GC_MARK(bc);
  *bcp = oldBCs;
  oldBCs = allocBCs;
  allocBCs = NULL;
}

PUBLIC void refreshOldBCs()
{
  REGISTER byteCodeChunk **bcp,*bc,*last;

  last = BCReserve[generation_count];
  for (bcp = &oldBCs; bc = *bcp, bc!=last; bcp = &bc->next)
    {
      GC_UNMARK(bc);
    }
  *bcp = allocBCs;
  allocBCs = oldBCs;
  oldBCs = bc;
}

PUBLIC void unmarkOldBCs()
{
  REGISTER byteCodeChunk *bc;

  for (bc = oldBCs; bc!=NULL; bc = bc->next)
    GC_UNMARK(bc);
}

/***** Functions that output bytecode. *****/

PRIVATE void bccSweep(extdata,doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  bccBuffer *bcc;

  bcc = (bccBuffer *)extdata;
  switch (doext->xtag)
    {
    case X_SAVE:
      /*fallthrough*/
    case X_MARK:
    case X_MARKDATA:
      for (; bcc!=NULL; bcc=bcc->next)
	{
	  REGISTER COUNTER i;
	  REGISTER ExprPtr *ep;

	  for (i=0, ep = bcc->data; i<bcc->n; i++)
	    (*doext->action)(*ep++);
	}
      break;
    case X_FREE:
      for (; bcc!=NULL; )
	{
	  bccBuffer *next;

	  next = bcc->next;
	  MARKFREE(bcc);
	  bcc = next;
	}
      break;
    }
}

PRIVATE extDesc bccDesc = {
  (CHAR *)"bccBuffer",
  { 'b','c','c',' ' }, NULL,
  NULL, /* not shared */
  NULL, /* not shared */
  NULL, /* print as default */
  bccSweep,
  NULL, /* do not load */
  };

PUBLIC bcFile *bcOpen()
{
  bcBuffer *buf, *dbuf;
  bccBuffer *cbuf;
  bcuBuffer *ubuf;
  bcFile *f;

  buf = (bcBuffer *)MALLOC(sizeof(bcBuffer));
  buf->next = NULL;
  buf->n = 0;
  dbuf = (bcBuffer *)MALLOC(sizeof(bcBuffer));
  dbuf->next = NULL;
  dbuf->n = 0;
  cbuf = (bccBuffer *)MALLOC(sizeof(bccBuffer));
  cbuf->next = NULL;
  cbuf->n = 0;
  ubuf = (bcuBuffer *)MALLOC(sizeof(bcuBuffer));
  ubuf->next = NULL;
  ubuf->n = 0;
  f = (bcFile *)MALLOC(sizeof(bcFile));
  f->nargs = 1;
  f->ntargs = -1;
  f->stackDepth = f->maxStackDepth = 2; /* room for the calling environment + one argument */
  f->mapStackDepth = f->maxMapStackDepth = 0;
  f->dead = 0;
  f->buf = f->current = buf;
  f->dbuf = f->dcurrent = dbuf;
  f->cbuf = f->ccurrent = cbuf;
  f->cpos = 0;
  f->ubuf = f->ucurrent = ubuf;
#if DEBUGGING!=0
  f->lastLine = -1;
  f->lastPos = 0;
  f->lastDebugZone = NULL;
#endif
  {
    ExprPtr e;

    e = f->cbuf_e = MakeExternal(&bccDesc,(char *)cbuf);
    (void) KEEP(e); /* protect all data in bccBuffers to come from GC. */
  }
  return f;
}

#ifdef HAS_PROTOTYPES
PUBLIC void bcPut(IN OPCODE c, INOUT bcFile *f)
#else
PUBLIC void bcPut(c,f)
IN OPCODE c;
INOUT bcFile *f;
#endif
{
  REGISTER bcBuffer *current;

  if (f->dead)
    return;
  current = f->current;
  if (current->n==BC_BUFSIZE)
    {
      bcBuffer *next;

      next = (bcBuffer *)MALLOC(sizeof(bcBuffer));
      next->n = 0;
      next->next = NULL;
      current->next = next;
      f->current = current = next;
    }
  current->instr[current->n++] = c;
}

PUBLIC INT bcTell(f)
IN CONST bcFile *f;
{
  REGISTER bcBuffer *buf;
  REGISTER INT n;

  for (n=0, buf=f->buf; buf!=NULL; n += buf->n, buf = buf->next);
  return n;
}

#ifdef HAS_PROTOTYPES
PUBLIC void bcdPut(IN OPCODE c, INOUT bcFile *f)
#else
PUBLIC void bcdPut(c,f)
IN OPCODE c;
INOUT bcFile *f;
#endif
{
  REGISTER bcBuffer *dcurrent;

  /*
    if (f->dead)
    return;
    */
  dcurrent = f->dcurrent;
  if (dcurrent->n==BC_BUFSIZE)
    {
      bcBuffer *next;

      next = (bcBuffer *)MALLOC(sizeof(bcBuffer));
      next->n = 0;
      next->next = NULL;
      dcurrent->next = next;
      f->dcurrent = dcurrent = next;
    }
  dcurrent->instr[dcurrent->n++] = c;
}

PUBLIC INT bcdTell(f)
IN CONST bcFile *f;
{
  REGISTER bcBuffer *dbuf;
  REGISTER INT n;

  for (n=0, dbuf=f->dbuf; dbuf!=NULL; n += dbuf->n, dbuf = dbuf->next);
  return n;
}

PUBLIC void bcdGetState(f,st)
IN CONST bcFile *f;
OUT bcdState *st;
{
  st->dcurrent = f->dcurrent;
  st->n = f->dcurrent->n;
}

PUBLIC void bcdRewindState(f,st)
INOUT bcFile *f;
IN CONST bcdState *st;
{
  bcBuffer *bc,*next;

  st->dcurrent->n = st->n;
  for (bc = st->dcurrent, next = bc->next; next!=NULL;)
    {
       bc = next;
       next = bc->next;
       MARKFREE(bc);
    }
  st->dcurrent->next = NULL;
  f->dcurrent = st->dcurrent;
}

PUBLIC void bcdPatch(addr,f)
IN REGISTER INT addr; /* address to patch */
INOUT bcFile *f;
{
  REGISTER INT here;
  REGISTER OPCODE offset;
  REGISTER bcBuffer *dbuf;

  here = bcdTell(f);
  offset = here-addr;
  if (addr+offset!=here)
    berror(message(671));
  for (dbuf=f->dbuf; dbuf!=NULL; dbuf = dbuf->next)
    {
      if (addr<dbuf->n)
	break;
      addr -= dbuf->n;
    }
  dbuf->instr[addr] = offset;
}

PUBLIC bcFile *bcOpenE(env)
IN CONST compEnv *env;
{
  bcFile *bc;

  bc = bcOpen();
  bccPut(CE_PROF(env),bc);
  return bc;
}

PUBLIC Boolean gotoOp0(opp)
INOUT OPCODE *opp;
{
  switch (*opp)
    {
    case BC_GOTO: *opp = BC_GOTO_0; break;
    default:
      return false;
    }
  return true;
}

PUBLIC void bcPatchJump(addr,excess,f)
IN REGISTER INT addr; /* address just after the opcode for a branching instruction */
IN INT excess;
INOUT bcFile *f;
{
  REGISTER INT here;
  REGISTER OPCODE offset;
  REGISTER bcBuffer *buf;

  here = bcTell(f);
  offset = here-addr;
  if (addr+offset!=here)
    berror(message(94));
  addr -= 2; /* find place where opcode should be stored */
  for (buf=f->buf; buf!=NULL; buf = buf->next)
    {
      if (addr<buf->n)
	break;
      addr -= buf->n;
    }
  if (excess==0) /* then compile special case instructions that do not drop the stack */
    {
      if (gotoOp0(&buf->instr[addr]))
	offset++; /* because pc will be 1 too low when taking the branch in that case */
    }
  if (++addr>=buf->n)
    { buf = buf->next; addr = 0; /* find location of excess parameter */ }
  buf->instr[addr] = (OPCODE)excess;
  if (++addr>=buf->n)
    { buf = buf->next; addr = 0; /* find location of offset parameter */ }
  buf->instr[addr] = offset;
}

PUBLIC void bcPatchSwitchEntry(addr,f)
IN REGISTER INT addr; /* address just after the opcode for a branching instruction */
INOUT bcFile *f;
{
  REGISTER INT here;
  REGISTER OPCODE offset;
  REGISTER bcBuffer *buf;

  here = bcTell(f);
  offset = here-addr;
  if (addr+offset!=here)
    berror(message(94));
  for (buf=f->buf; buf!=NULL; buf = buf->next)
    {
      if (addr<buf->n)
	break;
      addr -= buf->n;
    }
  buf->instr[addr] = offset;
}

PUBLIC void bccPut(e,f)
IN CONST ExprPtr e;
INOUT bcFile *f;
{
  REGISTER bccBuffer *current;

  if (f->dead)
    return;
  current = f->ccurrent;
  if (current->n==BCC_BUFSIZE)
    {
      bccBuffer *next;

      next = (bccBuffer *)MALLOC(sizeof(bccBuffer));
      next->n = 0;
      next->next = NULL;
      current->next = next;
      f->ccurrent = current = next;
    }
  current->data[current->n++] = e;
  f->cpos++;
  HITBARRIER (f->cbuf_e);
}

PUBLIC void bcuPut(data,size,align,f)
IN CONST CHAR *data;
IN INT size;
IN INT align;
INOUT bcFile *f;
{
  REGISTER bcuBuffer *current;

  /*
    if (f->dead)
    return;
    */
  current = f->ucurrent;
  if (align>1)
    {
      current->n = ((current->n+align-1) / align) * align;
    }
  if (current->n+size>BCU_BUFSIZE)
    {
      bcuBuffer *next;

      next = (bcuBuffer *)MALLOC(sizeof(bcuBuffer));
      next->n = 0;
      next->next = NULL;
      current->next = next;
      f->ucurrent = current = next;
    }
  safememcpy(current->data+current->n,data,size);
  current->n += size;
}

PUBLIC INT bcuTell(f)
IN CONST bcFile *f;
{
  REGISTER bcuBuffer *buf;
  REGISTER INT n;

  for (n=0, buf=f->ubuf; buf!=NULL; n += buf->n, buf = buf->next);
  return n;
}

PRIVATE INT byteCodeSize(n,dn,cn,un,rawdata,instr,debug)
IN INT n,dn,cn,un;
OUT COUNTER *rawdata,*instr,*debug;
{
  INT size;

  size = VARSIZE(byteCodeChunk,consts,cn);
  if (rawdata!=NULL)
    *rawdata = (COUNTER)size;
  size += un;
  size = ((size+sizeof(OPCODE)-1)/sizeof(OPCODE))*sizeof(OPCODE); /* align on OPCODE */
  if (instr!=NULL)
    *instr = size;
  size += n*sizeof(OPCODE);
  if (debug!=NULL)
    *debug = size;
  size += dn*sizeof(OPCODE);
  return size;
}

PRIVATE byteCodeChunk *bcClose1(f)
INOUT bcFile *f;
{
  INT n,dn,cn,un;
  COUNTER rawdata,instr,size,debug;
  byteCodeChunk *bc;

  n = bcTell(f);
  dn = bcdTell(f);
  cn = bccTell(f);
  un = bcuTell(f);
  size = byteCodeSize(n,dn,cn,un,&rawdata,&instr,&debug);
  bc = bcAlloc(size);
  bc->mapStackOff = VARSIZE(gcStackView,p,f->maxStackDepth);
  bc->excStackOff = bc->mapStackOff + f->maxMapStackDepth*sizeof(ExprPtr *);
  bc->nargs = f->nargs;
  bc->ntargs = f->ntargs;
  bc->stackSize = f->maxStackDepth;
  bc->nConsts = cn;
  bc->rawData = (CHAR *)(((char *)bc)+rawdata);
  bc->instr = (OPCODE *)(((char *)bc)+instr);
  bc->debug = (OPCODE *)(((char *)bc)+debug);
  bc->size = size;
  {
    bccBuffer *buf,*next;
    ExprPtr *to;

    for (to = bc->consts, buf = f->cbuf; buf!=NULL; buf=next)
      {
	safememcpy(to,buf->data,buf->n*sizeof(ExprPtr));
	next = buf->next;
	to += buf->n;
	/*MARKFREE(buf); // no: GC will get back the memory */
      }
  }
  {
    bcuBuffer *buf,*next;
    CHAR *to;

    for (to = (CHAR *)(((char *)bc)+rawdata), buf = f->ubuf; buf!=NULL; buf=next)
      {
	safememcpy(to,buf->data,buf->n);
	next = buf->next;
	to += buf->n;
	MARKFREE(buf);
      }
  }
  {
    bcBuffer *buf,*next;
    OPCODE *to;

    for (to = (OPCODE *)(((char *)bc)+instr), buf=f->buf; buf!=NULL; buf=next)
      {
	safememcpy(to,buf->instr,buf->n*sizeof(OPCODE));
	next = buf->next;
	to += buf->n;
	MARKFREE(buf);
      }
  }
  {
    bcBuffer *dbuf,*next;
    OPCODE *to;

    for (to = (OPCODE *)(((char *)bc)+debug), dbuf=f->dbuf; dbuf!=NULL; dbuf=next)
      {
	safememcpy(to,dbuf->instr,dbuf->n*sizeof(OPCODE));
	next = dbuf->next;
	to += dbuf->n;
	MARKFREE(dbuf);
      }
  }
  MARKFREE(f);
  return bc;
}

PUBLIC ExprPtr bcClose(f)
IN bcFile *f;
{
  byteCodeChunk *bc;
  ExprPtr e;

  BEGINQUICKKEEP();
  bc = bcClose1(f);
  QUICKKEEP((ExprPtr)bc);
  e = MakePair(E_TUPLE_2,(ExprPtr)bc,NIL);
  ENDQUICKKEEP();
  (void) KEEP(e);
  return e;
}

PUBLIC void bcPush(amount,f)
IN COUNTER amount;
IN bcFile *f;
{
  f->stackDepth += amount;
  if (f->dead)
    return;
  if (f->stackDepth>f->maxStackDepth)
    {
      f->maxStackDepth = f->stackDepth;
      if (f->maxStackDepth>BC_STACK_SIZE)
	{
	  type_error((TZone *)0L,(CHAR *)message(647));
	}
    }
}

PUBLIC void bcMapPush(amount,f)
IN COUNTER amount;
IN bcFile *f;
{
  f->mapStackDepth += amount;
  if (f->dead)
    return;
  if (f->mapStackDepth>f->maxMapStackDepth)
    f->maxMapStackDepth = f->mapStackDepth;
}

/* Computing sets of free variables: */

PRIVATE MapPtr freeVarsPattern P((IN CONST ExprPtr pat,INOUT compEnv *env));

PRIVATE MapPtr freeVarsVec(e,env)
IN CONST ExprPtr e;
INOUT compEnv *env;
{
  COUNTER i,n;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[2]; } _x;

  _x.n = 2;
  _x.next = theStackFrames;
  _x.p[0] = NewMap();
  _x.p[1] = NOEXPR;
  theStackFrames = (gcStackView *)&_x;

  for (i=0, n=AVAL(e)->n; i<n; i++)
    {
      _x.p[1] = freeVars(AVAL(e)->expr[i],env);
      setMapMerge(&_x.p[0],_x.p[1]);
    }

  theStackFrames = _x.next;
  return _x.p[0];
}

DEF_PRIVATE_ACTION_1ARG(freevarspat)
{
  struct freevars *data;

  data = (struct freevars *)ENV_UNDERLYING(env);
  return freeVarsPattern(data->code,data->env);
}
ACTION_END

PRIVATE MapPtr freeVarsPat1(pat,env)
IN CONST ExprPtr pat;
INOUT compEnv *env;
{
  struct freevars *d = (struct freevars *)alloctmp(sizeof(struct freevars));

  d->code = pat;
  d->env = env;
  return CallCC(ACTION_NAME(freevarspat),(ExprPtr)d,true);
}

PRIVATE MapPtr freeVarsVar(var,env)
IN CONST ExprPtr var;
INOUT compEnv *env;
{
  COUNTER level;

  level = env->level-VARLEVEL(var);
  if (env->globenvs & (1L << level)) /* level is GLOBENV (or a copy of it) */
    return NewMap(); /* we don't take globals into account. */
  {
    ExprPtr i_l;
    ENTER;

    i_l = KINT(VARRANK(var));
    i_l = KONS(i_l,KINT(level)); /* level is necessarily !=0 */
    i_l = MapletKons(i_l,UNIT);
    QUIT;
    return i_l;
  }
}

PRIVATE MapPtr freeVarsPattern(pat,env)
IN CONST ExprPtr pat;
INOUT compEnv *env;
{
  if (STKFULL(pat))
    return freeVarsPat1(pat,env);
again:
  switch (pat->kind)
    {
    case E_VARINTRO:
    case E_NIL: case E_TUPLE_0: case E_RECORD_0:
    case E_EMPTY: case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
    case E_EXCEPTION:
    case E_COMPILED_CLOSURE: case E_INTERPRETED_CLOSURE:
#ifdef OBSOLETE
    case E_MEMO_CLOSURE:
#endif
    case E_QUOTE: case E_PAT_ANY:
      return NewMap();
    case E_DEBUGNAME:
      pat = DNPAT(pat);
      goto again;
    case E_VARREF:
      return freeVarsVar(pat,env);
    case E_PATCHOICE:
      {
	struct _x { COUNTER n; gcStackView *next; ExprPtr p[2]; } _x;

	_x.n = 1;
	_x.next = theStackFrames;
	_x.p[0] = freeVarsPattern(CAR(CDR(pat)),env);
	theStackFrames = (gcStackView *)&_x;

	env->level++;
	_x.p[1] = freeVars(CDR(CDR(pat)),env);
	env->level--;
	_x.n = 2;
	_x.p[0] = MapOverwrite(_x.p[0],_x.p[1]);
	_x.n = 1;
	_x.p[1] = freeVarsPattern(CAR(pat),env);
	_x.n = 2;
	_x.p[0] = MapOverwrite(_x.p[0],_x.p[1]);

	theStackFrames = _x.next;
	return _x.p[0];
      }
    case E_CONS: case E_UNSHARED_CONS:
    case E_MAPLET: case E_PAT_REVCONS: case E_PAT_CONS:
    case E_TUPLE_2: case E_UNSHARED_TUPLE_2:
    case E_PAT_TUPLE_GNRL_CONS: case E_PAT_REVTUPLE_GNRL_CONS:
    case E_PAT_REVTUPLE_2:
    case E_PAT_RECORD_GNRL_CONS: case E_PAT_REVRECORD_GNRL_CONS:
    case E_APPL: case E_PAT_AS:
    case E_PAT_RECORD_MAPADD: case E_PAT_MAPADD:
      {
	struct _x { COUNTER n; gcStackView *next; ExprPtr p[2]; } _x;

	_x.n = 1;
	_x.next = theStackFrames;
	_x.p[0] = freeVarsPattern(CAR(pat),env);
	theStackFrames = (gcStackView *)&_x;

	_x.p[1] = freeVarsPattern(CDR(pat),env);
	_x.n = 2;
	_x.p[0] = MapOverwrite(_x.p[0],_x.p[1]);

	theStackFrames = _x.next;
	return _x.p[0];
      }
#ifdef DATACONPAT
    case E_DATACON:
#endif
    case E_TUPLE_1: case E_UNSHARED_TUPLE_1: case E_REF:
      pat = REFVAL(pat);
      goto again;
    case E_DATA:
    case E_RECORD_1: case E_UNSHARED_RECORD_1:
      pat = CDR(pat);
      goto again;
    case E_PAT_CHECK:
      pat = PATCHECKVAL(pat);
      goto again;
    case E_PAT_PACK:
      pat = PACKVAL(pat);
      goto again;
    case E_PAT_CARD_EQ: case E_PAT_CARD_GEQ:
      pat = COUNTEDVAL(pat);
      goto again;
    default:
      kinderror(message(662),pat->kind);
      /*NOTREACHED*/
      return NewMap();
    }
}

#ifdef OBSOLETE
MAPCOLLECTFUNC(do_freevars)
{
  compEnv *env;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[2]; } _x;

  _x.n = 0;
  _x.next = theStackFrames;
  theStackFrames = (gcStackView *)&_x;

  env = MAPDATA(compEnv *);
  _x.p[0] = freeVars(x,env);
  _x.n = 1;
  _x.p[1] = freeVars(y,env);
  _x.n = 2;
  _x.p[0] = MapOverwrite(_x.p[0],_x.p[1]);

  theStackFrames = _x.next;
  return _x.p[0];
}

PRIVATE MapPtr freeVarsMap(code,env)
IN CONST ExprPtr code;
INOUT compEnv *env;
{
  return MapCollect(code,do_freevars,(VOIDPTR)env);
}
#endif

MAPCOLLECTFUNC(do_chopfv)
{
  compEnv *env;

  env = MAPDATA(compEnv *);
  if (IVAL(CDR(x))<=env->level)
    return MapletKons(x,y);
  return NewMap();
}

PRIVATE MapPtr chopFv(fv,env)
IN CONST MapPtr fv;
IN CONST compEnv *env;
{
  return MapCollect(fv,do_chopfv,(VOIDPTR)env);
}

PRIVATE MapPtr freeVarsBindings P((IN CONST ExprPtr bindlist,INOUT compEnv *env));

PRIVATE MapPtr freeVarsLet(bindings,body,env,fun)
IN CONST ExprPtr bindings,body;
INOUT compEnv *env;
IN CONST freeVarsFun fun;
{
  COUNTER n,savelevel,saveglob;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[2]; } _x;

  savelevel = env->level;
  saveglob = env->globenvs;
  n = IVAL(CAR(bindings));
  if (n<0)
    {
      env->level++;
      env->globenvs |= (1L << env->level);
    }
  env->level++;

  _x.n = 1;
  _x.p[0] = freeVarsBindings(CDR(bindings),env);
  _x.next = theStackFrames;
  theStackFrames = (gcStackView *)&_x;

  _x.p[1] = (*fun)(body,env);
  _x.n = 2;
  _x.p[0] = MapOverwrite(_x.p[0],_x.p[1]);
  _x.n = 1;
  env->level = savelevel;
  env->globenvs = saveglob;

  _x.p[0] = chopFv(_x.p[0],env);

  theStackFrames = _x.next;
  return _x.p[0];
}

PRIVATE MapPtr freeVarsBindings(bindlist,env)
IN CONST ExprPtr bindlist;
INOUT compEnv *env;
{
  ExprPtr binding;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[2]; } _x;

  _x.n = 2;
  _x.next = theStackFrames;
  _x.p[0] = NewMap();
  _x.p[1] = NOEXPR;
  theStackFrames = (gcStackView *)&_x;

  while (bindlist!=NIL)
    {
      binding = CAR(bindlist);
      switch (binding->kind)
	{
	case E_LET:
	  _x.p[1] = freeVarsLet(LETBIND(binding),LETBODY(binding),
				env,freeVarsBindings);
	  break;
	case E_DEBUGDECL:
	  _x.p[1] = NewMap();
	  break;
	default:
	  _x.p[1] = freeVars(CDR(binding),env);
	  break;
	}
      _x.p[0] = MapOverwrite(_x.p[0],_x.p[1]);
      bindlist = CDR(bindlist);
    }

  theStackFrames = _x.next;
  return _x.p[0];
}

PRIVATE MapPtr freeVarsDoms(domlist,env)
IN CONST ExprPtr domlist;
INOUT compEnv *env;
{
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[2]; } _x;

  _x.n = 2;
  _x.next = theStackFrames;
  _x.p[0] = NewMap();
  _x.p[1] = NOEXPR;
  theStackFrames = (gcStackView *)&_x;

  while (domlist!=NIL)
    switch (domlist->kind)
      {
      case E_PAT_RECORD_GNRL_CONS:
      case E_PAT_REVRECORD_GNRL_CONS:
      case E_PAT_RECORD_MAPADD:
	_x.p[1] = freeVars(CDR(CAR(domlist)),env);
	_x.p[0] = MapOverwrite(_x.p[0],_x.p[1]);
	domlist = CDR(domlist);
	break;
      case E_TUPLE_1:
	env->level++;
	_x.p[1] = freeVarsBindings(REFVAL(domlist),env);
	env->level--;
	_x.p[0] = MapOverwrite(_x.p[0],_x.p[1]);
	domlist = NIL;
	break;
      }

  theStackFrames = _x.next;
  return _x.p[0];
}

PRIVATE MapPtr freeVarsFunc(fun,env)
IN CONST ExprPtr fun;
INOUT compEnv *env;
{
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[2]; } _x;
  ExprPtr match;
  Boolean found;

  _x.p[0] = MapGet(fun,CE_FV(env),&found);
  if (!found)
    {
      match = COUNTEDVAL(fun);
      env->level++;
      _x.n = 2;
      _x.next = theStackFrames;
      _x.p[1] = NOEXPR;
      _x.p[0] = freeVars(CDR(match),env);
      theStackFrames = (gcStackView *)&_x;

      env->level--;
      _x.p[1] = freeVarsPattern(CAR(match),env);
      _x.p[0] = MapOverwrite(_x.p[0],_x.p[1]);
      _x.n = 1;
      _x.p[0] = chopFv(_x.p[0],env);
      setMapPut(fun,_x.p[0],&CE_FV(env));

      theStackFrames = _x.next;
    }
  return _x.p[0];
}

PRIVATE MapPtr freeVarsQuant(doms,fp,env)
IN CONST ExprPtr doms,fp;
INOUT compEnv *env;
{
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[2]; } _x;

  _x.n = 1;
  _x.next = theStackFrames;
  _x.p[0] = freeVarsDoms(doms,env);
  theStackFrames = (gcStackView *)&_x;

  env->level++;
  _x.p[1] = freeVars(fp,env);
  _x.n = 2;
  env->level--;
  _x.p[0] = MapOverwrite(_x.p[0],_x.p[1]);
  _x.p[0] = chopFv(_x.p[0],env);

  theStackFrames = _x.next;
  return _x.p[0];
}

DEF_PRIVATE_ACTION_1ARG(freevars)
{
  struct freevars *data;

  data = (struct freevars *)ENV_UNDERLYING(env);
  return freeVars(data->code,data->env);
}
ACTION_END

PRIVATE MapPtr freeVars1(code,env)
IN CONST ExprPtr code;
INOUT compEnv *env;
{
  struct freevars *d = (struct freevars *)alloctmp(sizeof(struct freevars));

  d->code = code;
  d->env = env;
  return CallCC(ACTION_NAME(freevars),(ExprPtr)d,true);
}

PRIVATE MapPtr freeVarsCouple(code,env)
IN CONST ExprPtr code;
INOUT compEnv *env;
{
  struct _x { COUNTER n; gcStackView * next; ExprPtr p[2]; } _x;

  _x.n = 1;
  _x.next = theStackFrames;
  _x.p[0] = freeVars(CAR(code),env);
  theStackFrames = (gcStackView *)&_x;

  _x.p[1] = freeVars(CDR(code),env);
  _x.n = 2; /*###*/
  _x.p[0] = MapOverwrite(_x.p[0],_x.p[1]);

  theStackFrames = _x.next;
  return _x.p[0];
}

PUBLIC MapPtr freeVars(code,env)
IN CONST ExprPtr code;
INOUT compEnv *env;
{
  if (STKFULL(code))
    return freeVars1(code,env);
again:
  switch (code->kind)
    {
    case E_NIL: case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
    case E_EMPTY: case E_TUPLE_0: case E_RECORD_0:
    case E_ENV: case E_STACK:
    case E_COMPILED_CLOSURE: case E_INTERPRETED_CLOSURE:
#ifdef OBSOLETE
    case E_MEMO_CLOSURE:
#endif
    case E_QUOTE: case E_EXCEPTION: case E_BYTECODE:
    case E_DATA:
      return NewMap();
    case E_VARREF:
      return freeVarsVar(code,env);
    case E_MEMO_FN:
      code = MEMOFUN(code);
      goto again;
    case E_DELAY:
      code = DCODE(code);
      goto again;
    case E_DATACON:
      code = COUNTEDVAL(code);
      goto again;
    case E_POLYFUN:
      /*case E_N_LAMBDA:*/
      return freeVarsFunc(code,env);
    case E_REF: case E_TUPLE_1: case E_UNSHARED_TUPLE_1:
    case E_GET: case E_NOT:
      code = REFVAL(code);
      goto again;
    case E_RECORD_1: case E_UNSHARED_RECORD_1:
    case E_TSELECT: case E_TSEL1: case E_TSEL2: case E_TSELGNRL:
    case E_RSELECT: case E_RSEL1: case E_RSELGNRL:
      code = CDR(code);
      goto again;
    case E_TUPLE_2: case E_UNSHARED_TUPLE_2:
    case E_CONS: case E_UNSHARED_CONS:
    case E_MAPLET: case E_OVERWRITE: case E_UNDERWRITE:
    case E_SET: case E_MAPGET:
    case E_RMU1: case E_UNSHARED_RMU1:
    case E_RMUX: case E_UNSHARED_RMUX:
    case E_IF: case E_PROG: case E_AND: case E_OR:
    case E_EQUAL: case E_NEQUAL:
    case E_APPL: case E_GOTO:
      /*case E_N_APPL:*/
    case E_PROG_WHILE: case E_LIST_WHILE: case E_UNSHARED_LIST_WHILE:
    case E_SET_WHILE: case E_APPEND_WHILE: case E_UNSHARED_APPEND_WHILE:
    case E_UNION_WHILE: case E_UNION_UNDER_WHILE: case E_INTER_WHILE:
    case E_MAP_WHILE: case E_MAP_UNDER_WHILE: case E_ALL_WHILE:
    case E_EXISTS_WHILE: case E_SOME_WHILE: case E_UNSHARED_SOME_WHILE:
    case E_HANDLE:
    case E_PAT_AS: /* in this case, we only need the free vars of CDR(code)
		      if we inline it, or of CAR(code) if we don't, but
		      we cannot decide yet, so we take the union of both. */
      return freeVarsCouple(code,env);
    case E_TUPLE_GNRL: case E_UNSHARED_TUPLE_GNRL:
    case E_RECORD_GNRL: case E_UNSHARED_RECORD_GNRL:
      return freeVarsVec(code,env);
    case E_LET:
      return freeVarsLet(LETBIND(code),LETBODY(code),env,freeVars);
    case E_PROG_OF: case E_LIST_OF: case E_UNSHARED_LIST_OF:
    case E_SET_OF: case E_APPEND_OF: case E_UNSHARED_APPEND_OF:
    case E_UNION_OF: case E_UNION_UNDER_OF: case E_INTER_OF:
    case E_MAP_OF: case E_MAP_UNDER_OF: case E_ALL_OF:
    case E_EXISTS_OF: case E_SOME_OF: case E_UNSHARED_SOME_OF:
    case E_PROG_MAP: case E_LIST_MAP: case E_UNSHARED_LIST_MAP:
    case E_SET_MAP: case E_APPEND_MAP: case E_UNSHARED_APPEND_MAP:
    case E_UNION_MAP: case E_UNION_UNDER_MAP: case E_INTER_MAP: case E_MAP_MAP:
    case E_MAP_UNDER_MAP: case E_ALL_MAP: case E_EXISTS_MAP:
    case E_SOME_MAP: case E_UNSHARED_SOME_MAP:
      return freeVarsQuant(CDR(OFDOM(code)),OFFP(code),env);
    case E_PACK:
      code = PACKVAL(code);
      goto again;
    case E_RAISE:
      code = RBODY(code);
      goto again;
    case E_DEBUGZONE:
      code = DZCODE(code);
      goto again;
    case E_PROFILE:
      code = CAR(code);
      goto again;
    default:
      kinderror(message(663),code->kind);
      /*NOTREACHED*/
      return false;
    }
}

/* Testing whether an expression needs to allocate its environment in the heap: */

PRIVATE Boolean needEnvVec(e,env)
IN CONST ExprPtr e;
INOUT compEnv *env;
{
  COUNTER i,n;

  for (i=0, n=AVAL(e)->n; i<n; i++)
    {
      if (needEnv(AVAL(e)->expr[i],env))
	return true;
    }
  return false;
}

DEF_PRIVATE_ACTION_1ARG(needenvpat)
{
  struct needenv *data;

  data = (struct needenv *)ENV_UNDERLYING(env);
  data->res = needEnvPattern(data->code,data->env);
  return NIL;
}
ACTION_END

PRIVATE Boolean needEnvPat1(pat,env)
IN CONST ExprPtr pat;
INOUT compEnv *env;
{
  struct needenv *d = (struct needenv *)alloctmp(sizeof(struct needenv));

  d->code = pat;
  d->env = env;
  (void) CallCC(ACTION_NAME(needenvpat),(ExprPtr)d,true);
  return d->res;
}

PUBLIC Boolean needEnvPattern(pat,env)
IN CONST ExprPtr pat;
INOUT compEnv *env;
{
  if (STKFULL(pat))
    return needEnvPat1(pat,env);
again:
  switch (pat->kind)
    {
    case E_VARINTRO:
    case E_NIL: case E_TUPLE_0: case E_RECORD_0:
    case E_EMPTY: case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
    case E_EXCEPTION:
    case E_COMPILED_CLOSURE: case E_INTERPRETED_CLOSURE:
#ifdef OBSOLETE
    case E_MEMO_CLOSURE:
#endif
    case E_QUOTE: case E_VARREF: case E_PAT_ANY:
      return false;
    case E_DEBUGNAME:
      pat = DNPAT(pat);
      goto again;
    case E_PATCHOICE:
      if (needEnvPattern(CAR(CDR(pat)),env))
	return true;
      env->level++;
      if (needEnv(CDR(CDR(pat)),env))
	{
	  env->level--;
	  return true;
	}
      env->level--;
      pat = CAR(pat);
      goto again;
    case E_CONS: case E_UNSHARED_CONS:
    case E_MAPLET: case E_PAT_REVCONS: case E_PAT_CONS:
    case E_TUPLE_2: case E_UNSHARED_TUPLE_2:
    case E_PAT_TUPLE_GNRL_CONS: case E_PAT_REVTUPLE_GNRL_CONS:
    case E_PAT_REVTUPLE_2:
    case E_PAT_RECORD_GNRL_CONS: case E_PAT_REVRECORD_GNRL_CONS:
    case E_APPL: case E_PAT_AS:
    case E_PAT_RECORD_MAPADD: case E_PAT_MAPADD:
      if (needEnvPattern(CAR(pat),env))
	return true;
      pat = CDR(pat);
      goto again;
#ifdef DATACONPAT
    case E_DATACON:
#endif
    case E_TUPLE_1: case E_UNSHARED_TUPLE_1: case E_REF:
      pat = REFVAL(pat);
      goto again;
    case E_DATA:
    case E_RECORD_1: case E_UNSHARED_RECORD_1:
      pat = CDR(pat);
      goto again;
    case E_PAT_CHECK:
      pat = PATCHECKVAL(pat);
      goto again;
    case E_PAT_PACK:
      pat = PACKVAL(pat);
      goto again;
    case E_PAT_CARD_EQ: case E_PAT_CARD_GEQ:
      pat = COUNTEDVAL(pat);
      goto again;
    default:
      kinderror(message(642),pat->kind);
      /*NOTREACHED*/
      return false;
    }
}

#ifdef OBSOLETE
MAPSWEEPFUNC(do_needenv)
{
  compEnv *env;

  env = MAPDATA(compEnv *);
  return needEnv(x,env) || needEnv(y,env);
}

PRIVATE Boolean needEnvMap(code,env)
IN CONST ExprPtr code;
INOUT compEnv *env;
{
  return MapSweep(code,do_needenv,env);
}
#endif

PRIVATE Boolean needEnvLet(bindings,body,env,fun)
IN CONST ExprPtr bindings,body;
INOUT compEnv *env;
IN CONST needEnvFun fun;
{
  COUNTER n,savelevel,saveglob;
  Boolean needenv;

  savelevel = env->level;
  saveglob = env->globenvs;
  n = IVAL(CAR(bindings));
  if (n<0)
    {
      env->level++;
      env->globenvs |= (1L << env->level);
    }
  env->level++;

  needenv = needEnvBindings(CDR(bindings),env);
  if (!needenv)
    needenv = (*fun)(body,env);
  env->level = savelevel;
  env->globenvs = saveglob;
  return needenv;
}

PUBLIC Boolean needEnvBindings(bindlist,env)
IN CONST ExprPtr bindlist;
INOUT compEnv *env;
{
  ExprPtr binding;

  while (bindlist!=NIL)
    {
      binding = CAR(bindlist);
      switch (binding->kind)
	{
	case E_LET:
	  return needEnvLet(LETBIND(binding),LETBODY(binding),
			    env,needEnvBindings);
	case E_DEBUGDECL:
	  break;
	default:
	  if (needEnv(CDR(binding),env))
	    return true;
	  break;
	}
      bindlist = CDR(bindlist);
    }
  return false;
}

PUBLIC Boolean needEnvDoms(domlist,env)
IN CONST ExprPtr domlist;
INOUT compEnv *env;
{
  while (domlist!=NIL)
    switch (domlist->kind)
      {
      case E_PAT_RECORD_GNRL_CONS:
      case E_PAT_REVRECORD_GNRL_CONS:
      case E_PAT_RECORD_MAPADD:
	if (needEnv(CDR(CAR(domlist)),env))
	  return true;
	domlist = CDR(domlist);
	break;
      case E_TUPLE_1:
	{
	  env->level++;
	  if (needEnvBindings(REFVAL(domlist),env))
	    {
	      env->level--;
	      return true;
	    }
	  env->level--;
	  domlist = NIL;
	}
	break;
      }
  return false;
}

PRIVATE Boolean needEnvFunc(fun,env,arglist)
IN CONST ExprPtr fun;
INOUT compEnv *env;
IN CONST ExprPtr arglist;
{
  ExprPtr body;

  while (true)
    switch (fun->kind)
      {
      case E_PAT_AS:
	fun = CDR(fun); /* This one is always inlined (at a call site) */
	break;
      case E_POLYFUN:
	/*case E_N_LAMBDA:*/
polyfun:
	fun = COUNTEDVAL(fun);
	/*FALLTHROUGH*/
	if (needEnvPattern(CAR(fun),env))
	  return true;
	env->level++;
	body = CDR(fun);
	if (needEnv(CDR(fun),env))
	  {
	    env->level--;
	    if (arglist!=NIL && !hasChoice(CAR(fun)))
	      {
reapply:
		switch (body->kind)
		  {
		  case E_DEBUGZONE:
		    body = DZCODE(body);
		    goto reapply;
		  case E_PROFILE:
		    body = CAR(body);
		    goto reapply;
		  case E_PAT_AS:
		    body = CDR(body);
		    goto reapply;
		  case E_POLYFUN:
		    /*case E_N_LAMBDA:*/
		    arglist = CDR(arglist);
		    fun = body;
		    goto polyfun;
		  }
	      }
	    return true;
	  }
	env->level--;
	return false;
      case E_MEMO_FN:
	fun = MEMOFUN(fun);
	break;
      case E_DELAY:
	fun = DCODE(fun);
	break;
      case E_PROFILE:
	fun = CAR(fun);
	break;
      case E_APPL: case E_GOTO:
	/*case E_N_APPL:*/
	if (needEnv(APPLARG(fun),env))
	  return true;
	arglist = MakePair(E_CONS,APPLARG(fun),arglist);
	(void) KEEP(arglist);
	fun = APPLFUN(fun);
	break;
      case E_VARREF: /* shortcut for frequent case */
	return false;
      default:
	return needEnv(fun,env);
      }
}

PRIVATE Boolean needEnvQuant(doms,fp,env)
IN CONST ExprPtr doms,fp;
INOUT compEnv *env;
{
  if (needEnvDoms(doms,env))
    return true;
  env->level++;
  if (needEnv(fp,env))
    {
      env->level--;
      return true;
    }
  env->level--;
  return false;
}

DEF_PRIVATE_ACTION_1ARG(needenv)
{
  struct needenv *data;

  data = (struct needenv *)ENV_UNDERLYING(env);
  data->res = needEnv(data->code,data->env);
  return NIL;
}
ACTION_END

PRIVATE Boolean needEnv1(code,env)
IN CONST ExprPtr code;
INOUT compEnv *env;
{
  struct needenv *d = (struct needenv *)alloctmp(sizeof(struct needenv));

  d->code = code;
  d->env = env;
  (void) CallCC(ACTION_NAME(needenv),(ExprPtr)d,true);
  return d->res;
}

PUBLIC Boolean needEnv(code,env)
IN CONST ExprPtr code;
INOUT compEnv *env;
{
  if (STKFULL(code))
    return needEnv1(code,env);
again:
  switch (code->kind)
    {
    case E_NIL: case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
    case E_EMPTY: case E_TUPLE_0: case E_RECORD_0:
    case E_ENV: case E_STACK:
    case E_COMPILED_CLOSURE: case E_INTERPRETED_CLOSURE:
#ifdef OBSOLETE
    case E_MEMO_CLOSURE:
#endif
    case E_DATACON: case E_QUOTE: case E_VARREF: case E_RAISE: case E_EXCEPTION:
      return false;
    case E_MEMO_FN:
      code = MEMOFUN(code);
      goto again;
    case E_DELAY:
      code = DCODE(code);
      goto again;
    case E_BYTECODE:
      return true;
    case E_POLYFUN:
      /*case E_N_LAMBDA:*/
      return freeVars(code,env)!=NewMap();
    case E_REF: case E_TUPLE_1: case E_UNSHARED_TUPLE_1:
    case E_GET: case E_NOT:
      code = REFVAL(code);
      goto again;
    case E_RECORD_1: case E_UNSHARED_RECORD_1:
    case E_TSELECT: case E_TSEL1: case E_TSEL2: case E_TSELGNRL:
    case E_RSELECT: case E_RSEL1: case E_RSELGNRL:
      code = CDR(code);
      goto again;
    case E_PAT_AS:
      /* In pat_as(var,func), we don't know whether we shall inline (use fun)
	 or not (use var), so we have to be conservative, and decide that
	 we need an environment if we need one in one of these two cases. */
    case E_TUPLE_2: case E_UNSHARED_TUPLE_2:
    case E_CONS: case E_UNSHARED_CONS:
    case E_MAPLET: case E_OVERWRITE: case E_UNDERWRITE:
    case E_SET: case E_MAPGET:
    case E_RMU1: case E_UNSHARED_RMU1:
    case E_RMUX: case E_UNSHARED_RMUX:
    case E_IF: case E_PROG: case E_AND: case E_OR:
    case E_EQUAL: case E_NEQUAL:
      if (needEnv(CAR(code),env))
	return true;
      code = CDR(code);
      goto again;
    case E_TUPLE_GNRL: case E_UNSHARED_TUPLE_GNRL:
    case E_RECORD_GNRL: case E_UNSHARED_RECORD_GNRL:
      return needEnvVec(code,env);
    case E_LET:
      return needEnvLet(LETBIND(code),LETBODY(code),env,needEnv);
    case E_APPL: case E_GOTO:
      /*case E_N_APPL:*/
      if (needEnvFunc(APPLFUN(code),env,NIL))
	return true;
      code = APPLARG(code);
      goto again;
    case E_PROG_OF: case E_LIST_OF: case E_UNSHARED_LIST_OF:
    case E_SET_OF: case E_APPEND_OF: case E_UNSHARED_APPEND_OF:
    case E_UNION_OF: case E_UNION_UNDER_OF: case E_INTER_OF:
    case E_MAP_OF: case E_MAP_UNDER_OF: case E_ALL_OF:
    case E_EXISTS_OF: case E_SOME_OF: case E_UNSHARED_SOME_OF:
    case E_PROG_MAP: case E_LIST_MAP: case E_UNSHARED_LIST_MAP:
    case E_SET_MAP: case E_APPEND_MAP: case E_UNSHARED_APPEND_MAP:
    case E_UNION_MAP: case E_UNION_UNDER_MAP: case E_INTER_MAP: case E_MAP_MAP:
    case E_MAP_UNDER_MAP: case E_ALL_MAP: case E_EXISTS_MAP:
    case E_SOME_MAP: case E_UNSHARED_SOME_MAP:
      return needEnvQuant(CDR(OFDOM(code)),OFFP(code),env);
    case E_PROG_WHILE: case E_LIST_WHILE: case E_UNSHARED_LIST_WHILE:
    case E_SET_WHILE: case E_APPEND_WHILE: case E_UNSHARED_APPEND_WHILE:
    case E_UNION_WHILE: case E_UNION_UNDER_WHILE: case E_INTER_WHILE:
    case E_MAP_WHILE: case E_MAP_UNDER_WHILE: case E_ALL_WHILE:
    case E_EXISTS_WHILE: case E_SOME_WHILE: case E_UNSHARED_SOME_WHILE:
      if (needEnv(WCOND(code),env))
	return true;
      code = WBODY(code);
      goto again;
    case E_HANDLE:
      if (needEnv(HBODY(code),env))
	return true;
      return needEnvFunc(HHANDLERS(code),env,NIL);
    case E_PACK:
      code = PACKVAL(code);
      goto again;
    case E_DEBUGZONE:
      code = DZCODE(code);
      goto again;
    case E_PROFILE:
      code = CAR(code);
      goto again;
    default:
      kinderror(message(632),code->kind);
      /*NOTREACHED*/
      return false;
    }
}
#endif
