/* Directory-related functions.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#include "exeproto.h"
#include "dirproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <unistd.h>
#ifdef HAS_DIRENT
#include <dirent.h>
#endif
#else
#if OS==AMIGA_OS
#undef SHORT
#include <dos.h>
#include <sys/dir.h>
#include <sys/stat.h>
#define HAS_DIRENT
#include <dos/dos.h>
#include <proto/dos.h>
#endif
#include <errno.h>
#endif
#ifdef HAS_PROTOTYPES
#include "sdir.h"
#endif

EXTERN ExprPtr voidString;

#if OS==UNIX_SYS_V || OS==UNIX_BSD || OS==AMIGA_OS
IMPORT int wait_pid P((pid_type pid));

/* return set of file names in the directory specified by dirname.
   Note that it does not KEEP its result. */
PUBLIC MapPtr listDir(dirname)
IN CONST CHAR *dirname;
{
  MapPtr dir;
  ExprPtr name;
#ifdef HAS_DIRENT
  DIR *dirp;
  struct dirent *dp;
  ENTER;

  dir = NewMap();
  dirp = opendir((char *)dirname);
  if (dirp==NULL)
    return raiseIO();
  while (errno=0, dp=readdir(dirp), dp!=NULL)
    {
      if (dp->d_name[0]=='.' && (dp->d_name[1]==0 ||
				 (dp->d_name[1]=='.' && dp->d_name[2]==0)))
	continue; /* ignore "." and ".." */
      name = Kstring((INT)strlen(dp->d_name),(CHAR *)dp->d_name);
      (void) KEEP(name);
      dir = SetAdd(name,dir);
      KEEPONLY(dir);
    }
  if (errno!=0)
    {
      int saveerrno;

      saveerrno = errno;
      (void) closedir(dirp);
      errno = saveerrno;
      return raiseIO();
    }
  closedir(dirp);
#else
  pid_type pid;
  int redirect[2];
  char *argv[5];
  ENTER;

  /* otherwise, we simulate all this by exec'ing an ls -A -b -1 command
     (-A prints all files in the directory,  -1 lists one file per line) */
  /* popen() might be easier to use if we didn't have to reimplement the stdio
     library to avoid an obscure bug on the DPX20 (because we use our own
     sbrk()-based memory management scheme, we have to override libc.a's malloc()
     so that functions like those in stdio won't crash the machine; on the DPX20,
     such malloc() calls inside stdio functions are inlined [severe brain damage!],
     so that we cannot override them unless we use a wholly different stdio library). */
  dir = NewMap();
  if (pipe(redirect))
    return raiseIO();
  switch (pid = fork())
    {
    case -1:
      {
	int saveerrno;

	saveerrno = errno;
	(void) close(redirect[0]);
	(void) close(redirect[1]);
	errno = saveerrno;
      }
      return raiseIO();
    case 0: /* child */
      argv[0] = "/bin/ls";
      argv[1] = "-A";
      argv[2] = "-1";
      argv[3] = dirname;
      argv[4] = NULL;
      if (close(redirect[0]) || close(1) || dup2(redirect[1],1)==-1 || close(redirect[1]))
	/* redirect output */
	exit(errno);
      (void) execv(argv[0],argv);
      exit(errno);
    default: /* parent */
      (void) close(redirect[1]);
      {
#define BUFS 512
	int size;
	char *buf;
	CHAR *buff,*s,*t,*tend;

	buf = (char *)alloctmp(2*BUFS);
	buff = (CHAR *)(buf+BUFS);
	s = buff;
	while (size = read(redirect[0],buf,BUFS))
	  {
	    t = (CHAR *)buf;
	    tend = t+size;
	    while (t<tend)
	      switch (*t)
		{
		case '\n':
		  if (s!=buff)
		    {
		      name = Kstring((INT)(s-buff),(CHAR *)buff);
		      (void) KEEP(name);
		      dir = SetAdd(name,dir);
		      KEEPONLY(dir);
		    }
		  s = buff;
		  t++;
		  break;
		default:
		  *s++ = *t++;
		  break;
		}
	  }
	if (s!=buff)
	  {
	    name = Kstring((INT)(s-buff),(CHAR *)buff);
	    (void) KEEP(name);
	    dir = SetAdd(name,dir);
	    KEEPONLY(dir);
	  }
	(void) close(redirect[0]);
	errno = 0;
	size = wait_pid(pid);
	if (errno!=0 || (size!=0 && (errno=(size>>8))))
	  raiseIO();
#undef BUFS
      }
      break;
    }
#endif
  QUIT;
  return dir;
}

PUBLIC int changeDir(dirname)
IN CONST CHAR *dirname;
{
  return chdir((char *)dirname);
}

PUBLIC CHAR *getDir(buf,size)
INOUT CHAR *buf;
IN COUNTER size;
{
#ifdef HAS_CWD
  return (CHAR *)getcwd((char *)buf,(int)size);
#else
#ifdef HAS_WD
  char *buf2;
  int len;

#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif

  buf2 = (char *)alloctmp(MAXPATHLEN+1);
  if (getwd(buf2) == NULL) {
    /*
     * There's not much we can do besides guess an errno to
     * use for the result (the error message in buf2 isn't
     * much use...).
     */
    errno = EACCES;
    return NULL;
  }
  len = strlen(buf2);
  if (len >= size) {
    errno = ERANGE;
    return NULL;
  }
  strcpy((char *)buf, buf2);
  return buf;
#else
  buf[0] = 0;
  return buf;
#endif
#endif
}

PUBLIC int makeDir(dirname)
IN CONST CHAR *dirname;
{
  return mkdir((char *)dirname
#if OS==UNIX_SYS_V || OS==UNIX_BSD
	       ,0666
#endif
	       );
}

PUBLIC int rmDir(dirname)
IN CONST CHAR *dirname;
{
  return rmdir((char *)dirname);
}

/* return set of strings telling which properties the file whose name is given has.
   Note that it does not KEEP its result. */
PUBLIC MapPtr fileBits(name)
IN CONST CHAR *name;
{
  MapPtr bits;
  CHAR c;
#if OS==AMIGA_OS
  BPTR lock;
  __aligned struct FileInfoBlock fib;
  COUNTER i;
  ENTER;

  lock = Lock(name,ACCESS_READ);
  if (lock==0)
    return raiseIO();
  if (Examine(lock,&fib)==0)
    {
      UnLock(lock);
      return raiseIO();
    }
  UnLock(lock);
  bits = NewMap();
  if (fib.fib_DirEntryType>0)
    c = 'd'; /* directory */
  else
    c = 'n';
  bits = SetAdd(Kstring(1L,&c),bits);
  for (i=0; i<7; i++)
    {
      if (fib.fib_Protection & (1 << i))
	bits = SetAdd(Kstring(1L,"exwraps"+i),bits);
    }
#else
  struct stat st;
  CHAR cc[2];
  ENTER;

  if (stat((char *)name,&st))
    return raiseIO();
  bits = NewMap();
#ifdef S_ISDIR
  if (S_ISDIR(st.st_mode))
    { c = 'd'; bits = SetAdd(Kstring(1L,&c),bits); }
#else
#ifdef S_IFDIR
  if ((st.st_mode & S_IFDIR)==S_IFDIR)
    { c = 'd'; bits = SetAdd(Kstring(1L,&c),bits); }
#endif
#endif
#ifdef S_ISBLK
  if (S_ISBLK(st.st_mode))
    { c = 'b'; bits = SetAdd(Kstring(1L,&c),bits); }
#else
#ifdef S_IFBLK
  if ((st.st_mode & S_IFBLK)==S_IFBLK)
    { c = 'b'; bits = SetAdd(Kstring(1L,&c),bits); }
#endif
#endif
#ifdef S_ISCHR
  if (S_ISCHR(st.st_mode))
    { c = 'c'; bits = SetAdd(Kstring(1L,&c),bits); }
#else
#ifdef S_IFCHR
  if ((st.st_mode & S_IFCHR)==S_IFCHR)
    { c = 'c'; bits = SetAdd(Kstring(1L,&c),bits); }
#endif
#endif
#ifdef S_ISFIFO
  if (S_ISFIFO(st.st_mode))
    { c = 'f'; bits = SetAdd(Kstring(1L,&c),bits); }
#else
#ifdef S_IFFIFO
  if ((st.st_mode & S_IFFIFO)==S_IFFIFO)
    { c = 'f'; bits = SetAdd(Kstring(1L,&c),bits); }
#endif
#endif
#ifdef S_ISREG
  if (S_ISREG(st.st_mode))
    { c = 'n'; bits = SetAdd(Kstring(1L,&c),bits); }
#else
#ifdef S_IFREG
  if (st.st_mode & S_IFREG)
    { c = 'n'; bits = SetAdd(Kstring(1L,&c),bits); }
#endif
#endif
#ifdef S_ISCRIPT
  if ((st.st_mode & S_ISCRIPT)==S_ISCRIPT)
    { c = 's'; bits = SetAdd(Kstring(1L,&c),bits); }
#endif
#ifdef S_IPURE
  if ((st.st_mode & S_IPURE)==S_IPURE)
    { c = 'p'; bits = SetAdd(Kstring(1L,&c),bits); }
#endif
#ifdef S_IARCHIVE
  if ((st.st_mode & S_IARCHIVE)==S_IARCHIVE)
    { c = 'a'; bits = SetAdd(Kstring(1L,&c),bits); }
#endif
#ifdef S_IRUSR
  if ((st.st_mode & S_IRUSR)==S_IRUSR)
    { c = 'r'; bits = SetAdd(Kstring(1L,&c),bits); }
#else
#ifdef S_IREAD
  if ((st.st_mode & S_IREAD)==S_IREAD)
    { c = 'r'; bits = SetAdd(Kstring(1L,&c),bits); }
#endif
#endif
#ifdef S_IWUSR
  if ((st.st_mode & S_IWUSR)==S_IWUSR)
    { c = 'w'; bits = SetAdd(Kstring(1L,&c),bits);
      c = 'e'; bits = SetAdd(Kstring(1L,&c),bits); }
#else
#ifdef S_IWRITE
  if ((st.st_mode & S_IWRITE)==S_IWRITE)
    { c = 'w'; bits = SetAdd(Kstring(1L,&c),bits);
      c = 'e'; bits = SetAdd(Kstring(1L,&c),bits); }
#endif
#endif
#ifdef S_IXUSR
  if ((st.st_mode & S_IXUSR)==S_IXUSR)
    { c = 'x'; bits = SetAdd(Kstring(1L,&c),bits); }
#else
#ifdef S_IEXECUTE
  if ((st.st_mode & S_IEXECUTE)==S_IEXECUTE)
    { c = 'x'; bits = SetAdd(Kstring(1L,&c),bits); }
#endif
#endif
#ifdef S_IDELETE
  if ((st.st_mode & S_IDELETE)==S_IDELETE)
    { c = 'e'; bits = SetAdd(Kstring(1L,&c),bits); }
#endif
  cc[1] = 'g';
#ifdef S_IRGRP
  if ((st.st_mode & S_IRGRP)==S_IRGRP)
    { cc[0] = 'r'; bits = SetAdd(Kstring(2L,cc),bits); }
#endif
#ifdef S_IWGRP
  if ((st.st_mode & S_IWGRP)==S_IWGRP)
    { cc[0] = 'w'; bits = SetAdd(Kstring(2L,cc),bits);
      cc[0] = 'e'; bits = SetAdd(Kstring(2L,cc),bits); }
#endif
#ifdef S_IXGRP
  if ((st.st_mode & S_IXGRP)==S_IXGRP)
    { cc[0] = 'x'; bits = SetAdd(Kstring(2L,cc),bits); }
#endif
  cc[1] = 'o';
#ifdef S_IROTH
  if ((st.st_mode & S_IROTH)==S_IROTH)
    { cc[0] = 'r'; bits = SetAdd(Kstring(2L,cc),bits); }
#endif
#ifdef S_IWOTH
  if ((st.st_mode & S_IWOTH)==S_IWOTH)
    { cc[0] = 'w'; bits = SetAdd(Kstring(2L,cc),bits);
      cc[0] = 'e'; bits = SetAdd(Kstring(2L,cc),bits); }
#endif
#ifdef S_IXOTH
  if ((st.st_mode & S_IXOTH)==S_IXOTH)
    { cc[0] = 'x'; bits = SetAdd(Kstring(2L,cc),bits); }
#endif
#ifdef S_ISUID
  if ((st.st_mode & S_ISUID)==S_ISUID)
    { c = 'u'; bits = SetAdd(Kstring(1L,&c),bits); }
#endif
#ifdef S_ISGID
  if ((st.st_mode & S_ISGID)==S_ISGID)
    { c = 'g'; bits = SetAdd(Kstring(1L,&c),bits); }
#endif
#ifdef S_ISVTX
  if ((st.st_mode & S_ISGID)==S_ISGID)
    { c = 'v'; bits = SetAdd(Kstring(1L,&c),bits); }
#endif
#ifdef S_IFLNK
  if ((st.st_mode & S_IFLNK)==S_IFLNK)
    { c = 'l'; bits = SetAdd(Kstring(1L,&c),bits); }
#endif
#ifdef S_IFSOCK
  if ((st.st_mode & S_IFSOCK)==S_IFSOCK)
    { c = 'S'; bits = SetAdd(Kstring(1L,&c),bits); }
#endif
#endif
  QUIT;
  return bits;
}
#endif

#ifdef HAS_SYSERRLIST
/*extern char *sys_errlist[ ];*/
/*extern int sys_nerr;*/
#endif

PUBLIC char *ioerrormsg(code)
IN COUNTER code;
{
  char *err;

#ifdef HAS_STRERROR
  err = strerror((int)code);
#else
#ifdef HAS_SYSERROR
    if ((code <= sys_nerr) && (code > 0)) {
	return (char *)sys_errlist[error];
    }
#else
    switch (code) {
#ifdef E2BIG
	case E2BIG: err = "argument list too long"; break;
#endif
#ifdef EACCES
	case EACCES: err = "permission denied"; break;
#endif
#ifdef EADDRINUSE
	case EADDRINUSE: err = "address already in use"; break;
#endif
#ifdef EADDRNOTAVAIL
	case EADDRNOTAVAIL: err = "can't assign requested address"; break;
#endif
#ifdef EADV
	case EADV: err = "advertise error"; break;
#endif
#ifdef EAFNOSUPPORT
	case EAFNOSUPPORT: err = "address family not supported by protocol family"; break;
#endif
#ifdef EAGAIN
	case EAGAIN: err = "no more processes"; break;
#endif
#ifdef EALIGN
	case EALIGN: err = "EALIGN"; break;
#endif
#ifdef EALREADY
	case EALREADY: err = "operation already in progress"; break;
#endif
#ifdef EBADE
	case EBADE: err = "bad exchange descriptor"; break;
#endif
#ifdef EBADF
	case EBADF: err = "bad file number"; break;
#endif
#ifdef EBADFD
	case EBADFD: err = "file descriptor in bad state"; break;
#endif
#ifdef EBADMSG
	case EBADMSG: err = "not a data message"; break;
#endif
#ifdef EBADR
	case EBADR: err = "bad request descriptor"; break;
#endif
#ifdef EBADRPC
	case EBADRPC: err = "RPC structure is bad"; break;
#endif
#ifdef EBADRQC
	case EBADRQC: err = "bad request code"; break;
#endif
#ifdef EBADSLT
	case EBADSLT: err = "invalid slot"; break;
#endif
#ifdef EBFONT
	case EBFONT: err = "bad font file format"; break;
#endif
#ifdef EBUSY
	case EBUSY: err = "mount device busy"; break;
#endif
#ifdef ECHILD
	case ECHILD: err = "no children"; break;
#endif
#ifdef ECHRNG
	case ECHRNG: err = "channel number out of range"; break;
#endif
#ifdef ECOMM
	case ECOMM: err = "communication error on send"; break;
#endif
#ifdef ECONNABORTED
	case ECONNABORTED: err = "software caused connection abort"; break;
#endif
#ifdef ECONNREFUSED
	case ECONNREFUSED: err = "connection refused"; break;
#endif
#ifdef ECONNRESET
	case ECONNRESET: err = "connection reset by peer"; break;
#endif
#ifdef EDEADLK
#ifndef EWOULDBLOCK
#define RDLKAVOID
#endif
#if EDEADLK != EWOULDBLOCK
#define RDLKAVOID
#endif
#endif
#ifdef RDLKAVOID
	case EDEADLK: err = "resource deadlock avoided"; break;
#endif
#undef RDLKAVOID
#ifdef EDEADLOCK
#define RDLKAVOID
#ifdef EWOULDBLOCK
#if EDEADLOCK==EWOULDBLOCK
#undef RDLKAVOID
#endif
#endif
#ifdef EDEADLK
#if EDEADLOCK==EDEADLK
#undef RDLKAVOID
#endif
#endif
#endif
#ifdef RDLKAVOID
	case EDEADLOCK: err = "resource deadlock avoided"; break;
#endif
#undef RDLKAVOID
#ifdef EDESTADDRREQ
	case EDESTADDRREQ: err = "destination address required"; break;
#endif
#ifdef EDIRTY
	case EDIRTY: err = "mounting a dirty fs w/o force"; break;
#endif
#ifdef EDOM
	case EDOM: err = "math argument out of range"; break;
#endif
#ifdef EDOTDOT
	case EDOTDOT: err = "cross mount point"; break;
#endif
#ifdef EDQUOT
	case EDQUOT: err = "disk quota exceeded"; break;
#endif
#ifdef EDUPPKG
	case EDUPPKG: err = "duplicate package name"; break;
#endif
#ifdef EEXIST
	case EEXIST: err = "file already exists"; break;
#endif
#ifdef EFAULT
	case EFAULT: err = "bad address in system call argument"; break;
#endif
#ifdef EFBIG
	case EFBIG: err = "file too large"; break;
#endif
#ifdef EHOSTDOWN
	case EHOSTDOWN: err = "host is down"; break;
#endif
#ifdef EHOSTUNREACH
	case EHOSTUNREACH: err = "host is unreachable"; break;
#endif
#ifdef EIDRM
	case EIDRM: err = "identifier removed"; break;
#endif
#ifdef EINIT
	case EINIT: err = "initialization error"; break;
#endif
#ifdef EINPROGRESS
	case EINPROGRESS: err = "operation now in progress"; break;
#endif
#ifdef EINTR
	case EINTR: err = "interrupted system call"; break;
#endif
#ifdef EINVAL
	case EINVAL: err = "invalid argument"; break;
#endif
#ifdef EIO
	case EIO: err = "I/O error"; break;
#endif
#ifdef EISCONN
	case EISCONN: err = "socket is already connected"; break;
#endif
#ifdef EISDIR
	case EISDIR: err = "illegal operation on a directory"; break;
#endif
#ifdef EISNAME
	case EISNAM: err = "is a name file"; break;
#endif
#ifdef ELBIN
	case ELBIN: err = "ELBIN"; break;
#endif
#ifdef EL2HLT
	case EL2HLT: err = "level 2 halted"; break;
#endif
#ifdef EL2NSYNC
	case EL2NSYNC: err = "level 2 not synchronized"; break;
#endif
#ifdef EL3HLT
	case EL3HLT: err = "level 3 halted"; break;
#endif
#ifdef EL3RST
	case EL3RST: err = "level 3 reset"; break;
#endif
#ifdef ELIBACC
	case ELIBACC: err = "can not access a needed shared library"; break;
#endif
#ifdef ELIBBAD
	case ELIBBAD: err = "accessing a corrupted shared library"; break;
#endif
#ifdef ELIBEXEC
	case ELIBEXEC: err = "can not exec a shared library directly"; break;
#endif
#ifdef ELIBMAX
	case ELIBMAX: err =
		"attempting to link in more shared libraries than system limit"; break;
#endif
#ifdef ELIBSCN
	case ELIBSCN: err = ".lib section in a.out corrupted"; break;
#endif
#ifdef ELNRNG
	case ELNRNG: err = "link number out of range"; break;
#endif
#ifdef ELOOP
	case ELOOP: err = "too many levels of symbolic links"; break;
#endif
#ifdef EMFILE
	case EMFILE: err = "too many open files"; break;
#endif
#ifdef EMLINK
	case EMLINK: err = "too many links"; break;
#endif
#ifdef EMSGSIZE
	case EMSGSIZE: err = "message too long"; break;
#endif
#ifdef EMULTIHOP
	case EMULTIHOP: err = "multihop attempted"; break;
#endif
#ifdef ENAMETOOLONG
	case ENAMETOOLONG: err = "file name too long"; break;
#endif
#ifdef ENAVAIL
	case ENAVAIL: err = "not available"; break;
#endif
#ifdef ENET
	case ENET: err = "ENET"; break;
#endif
#ifdef ENETDOWN
	case ENETDOWN: err = "network is down"; break;
#endif
#ifdef ENETRESET
	case ENETRESET: err = "network dropped connection on reset"; break;
#endif
#ifdef ENETUNREACH
	case ENETUNREACH: err = "network is unreachable"; break;
#endif
#ifdef ENFILE
	case ENFILE: err = "file table overflow"; break;
#endif
#ifdef ENOANO
	case ENOANO: err = "anode table overflow"; break;
#endif
#ifdef ENOBUFS
#ifndef ENOSR
#define NOBUFSPACE
#endif
#if ENOBUFS != ENOSR
#define NOBUFSPACE
#endif
#endif
#ifdef NOBUFSPACE
	case ENOBUFS: err = "no buffer space available"; break;
#endif
#undef NOBUFSPACE
#ifdef ENOCSI
	case ENOCSI: err = "no CSI structure available"; break;
#endif
#ifdef ENODATA
	case ENODATA: err = "no data available"; break;
#endif
#ifdef ENODEV
	case ENODEV: err = "no such device"; break;
#endif
#ifdef ENOENT
	case ENOENT: err = "no such file or directory"; break;
#endif
#ifdef ENOEXEC
	case ENOEXEC: err = "exec format error"; break;
#endif
#ifdef ENOLCK
	case ENOLCK: err = "no locks available"; break;
#endif
#ifdef ENOLINK
	case ENOLINK: err = "link has be severed"; break;
#endif
#ifdef ENOMEM
	case ENOMEM: err = "not enough memory"; break;
#endif
#ifdef ENOMSG
	case ENOMSG: err = "no message of desired type"; break;
#endif
#ifdef ENONET
	case ENONET: err = "machine is not on the network"; break;
#endif
#ifdef ENOPKG
	case ENOPKG: err = "package not installed"; break;
#endif
#ifdef ENOPROTOOPT
	case ENOPROTOOPT: err = "bad protocol option"; break;
#endif
#ifdef ENOSPC
	case ENOSPC: err = "no space left on device"; break;
#endif
#ifdef ENOSR
	case ENOSR: err = "out of stream resources"; break;
#endif
#ifdef ENOSTR
	case ENOSTR: err = "not a stream device"; break;
#endif
#ifdef ENOSYM
	case ENOSYM: err = "unresolved symbol name"; break;
#endif
#ifdef ENOSYS
	case ENOSYS: err = "function not implemented"; break;
#endif
#ifdef ENOTBLK
	case ENOTBLK: err = "block device required"; break;
#endif
#ifdef ENOTCONN
	case ENOTCONN: err = "socket is not connected"; break;
#endif
#ifdef ENOTDIR
	case ENOTDIR: err = "not a directory"; break;
#endif
#ifdef ENOTEMPTY
	case ENOTEMPTY: err = "directory not empty"; break;
#endif
#ifdef ENOTNAM
	case ENOTNAM: err = "not a name file"; break;
#endif
#ifdef ENOTSOCK
	case ENOTSOCK: err = "socket operation on non-socket"; break;
#endif
#ifdef ENOTTY
	case ENOTTY: err = "inappropriate device for ioctl"; break;
#endif
#ifdef ENOTUNIQ
	case ENOTUNIQ: err = "name not unique on network"; break;
#endif
#ifdef ENXIO
	case ENXIO: err = "no such device or address"; break;
#endif
#ifdef EOPNOTSUPP
	case EOPNOTSUPP: err = "operation not supported on socket"; break;
#endif
#ifdef EPERM
	case EPERM: err = "not owner"; break;
#endif
#ifdef EPFNOSUPPORT
	case EPFNOSUPPORT: err = "protocol family not supported"; break;
#endif
#ifdef EPIPE
	case EPIPE: err = "broken pipe"; break;
#endif
#ifdef EPROCLIM
	case EPROCLIM: err = "too many processes"; break;
#endif
#ifdef EPROCUNAVAIL
	case EPROCUNAVAIL: err = "bad procedure for program"; break;
#endif
#ifdef EPROGMISMATCH
	case EPROGMISMATCH: err = "program version wrong"; break;
#endif
#ifdef EPROGUNAVAIL
	case EPROGUNAVAIL: err = "RPC program not available"; break;
#endif
#ifdef EPROTO
	case EPROTO: err = "protocol error"; break;
#endif
#ifdef EPROTONOSUPPORT
	case EPROTONOSUPPORT: err = "protocol not suppored"; break;
#endif
#ifdef EPROTOTYPE
	case EPROTOTYPE: err = "protocol wrong type for socket"; break;
#endif
#ifdef ERANGE
	case ERANGE: err = "math result unrepresentable"; break;
#endif
#ifdef EREFUSED
#ifndef ECONNREFUSED
#define CONNXREFUSED
#endif
#if EREFUSED != ECONNREFUSED
#define CONNXREFUSED
#endif
#endif
#ifdef CONNXREFUSED
	case EREFUSED: err = "EREFUSED"; break;
#endif
#undef CONNXREFUSED
#ifdef EREMCHG
	case EREMCHG: err = "remote address changed"; break;
#endif
#ifdef EREMDEV
	case EREMDEV: err = "remote device"; break;
#endif
#ifdef EREMOTE
	case EREMOTE: err = "pathname hit remote file system"; break;
#endif
#ifdef EREMOTEIO
	case EREMOTEIO: err = "remote i/o error"; break;
#endif
#ifdef EREMOTERELEASE
	case EREMOTERELEASE: err = "EREMOTERELEASE"; break;
#endif
#ifdef EROFS
	case EROFS: err = "read-only file system"; break;
#endif
#ifdef ERPCMISMATCH
	case ERPCMISMATCH: err = "RPC version is wrong"; break;
#endif
#ifdef ERREMOTE
	case ERREMOTE: err = "object is remote"; break;
#endif
#ifdef ESHUTDOWN
	case ESHUTDOWN: err = "can't send afer socket shutdown"; break;
#endif
#ifdef ESOCKTNOSUPPORT
	case ESOCKTNOSUPPORT: err = "socket type not supported"; break;
#endif
#ifdef ESPIPE
	case ESPIPE: err = "invalid seek"; break;
#endif
#ifdef ESRCH
	case ESRCH: err = "no such process"; break;
#endif
#ifdef ESRMNT
	case ESRMNT: err = "srmount error"; break;
#endif
#ifdef ESTALE
	case ESTALE: err = "stale remote file handle"; break;
#endif
#ifdef ESUCCESS
	case ESUCCESS: err = "Error 0"; break;
#endif
#ifdef ETIME
	case ETIME: err = "timer expired"; break;
#endif
#ifdef ETIMEDOUT
	case ETIMEDOUT: err = "connection timed out"; break;
#endif
#ifdef ETOOMANYREFS
	case ETOOMANYREFS: err = "too many references: can't splice"; break;
#endif
#ifdef ETXTBSY
	case ETXTBSY: err = "text file or pseudo-device busy"; break;
#endif
#ifdef EUCLEAN
	case EUCLEAN: err = "structure needs cleaning"; break;
#endif
#ifdef EUNATCH
	case EUNATCH: err = "protocol driver not attached"; break;
#endif
#ifdef EUSERS
	case EUSERS: err = "too many users"; break;
#endif
#ifdef EVERSION
	case EVERSION: err = "version mismatch"; break;
#endif
#ifdef EWOULDBLOCK
#ifndef EAGAIN
#define OPWLDBLOCK
#endif
#if EWOULDBLOCK != EAGAIN
#define OPWLDBLOCK
#endif
#endif
#ifdef OPWLDBLOCK
	case EWOULDBLOCK: err = "operation would block"; break;
#endif
#undef OPWLDBLOCK
#ifdef EXDEV
	case EXDEV: err = "cross-domain link"; break;
#endif
#ifdef EXFULL
	case EXFULL: err = "message tables full"; break;
#endif
    }
#endif
#endif
  return err;
}

/* returns a string describing the error whose code is given.
   Note that it does not KEEP its result. */
PUBLIC ExprPtr ioErrorMsg(code)
IN COUNTER code;
{
  char *err;

  err = ioerrormsg(code);
  if (err==NULL)
    return voidString;
  return Kstr((CHAR *)err);
}

#if OS==AMIGA_OS
IMPORT char *init_cd P((void));
IMPORT void end_cd P((char *lock));

PUBLIC char *init_cd()
{
  BPTR pwd;

  pwd = CurrentDir(NULL);
  (void) CurrentDir(pwd);
  return (char *)DupLock(pwd);
}

PUBLIC void end_cd(lock)
IN char *lock;
{
  UnLock(CurrentDir((BPTR)lock));
}
#endif
