/* Type checking.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <math.h>
#include "eval.h"
#include "abssyn.h"
#include "gram.h"
#include "type.h"
#include "expr.h"
#include "mem.h"
#include "grmproto.h"
#include "typproto.h"
#include "prtproto.h"
#include "errproto.h"
#include "zproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "rtype.h"
#ifdef HAS_PROTOTYPES
#include "stype.h"
#endif

#undef DEBUG
#undef DEBUG1
#undef DEBUG_EXPLICIT

EXTERN SymPtr equalSym,equalitySym,discrSym;
EXTERN CHAR *yyfilename;

#if IMPERATIVE_TYPES==EFFECT_TYVAR
PUBLIC ExprPtr theFx; /* contains the computed effect of the expression */
	/* used as an accumulator */
#endif

#define TP(type) TypePrint(type,do_printfile,(char *)hstderr,NIL)
#define EP(expr) valPrint(expr,alpha,do_printfile,(char *)hstderr,0)

	/*****  type-checking functions *****/

#undef __SEG__
#define __SEG__ TYPECHECK

PRIVATE void setq(var,val)
INOUT ExprPtr *var;
IN CONST ExprPtr val;
{
  SETQ(*var,val);
}

#undef SETQ
#define SETQ(var,val) setq(&var,val)

PRIVATE void setcar(cell,val)
INOUT ExprPtr cell;
IN CONST ExprPtr val;
{
  SETCAR(cell,val);
}

#undef SETCAR
#define SETCAR(cell,val) setcar(cell,val)

PRIVATE void setcdr(cell,val)
INOUT ExprPtr cell;
IN CONST ExprPtr val;
{
  SETCDR(cell,val);
}

#undef SETCDR
#define SETCDR(cell,val) setcdr(cell,val)

PRIVATE ExprPtr Cons(a,b)
IN CONST ExprPtr a,b;
{
  ExprPtr res;

  MAKECONS(a,b,res);
  return KEEP(res);
}

#if IMPERATIVE_TYPES==EFFECT_TYVAR
PUBLIC void resetFx()
{
  SETQ(theFx,NewMap());
}

PRIVATE void addFx(fx)
IN CONST ExprPtr fx;
{
  ExprPtr aux;
  ENTER;

  aux = SetAdd(fx,theFx);
  SETQ(theFx,aux);
  QUIT;
}

PRIVATE void trimObsFx(oldassum,type,assum,state,memoref)
IN CONST ExprPtr oldassum;
IN CONST ExprPtr type;
IN CONST ExprPtr assum;
IN CONST ExprPtr state;
INOUT ExprPtr memoref;
{
  MapPtr vars,constr;
  ENTER;

  vars = SetUnion(A_NONGENS(oldassum),
		  tyVars(type,assum,state,0
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			 ,INF_STRENGTH
#endif
			 ,memoref));
  /* keep only these effects that are observable from
     the old environment oldassum or from the type type */
  constr = obsFx(theFx,vars,state);
  SETQ(theFx,constr);
  QUIT;
}

PRIVATE void cvtPatFx()
{
  ExprPtr fx;
  ENTER;

  fx = convertPatFx(theFx);
  SETQ(theFx,fx);
  QUIT;
}
#endif

PRIVATE ExprPtr tcString(expr,assum,state) /* A_STRING, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  return stringType;
}

PRIVATE ExprPtr tcNumber(expr,assum,state) /* A_NUMBER, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  if (CAR(ARGS(expr))==zzero) /* 0 : '#a, and for all other constants x, x : num */
    {
      ExprPtr alpha = freshInstfromFlags(F_NUM,trueValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
					 ,INF_STRENGTH,INF_STRENGTH
#endif
					 );
      ExprPtr type = KnumType (MapSingleton (alpha, one));
      return KEEP (type);
    }
  else return numType;
}

PRIVATE ExprPtr tcNonSense(var,assum,state) /* A_NONSENSE, return type */
IN CONST ExprPtr var;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  /* nonsensical expressions have type 'a (any type) */
  return freshInstfromFlags(0,falseValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			    ,INF_STRENGTH,INF_STRENGTH
#endif
			    );
}

PRIVATE ExprPtr tcVariable(var,assum,state) /* A_VARIABLE, return type */
IN CONST ExprPtr var;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  TZone *zone;
  ExprPtr sym,type;
  ENTER;

  sym = CAR(ARGS(var));
  zone = ZONE(var);
  getLongStruct(zone,CDR(ARGS(var)),&assum,state);
  type = varType(sym,assum,state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
		,false
#endif
		);
  if (type==NOEXPR)
    {
      type_soft_error(zone,(CHAR *)varMsg(message(54),sym));
      /* undefined variable x: then recover from the error, with x:'a,
	 and patch code to raise an exception. */
      TYPE(var) = A_NONSENSE;
      type = tcNonSense(var,assum,state);
    }
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcUnitPower(expr,assum,state)
   /* A_UNIT_POWER, return couple (scale,type) */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr unit,sc_q,exp,args;
  ENTER;

  args = ARGS(expr);
  unit = CAR(args); /* should be a A_UNIT_NAME */
  sc_q = TC(unit,assum,state); /* return (scale,dimension) */
  exp = CAR(ARGS(CAR(CDR(args)))); /* should be a number */
  expr = Kons(scalePower(CAR(sc_q),exp),
	      typePower(ZONE(unit),CDR(sc_q),exp,state));
  KEEPONLY(expr);
  return expr;
}

PRIVATE ExprPtr tcUnitStar(expr,assum,state)
	/* A_UNIT_STAR, return couple (scale,type) */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr unit1,unit2,sc_q1,sc_q2,args;
  ENTER;

  args = ARGS(expr);
  unit1 = CAR(args);
  unit2 = CAR(CDR(args));
  sc_q1 = TC(unit1,assum,state); /* return (scale,dimension) */
  sc_q2 = TC(unit2,assum,state); /* return (scale,dimension) */
  expr = Kons(scaleMul(CAR(sc_q1),CAR(sc_q2)),
	      typeMul(ZONE(unit1),ZONE(unit2),CDR(sc_q1),CDR(sc_q2),state));
  KEEPONLY(expr);
  return expr;
}

PRIVATE ExprPtr tcSetUnit(expr,assum,state)
 /* A_SET_UNIT, return type and scale number in place */
INOUT ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,num,unit,sc_q,scale;
  ENTER;

  /* expr = [z,A_SET_UNIT,[z1,A_NUMBER,num],unit] */
  args = ARGS(expr);
  num = CAR(ARGS(CAR(args)));
  unit = CAR(CDR(args));
  sc_q = TC(unit,assum,state); /* return (scale,dimension) */
  scale = scaleMul(KONS(num,NewMap()),CAR(sc_q));
  scale = instScale(scale,state);
  /* patch expr */
  TYPE(expr) = A_NUMBER;
  SETCAR(args,CAR(scale)); /* Here, we assume that a final scale does not
			      depend on variables. (This may only happen
			      with SML modules, extended to dimensions.) */
  if (CDR(scale)==NewMap())
    {
      SETCDR(args,NIL); /* now, expr = [z,A_NUMBER,num'] */
    }
  else
    {
      SETCAR(CDR(args),CDR(scale)); /* now, expr = [z,A_NUMBER,num',varscale] */
    }
  expr = CDR(sc_q);
  KEEPONLY(expr);
  return expr;
}

PRIVATE ExprPtr tcLabel(expr,assum,state) /* A_LABEL, return label */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  return CAR(ARGS(expr));
}

PRIVATE ExprPtr tcRecordMaplet(expr,assum,state) /* A_RECORD_MAPLET,
	return (label,type) */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,label,type,lbl,e;
  ENTER;

  args = ARGS(expr);
  lbl = CAR(args);
  args = CDR(args);
  e = CAR(args);
  label = TC(lbl,assum,state);
  type = TC(e,assum,state);
  expr = Kons(label,type);
  KEEPONLY(expr);
  return expr;
}

PRIVATE ExprPtr tcRecord(expr,assum,state) /* A_RECORD, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,l_t,e;
  MapPtr type;
  ENTER;

  type = NewMap();
  args = ARGS(expr);
  while (args!=NIL)
    {
      e = CAR(args);
      l_t = TC(e,assum,state);
      type = MapPut(CAR(l_t),CDR(l_t),type);
      args = CDR(args);
    }
  /*#if DEBUGGING!=0*/
  /*we need it for debugging and to determine if sharing is necessary
    in building tuples */
  addAnnotation(expr,type,A_FNLEVEL(assum),state);
  /*#endif*/
  KEEPONLY(type);
  return type;
}

#define tcFixity tcIllegal

PRIVATE ExprPtr tcTuple1(args,assum,state)
IN CONST ExprPtr args;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr e;
  ExprPtr *exprs,*ep;
  COUNTER i,n;

  if (args==NIL)
    return NIL;
  n = ListLength(args);
  exprs = (ExprPtr *)alloctmp(n*sizeof(ExprPtr));
  for (i=0, ep=exprs; i<n; i++, args=CDR(args))
    {
      e = CAR(args);
      *ep++ = TC(e,assum,state);
    }
  e = NIL;
  for (; ep!=exprs; )
    e = KONS(*--ep,e);
  return e;
}

PRIVATE ExprPtr tcTuple(expr,assum,state) /* A_TUPLE, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr type;
  ENTER;

  type = tcTuple1(ARGS(expr),assum,state);
/*#if DEBUGGING!=0*/
/*we need it for debugging and to determine if sharing is necessary
  in building tuples */
  addAnnotation(expr,type,A_FNLEVEL(assum),state);
/*#endif*/
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcProgn(expr,assum,state) /* A_PROGN, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,type,e;
  ENTER;

  type = NIL; /* useless since args should be non-empty */
  for (args=ARGS(expr); args!=NIL; args=CDR(args))
    {
      e = CAR(args);
      type = TC(e,assum,state);
    }
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcList(expr,assum,state) /* A_LIST, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,type,e;
  ENTER;

  args = ARGS(expr);
  if (args==NIL)
    type = listType(freshInstfromFlags(0,falseValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
				       ,INF_STRENGTH,INF_STRENGTH
#endif
				       ));
  else
    {
      e = CAR(args);
      type = TC(e,assum,state);
      while (args = CDR(args), args!=NIL)
	{
	  e = CAR(args);
	  (void) coerceTypes(ZONE(e),TC(e,assum,state),type,assum,state,false);
	}
      type = listType(type);
    }
#if DEBUGGING!=0
  addAnnotation(expr,type,A_FNLEVEL(assum),state);
#endif
  KEEPONLY(type);
  return type;
}

PRIVATE void checkRep(oldve,ve,vz)
IN CONST MapPtr oldve,ve,vz;
{
  ExprPtr errvars,x;
  ExprPtr zonex;

  errvars = SetInter(oldve,ve);
  if (errvars!=NewMap())
    {
      x = ChooseInDom(errvars);
      zonex = MapGetSure(x,vz);
      type_error(ZONE(zonex),(CHAR *)varMsg(message(55),x));
    }
}

PRIVATE ExprPtr tcDeclList1(args,assum,state)
	/* return assumptions */
IN CONST ExprPtr args;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr res,decl_assum,e;
  ENTER;

  if (args==NIL)
    {
      res = NewTypeAssum();
      A_KIND(res) = A_KIND(assum);
      A_FLAGS(res) = A_FLAGS(assum);
    }
  else
    {
      e = CAR(args);
      res = decl_assum = TC(e,assum,state);
      A_KIND(res) = A_KIND(assum);
      A_FLAGS(res) = A_FLAGS(assum);
      while (args=CDR(args), args!=NIL)
	{
	  AddTypeAssum(assum,decl_assum);
	  e = CAR(args);
	  decl_assum = TC(e,assum,state);
	  AddTypeAssum(res,decl_assum);
	}
    }
  KEEPONLY(res);
  return res;
}

PRIVATE ExprPtr beginAssum(assum)
IN CONST ExprPtr assum;
{
  ExprPtr new_assum;

  new_assum = DupTypeAssum(assum);
  A_KIND(new_assum) = TA_REGULAR;
  SETQ(A_TYEXTS(new_assum),NewMap()); /* so that extern datatypes are converted
					 to be used as regular datatypes, locally. */
  return new_assum;
}

PRIVATE void endAssum(assum,new_assum)
INOUT ExprPtr assum;
IN CONST ExprPtr new_assum;
{
  ENTER;

  /* get back all those newly created non generic type variables: */
  SETQ(A_NONGENS(assum),A_NONGENS(new_assum));
  /* then also get back the info that some new singular variables were
     used, so that they must get the imported status: */
  {
    MapPtr imported;

    imported = MapDomRestrTo(A_IDS(new_assum),A_IDEXTS(new_assum));
    setMapMerge(&A_IDEXTS(assum),imported);
    setMapMerge(&A_IDS(assum),imported);
  }
  /* the extern datatypes get dealt with by beginAssum(), so we don't do
     anything special here. */
  QUIT;
}

PRIVATE void updateAssum(assum,new_assum)
INOUT ExprPtr assum;
IN CONST ExprPtr new_assum;
{
  ENTER;

  endAssum(assum,new_assum);
  QUIT;
}

PRIVATE ExprPtr tcOfMap1(expr,assum,state)
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,f,p,new_assum,ve,ve_vz,type,e;
  ENTER;

  args = ARGS(expr);
  f = CAR(args);
  args = CDR(args);
  p = CAR(args);
  args = CDR(args);
  ve = NewMap();
  for (; args!=NIL; args=CDR(args))
    {
      e = CAR(args);
      if (TYPE(e)!=A_IN_SET && TYPE(e)!=A_IN_LIST && TYPE(e)!=A_SUB_MAP)
	break;
      ve_vz = TC(e,assum,state); /* return (varenv,varzones) */
      checkRep(ve,CAR(ve_vz),CDR(ve_vz));
      ve = MapMerge(ve,CAR(ve_vz));
    }
  new_assum = beginAssum(assum);
  setMapMerge(&A_IDS(new_assum),ve);
  setMapDomRestrBy(&A_IDEXTS(new_assum),ve); /* new variables shadow externals */
  AddTypeAssum(new_assum, tcDeclList1(args,new_assum,state));
  /* don't modify A_NONGENS(new_assum) : all new type variables
      are generic */
  type = TC(p,new_assum,state);
  (void) coerceTypes(ZONE(p),type,boolType,new_assum,state,false);
  type = TC(f,new_assum,state);
  endAssum(assum,new_assum);
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcListOfMap(expr,assum,state) /* A_LIST_OF, A_LIST_MAP return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr type;
  ENTER;

  type = listType(tcOfMap1(expr,assum,state));
  addAnnotation(expr,type,A_FNLEVEL(assum),state); /* kept for compilation */
  KEEPONLY(type);
  return type;
}

#define tcListOf tcListOfMap
#define tcListMap tcListOfMap

PRIVATE ExprPtr tcInSet(expr,assum,state)
	/* A_IN_SET, return (ve,vz),
	where ve is a map from variables to their types
	and vz is a map from variables to the zone of their definition */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,pat,dom,pattype,domtype,wanted,alpha,beta;
  ENTER;

  alpha = freshInstfromFlags(F_EQ,trueValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			     ,INF_STRENGTH,INF_STRENGTH
#endif
			     );
  beta = freshInstfromFlags(0,falseValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			    ,INF_STRENGTH,INF_STRENGTH
#endif
			    );
  wanted = mapType(alpha,beta);
  args = ARGS(expr);
  pat = CAR(args);
  dom = CAR(CDR(args));
  pattype = TC(pat,assum,state); /* = type1::type2::ve::vz */
  domtype = TC(dom,assum,state); /* = type */
  (void) coerceTypes(ZONE(dom),domtype,wanted,assum,state,false);
  (void) coerceTypes(ZONE(pat),CAR(pattype),alpha,assum,state,false);
  pattype = CDR(pattype);
  (void) coerceTypes(ZONE(pat),CAR(pattype),beta,assum,state,false);
  pattype = CDR(pattype);
  KEEPONLY(pattype);
  return pattype;
}

PRIVATE ExprPtr tcInList(expr,assum,state)
	/* A_IN_LIST, return (ve,vz),
	where ve is a map from variables to their types
	and vz is a map from variables to the zone of their definition */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,pat,dom,pattype,domtype,wanted;
  ENTER;

  wanted = listType(freshInstfromFlags(0,falseValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
				       ,INF_STRENGTH,INF_STRENGTH
#endif
				       ));
  args = ARGS(expr);
  pat = CAR(args);
  dom = CAR(CDR(args));
  pattype = TC(pat,assum,state); /* = type::ve::vz */
  domtype = TC(dom,assum,state); /* = type */
  (void) coerceTypes(ZONE(dom),domtype,wanted,assum,state,false);
  (void) coerceTypes(ZONE(pat),CAR(pattype),CAR(APPLARG(wanted)),assum,state,false);
  pattype = CDR(pattype);
  KEEPONLY(pattype);
  return pattype;
}

PRIVATE ExprPtr tcMaplet(expr,assum,state) /* A_MAPLET, return (type1,type2),
	where type1 and type2 are the types of the components
	of the maplet */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,domtype,rngtype,e;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  domtype = TC(e,assum,state);
  coerceFlagTypes(ZONE(CAR(args)),0/*F_EQ*/,trueValue,
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
		  INF_STRENGTH,INF_STRENGTH,
#endif
		  domtype,assum,state);
  e = CAR(CDR(args));
  rngtype = TC(e,assum,state);
  e = Kons(domtype,rngtype);
  KEEPONLY(e);
  return e;
}

PRIVATE ExprPtr tcMap(expr,assum,state) /* A_MAP, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,type,type1,type2,t12,e;
  TZone *zone;
  ENTER;

  args = ARGS(expr);
  if (args==NIL)
    type = mapType(freshInstfromFlags(F_EQ,trueValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
				      ,INF_STRENGTH,INF_STRENGTH
#endif
				      ),
		   freshInstfromFlags(0,falseValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
				      ,INF_STRENGTH,INF_STRENGTH
#endif
				      ));
  else
    {
      e = CAR(args);
      t12 = TC(e,assum,state);
      type1 = CAR(t12);
      type2 = CDR(t12);
      while (args = CDR(args), args!=NIL)
	{
	  e = CAR(args);
	  zone = ZONE(e);
	  t12 = TC(e,assum,state);
	  (void) coerceTypes(zone,CAR(t12),type1,assum,state,false);
	  (void) coerceTypes(zone,CDR(t12),type2,assum,state,false);
	}
      type = mapType(type1,type2);
    }
#if DEBUGGING!=0
  addAnnotation(expr,type,A_FNLEVEL(assum),state);
#endif
  KEEPONLY(type);
  return type;
}

#define tcMapUnder tcMap

PRIVATE ExprPtr tcMapOfMap(expr,assum,state) /* A_MAP_OF, A_MAP_MAP,
	return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr t12;
  ENTER;

  t12 = tcOfMap1(expr,assum,state);
  t12 = mapType(CAR(t12),CDR(t12));
#if DEBUGGING!=0
  addAnnotation(expr,t12,A_FNLEVEL(assum),state);
#endif
  KEEPONLY(t12);
  return t12;
}

#define tcMapOf tcMapOfMap
#define tcMapMap tcMapOfMap
#define tcMapUnderOf tcMapOfMap
#define tcMapUnderMap tcMapOfMap

PRIVATE ExprPtr tcProgOf(expr,assum,state)
	/* A_PROG_OF, A_PROG_MAP, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  (void) tcOfMap1(expr,assum,state);
#if DEBUGGING!=0
  addAnnotation(expr,NIL,A_FNLEVEL(assum),state);
#endif
  return NIL; /* type of the 0-ary cartesian product */
}

#define tcProgMap tcProgOf /* A_PROG_MAP */

PRIVATE ExprPtr tcQuantOfMap(expr,assum,state)
	/* A_ALL_OF, A_ALL_MAP, A_EXISTS_OF, A_EXISTS_MAP,
	return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr type;
  ENTER;

  type = tcOfMap1(expr,assum,state);
  (void) coerceTypes(ZONE(CAR(ARGS(expr))),type,boolType,assum,state,false);
#if DEBUGGING!=0
  addAnnotation(expr,boolType,A_FNLEVEL(assum),state);
#endif
  QUIT;
  return boolType;
}

#define tcAllOf tcQuantOfMap
#define tcAllMap tcQuantOfMap
#define tcExistsOf tcQuantOfMap
#define tcExistsMap tcQuantOfMap

PRIVATE ExprPtr tcSomeOf(expr,assum,state)
     /* A_SOME_OF, A_SOME_MAP */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr type;
  ENTER;

  type = KtypeAppl(optSym,KONS(tcOfMap1(expr,assum,state),NIL));
  (void) KEEP(type);
  addAnnotation(expr,type,A_FNLEVEL(assum),state); /*kept for compilation*/
  KEEPONLY(type);
  return type;
}

#define tcSomeMap tcSomeOf /* A_SOME_MAP */

PRIVATE ExprPtr tcLet(expr,assum,state) /* A_LET, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,local_assum,new_assum,type,e;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  new_assum = beginAssum(assum); /* was before just after TC(e,assum,state) (not new_assum) */
  local_assum = TC(e,new_assum,state); /* assumption */
  AddTypeAssum(new_assum,local_assum); /* is it needed? */
  e = CAR(CDR(args));
  type = TC(e,new_assum,state);
  endAssum(assum,new_assum);
#ifdef BREAK_ON_CONTROL
#if DEBUGGING!=0
  addAnnotation(expr,type,A_FNLEVEL(assum),state);
#endif
#endif
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcApply1(fzone,argzone,ftype,argtype,assum,state)
IN CONST TZone *fzone,*argzone;
IN CONST ExprPtr ftype,argtype;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr type;
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  ExprPtr funtype;
#endif
  ftype = correctTypeAppl(ftype,state);
  /*if (ftype==NIL)
    goto error;*/
  if (ftype->kind==(CHAR)E_VAR)
    ftype = _nongenValue(ftype,state);
  if (ftype->kind==(CHAR)E_VAR)
    {
      type = freshInstfromFlags(0,falseValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
				,INF_STRENGTH,INF_STRENGTH
#endif
				);
      (void) coerceTypes(fzone,ftype,
#if IMPERATIVE_TYPES==EFFECT_TYVAR
			 funtype = funTypeFx(argtype,type,
					     freshFxVar(NewMap(),effectSym,assum,state)),
#else
			 funType(argtype,type),
#endif
			 assum,state,false);
    }
  else if (ftype->kind==(CHAR)E_TYPE_APPL && fxSym(APPLFUN(ftype))==funSym)
    {
#if IMPERATIVE_TYPES==EFFECT_TYVAR
      funtype = ftype;
#endif
      (void) coerceTypes(argzone,argtype,CAR(APPLARG(ftype)),assum,state,false);
      type = CAR(CDR(APPLARG(ftype)));
    }
  else
    {
      /*error:*/
      typeError(fzone,(CHAR *)message(56),ftype,state);
      return freshInstfromFlags(0,falseValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
				,INF_STRENGTH,INF_STRENGTH
#endif
				);
    }
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  funtype = APPLFUN(funtype);
  if (funtype->kind==(CHAR)E_TYPEFX)
    addFx(CDR(funtype));
#endif
  return type;
}

PRIVATE Boolean complexFn(f,assum)
IN CONST ExprPtr f;
INOUT ExprPtr assum;
{
again:
  switch (TYPE(f))
    {
    case A_VARIABLE:
      {
	ExprPtr sym;
	Boolean found;

	sym = CAR(ARGS(f));
	(void) MapGet(sym,A_CONS(assum),&found);
	if (found) /* type constructor */
	  return false;
      }
      break;
    case A_CAST:
      f = CAR(ARGS(f));
      goto again;
    }
  return true;
}

PRIVATE ExprPtr tcApply(expr,assum,state) /* A_APPLY, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,f,arg,ftype,argtype,exp,alpha,type,ptype;
  ENTER;

  args = ARGS(expr);
  f = CAR(args);
  arg = CAR(CDR(args));
  ftype = TC(f,assum,state);
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  ftype = Weaken(ftype,assum,state,1L);
#endif
  argtype = TC(arg,assum,state);
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  if (complexFn(f,assum))
    argtype = Weaken(argtype,assum,state,INF_STRENGTH);
  /* all weak type variables in the argument may be created by the
     application; see (fn f => f nil nil) (fn x => fn y => ref x) */
  /* except when f is not complex (typically a datatype constructor,
     which cannot apply its argument to anything). */
#endif
  if (TYPE(f)==A_VARIABLE && CAR(ARGS(f))==powerSym &&
       TYPE(arg)==A_TUPLE && ListLength(ARGS(arg))==2 &&
       (exp = CAR(CDR(ARGS(arg))), TYPE(exp)==A_NUMBER))
    { /* if we analyse op **(a,b), where b is a number: */
      exp = CAR(ARGS(exp));
      ftype = newTypeInst(ftype,NIL,state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			  ,false
#endif
			  );
      /* (resolve bindings of type variables) */
      if (ftype==powerType)
	{ /* and if op **: num * num -> num, we assume that
		op ** is the original power operation; so, knowing the
		value of b, we extend the typing so that
		op **: '#a * num -> '#a^b */
	  alpha = freshInstfromFlags(F_NUM,trueValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
					 ,INF_STRENGTH,INF_STRENGTH
#endif
					 );
	  ptype = KnumType(MapSingleton(alpha,exp));
	  (void) KEEP(ptype);
	  alpha = KnumType(MapSingleton(alpha,one));
	  (void) KEEP(alpha);
	  ftype = funTypeFx(KONS(alpha,KONS(numType,NIL)),ptype,freshFxVar(NewMap(),effectSym,assum,state));
	}
    }
  type = tcApply1(ZONE(f),ZONE(arg),ftype,argtype,assum,state);
/*#if DEBUGGING!=0*/
/*we need it for debugging and to determine if sharing is necessary
  in building tuples */
  addAnnotation(expr,type,A_FNLEVEL(assum),state);
/*#endif*/
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcCast(expr,assum,state) /* A_CAST, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,type,type2,e;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  type = TC(e,assum,state);
  e = CAR(CDR(args));
  type2 = typeSubst(ZONE(e),TC(e,assum,state),A_TYVARS(assum),assum,state);
  (void) coerceTypes(ZONE(expr),type,type2,assum,state,false);
  KEEPONLY(type); /* must not return type2, which does not have any latent effects */
  return type;
}

PRIVATE ExprPtr tcBoolBin(expr,assum,state) /* A_AND, A_OR, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,e;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  (void) coerceTypes(ZONE(e),TC(e,assum,state),boolType,assum,state,false);
  args = CDR(args);
  e = CAR(args);
  (void) coerceTypes(ZONE(e),TC(e,assum,state),boolType,assum,state,false);
#ifdef BREAK_ON_CONTROL
#if DEBUGGING!=0
  addAnnotation(expr,boolType,A_FNLEVEL(assum),state);
#endif
#endif
  QUIT;
  return boolType;
}

#define tcAnd tcBoolBin
#define tcOr tcBoolBin

/* in the effect case, <code> handle <fn> 
	<code> : tau :: fx
	<fn> : exn[rho] ->[fx'] tau :: empty (because of syntax, the effect is empty)
	----------------------------------------------------------------
	<code> handle <fn> : tau :: fx U fx' U exn(rho,tau) U from(rho)

	<expr> : exn[rho] :: fx
	--------------------------------------
	raise <expr> : alpha :: fx U goto(rho)

	and (Jouvelot defined the effect of callcc to be fx' >= fx U from(rho,alpha)
		instead of the following):
	capture : (alpha cont[rho] ->[fx] alpha) ->[fx'] alpha,
		fx' >= fx U from(rho)
	callcc : (alpha cont[rho] ->[fx] alpha) ->[fx'] alpha,
		fx' >= fx U exn(rho,alpha) U from(rho)
	throw : alpha cont[rho] ->[fx] alpha ->[fx'] alpha',
		fx' >= fx U goto(rho)
	goto is the effect of (possibly non-local) control-flow alteration
	from is the effect of storing a continuation (with exception state);
		hence capture has no from effect.

		!!! actually, handle should not have an exn(rho,tau) effect,
		which should be left to the creation of the exception (tcException
		must include a side-effect, since exceptions are generated at
		run-time).
*/
PRIVATE ExprPtr tcHandle(expr,assum,state) /* A_HANDLE, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,type,exntype,e;
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  ExprPtr funtype,rho;
#endif
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  type = TC(e,assum,state);
  e = CAR(CDR(args));
  exntype = TC(e,assum,state);
  exntype = correctTypeAppl(exntype,state);
  /*if (exntype==NIL)
    goto error;*/
  if (exntype->kind==(CHAR)E_VAR)
    exntype = _nongenValue(exntype,state);
  if (exntype->kind==(CHAR)E_VAR)
    {
      (void) coerceTypes(ZONE(e),exntype,
#if IMPERATIVE_TYPES==EFFECT_TYVAR
			 funtype = funTypeFx(exnTypeFx(rho = freshFxVar(NewMap(),
									rgnSym,assum,state)),
					     type,
					     freshFxVar(NewMap(),effectSym,assum,state)),
#else
			 funType(exnType,type),
#endif
			 assum,state,false);
    }
  else if (fxSym(APPLFUN(exntype))==funSym)
    {
#if IMPERATIVE_TYPES==EFFECT_TYVAR
      funtype = exntype;
#endif
      (void) coerceTypes(ZONE(e),CAR(APPLARG(exntype)),
#if IMPERATIVE_TYPES==EFFECT_TYVAR
			 exnTypeFx(rho = freshFxVar(NewMap(),rgnSym,assum,state)),
#else
			 exnType,
#endif
			 assum,state,false);
      (void) coerceTypes(ZONE(expr),CAR(CDR(APPLARG(exntype))),type,assum,state,false);
    }
  else
    {
      /*error:*/
      typeError(ZONE(CAR(CDR(expr))),(CHAR *)message(57),exntype,state);
      type = freshInstfromFlags(0,falseValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
				,INF_STRENGTH,INF_STRENGTH
#endif
				);
      goto end;
    }
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  funtype = APPLFUN(funtype);
  if (funtype->kind==(CHAR)E_TYPEFX)
    addFx(CDR(funtype));
  addFx(exnFx(rho,type));
  addFx(fromFx(rho));
#endif
end:
#if DEBUGGING!=0
  addAnnotation(expr,type,A_FNLEVEL(assum),state);
#endif
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcRaise(expr,assum,state) /* A_RAISE, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,type,e;
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  ExprPtr rho;
#endif
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  type = TC(e,assum,state);
  (void) coerceTypes(ZONE(e),type,exnType,assum,state,false);
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  if (type->kind==(CHAR)E_VAR)
    type = _nongenValue(type,state); /* type is now an application */
  if (type->kind==(CHAR)E_TYPE_APPL)
    {
      rho = APPLFUN(type);
      if (rho->kind==(CHAR)E_TYPEFX)
	{
	  rho = CDR(rho);
	  addFx(gotoFx(rho));
	}
    }
#endif
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  type = freshInstfromFlags(0,falseValue,assum,state,INF_STRENGTH,INF_STRENGTH);
#else
  type = freshInstfromFlags(0,falseValue,assum,state);
#endif
#if DEBUGGING!=0
  addAnnotation(expr,type,A_FNLEVEL(assum),state);
#endif
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcMatch(expr,assum,state) /* A_MATCH, A_FN, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,type,e;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  ExprPtr oldmaxstrengths;
#endif
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  ExprPtr oldFx;
#endif
  ENTER;

#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  oldmaxstrengths = S_MAXSTRENGTH(state);
  (void) KEEP(oldmaxstrengths);
  setMapMerge(&S_MAXSTRENGTH(state),S_STRENGTH(state));
  /* to prevent strengthening to go over the limit
     imposed by the strengths of outer type variables;
     this is actually redundant, since maxstrengths
     could be inferred (slowly) from S_TYVARSUBS and
     the strengths of outer type variables (which are
     those which were created before entering the lambda) */
#endif
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  oldFx = KEEP(theFx);
  resetFx();
#endif
  args = ARGS(expr);
  if (args==NIL)
    type =
#if IMPERATIVE_TYPES==EFFECT_TYVAR
      funTypeFx
#else
	funType
#endif
	  (freshInstfromFlags(0,falseValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			      ,INF_STRENGTH,INF_STRENGTH
#endif
			      ),
	   freshInstfromFlags(0,falseValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			      ,INF_STRENGTH,INF_STRENGTH
#endif
			      )
#if IMPERATIVE_TYPES==EFFECT_TYVAR
	   ,freshFxVar(NewMap(),effectSym,assum,state)
#endif
	   );
  else
    {
      e = CAR(args);
      type = TC(e,assum,state);
      while (args = CDR(args), args!=NIL)
	{
	  e = CAR(args);
	  (void) coerceTypes(ZONE(e),TC(e,assum,state),type,assum,state,false);
	}
    }
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  (void) Weaken(type,assum,state,-1L);
  setMapDomRestrTo(&S_MAXSTRENGTH(state),oldmaxstrengths);
#endif
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  SETQ(theFx,oldFx);
#endif
  KEEPONLY(type);
  return type;
}


PRIVATE ExprPtr tcFn(expr,assum,state) /* A_FN, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr type;

  type = tcMatch (expr, assum, state);
  addAnnotation(expr,type,A_FNLEVEL(assum),state); /* so as to detect
						      whether function is unary
						      or n-ary later on. */
  return type;
}

PRIVATE ExprPtr tcLambda(expr,assum,state) /* A_LAMBDA, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,pattype,valtype,new_assum,e,ve;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  A_FNLEVEL(assum)++;
  pattype = TC(e,assum,state); /* = type::ve::vz */
  new_assum = beginAssum(assum);
  ve = CAR(CDR(pattype));
  setMapMerge(&A_IDS(new_assum),ve);
  setMapDomRestrBy(&A_IDEXTS(new_assum),ve); /* new variables shadow externals */
  /* all new type variables introduced by the pattern
     are non generic: A_NONGENS(assum) is modified directly
     by the call to TC() above */
  e = CAR(CDR(args));
  valtype = TC(e,new_assum,state);
#if DEBUGGING!=0
  addAnnotation(e,valtype,A_FNLEVEL(new_assum),state);
#endif
  endAssum(assum,new_assum);
  valtype = funTypeFx(CAR(pattype),valtype,
		      freshFxVar(theFx,effectSym,assum,state));
  A_FNLEVEL(assum)--;
  KEEPONLY(valtype);
  return valtype;
}

PRIVATE ExprPtr tcIf(expr,assum,state) /* A_IF, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,type,e;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  (void) coerceTypes(ZONE(e),TC(e,assum,state),boolType,assum,state,false);
  args = CDR(args);
  e = CAR(args);
  type = TC(e,assum,state);
  args = CDR(args);
  e = CAR(args);
  (void) coerceTypes(ZONE(e),TC(e,assum,state),type,assum,state,false);
#ifdef BREAK_ON_CONTROL
#if DEBUGGING!=0
  addAnnotation(expr,type,A_FNLEVEL(assum),state);
#endif
#endif
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcWhile1(expr,assum,state) /* A_WHILE, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,e,type;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  (void) coerceTypes(ZONE(e),TC(e,assum,state),boolType,assum,state,false);
  args = CDR(args);
  e = CAR(args);
  type = TC(e,assum,state);
  args = CDR(args);
  e = CAR(args);
  (void) coerceTypes(ZONE(e),TC(e,assum,state),boolType,assum,state,false);
#ifdef BREAK_ON_CONTROL
#if DEBUGGING!=0
  addAnnotation(expr,type,A_FNLEVEL(assum),state);
#endif
#endif
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcWhile(expr,assum,state) /* A_WHILE, return NIL */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  (void) tcWhile1(expr,assum,state);
  return NIL; /* type of the 0-ary cartesian product */
}

PRIVATE ExprPtr tcWhileList(expr,assum,state) /* A_WHILE_LIST, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr type;
  ENTER;

  type = listType(tcWhile1(expr,assum,state));
  addAnnotation(expr,type,A_FNLEVEL(assum),state); /* kept for compilation */
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcWhileMap(expr,assum,state)
	/* A_WHILE_MAP, A_WHILE_MAP_UNDER, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr t12;
  ENTER;

  t12 = tcWhile1(expr,assum,state);
  t12 = mapType(CAR(t12),CDR(t12));
  KEEPONLY(t12);
  return t12;
}

#define tcWhileMapUnder tcWhileMap

PRIVATE ExprPtr tcWhileBool(expr,assum,state)
	/* A_WHILE_ALL, A_WHILE_EXISTS, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr type;
  ENTER;

  type = tcWhile1(expr,assum,state);
  (void) coerceTypes(ZONE(CAR(CDR(ARGS(expr)))),type,boolType,assum,state,false);
  QUIT;
  return boolType;
}

#define tcWhileAll tcWhileBool
#define tcWhileExists tcWhileBool

PRIVATE ExprPtr tcWhileSome(expr,assum,state)
	/* A_WHILE_SOME, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr type;
  ENTER;

  type = tcWhile1(expr,assum,state);
  type = KtypeAppl(optSym,KONS(type,NIL));
  addAnnotation(expr,type,A_FNLEVEL(assum),state); /*kept for compilation*/
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcDeclList(expr,assum,state)
	/* A_DECL_LIST, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  return tcDeclList1(ARGS(expr),assum,state);
}

PRIVATE ExprPtr tcLocal(expr,assum,state) /* A_LOCAL, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,new_assum,assum1,assum2,e;
  ENTER;

  args = ARGS(expr);
  new_assum = beginAssum(assum);
  e = CAR(args);
  assum1 = TC(e,new_assum,state);
  AddTypeAssum(new_assum,assum1);
  A_KIND(new_assum) = A_KIND(assum); /* if we were precompiling, then precompile
					the body of the local expression (not its
					local, hidden declarations). */
  setMapMerge(&A_TYEXTS(new_assum),MapDomRestrBy(A_TYEXTS(assum),A_TYFCNS(assum1)));
  /* if we were precompiling, then precompile the body knowing that the types
     that were declared extern beforehand, except those overriden by local
     declarations (in A_TYFCNS(assum1)), are considered extern. */
  e = CAR(CDR(args));
  assum2 = TC(e,new_assum,state);
  endAssum(assum,new_assum);
  {
    MapPtr imported,aux;

    /*
      Any imported identifier in new_assum should be added as imported
      to assum2.
    */
    imported = MapDomRestrTo (A_IDS(new_assum), A_IDEXTS(new_assum));
    aux = MapOverwrite (imported, A_IDS(assum2));
    SETQ(A_IDS(assum2), aux);
    aux = MapOverwrite (imported, A_IDEXTS(assum2));
    SETQ(A_IDEXTS(assum2), aux);
    imported = MapDomRestrTo (A_IDLOCS(new_assum), imported);
    aux = MapOverwrite (imported, A_IDLOCS(assum2));
    SETQ(A_IDLOCS(assum2), aux);
    /*
      Any imported type... similarly: no, this is useless.
    imported = A_TYEXTS(new_assum);
    setMapMerge (&A_TYEXTS(assum2), imported);
    setMapMerge (&A_TYFCNS(assum2), imported);
     */
  }
  KEEPONLY(assum2);
  return assum2;
}

MAPSWEEPFUNC(do_checkidexts)
{
  struct checkidexts *data;
  ExprPtr type;
  TZone *zone;
  Boolean found;

  data = MAPDATA(struct checkidexts *);
  (void) MapGet(x,A_IDS(data->assum),&found);
  if (found) /* then it was imported (declared extern, then used) or it was already defined */
    {
      (void) MapGet (x,A_IDEXTS(data->assum),&found);
      if (!found) /* if imported (found==true),
		     then this is not a redefinition, and
		    is in fact legal; so don't produce any error.
		    In fact, it won't be in data->vz, so trying to
		    produce an error would fail, additionally. */
	{
	  zone = ZONE(MapGetSure(x,data->vz));
	  (void) MapGet(x,A_IDEXTS(data->assum),&found);
	  type_soft_error(zone,(CHAR *)varMsg(message(found?785:781),x));
	}
    }
  else
    {
      type = MapGet(x,A_IDEXTS(data->assum),&found);
      if (found) /* then check its type */
	{
	  ExprPtr assum,state;
	  
	  assum = DupTypeAssum(data->assum);
	  state = DupTypeState(data->state);
	  y = newTypeInst(y,assum,state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			  ,false
#endif
			  );
	  zone = ZONE(MapGetSure(x,data->vz));
	  (void) coerceTypes(zone,y,type,assum,state,false);
	}
    }
  return false;
}

PRIVATE void checkIdExts(ve,vz,assum,state)
IN CONST MapPtr ve,vz;
IN CONST ExprPtr assum;
IN CONST ExprPtr state;
{
  if (A_KIND(assum)==TA_PRECOMPILE)
    {
      struct checkidexts *data =
	(struct checkidexts *)alloctmp(sizeof(struct checkidexts));

      data->assum  = assum;
      data->state = state;
      data->vz = vz;
      (void) MapSweep(ve,do_checkidexts,data);
      /* the change from singular to regular status is done through the use
	 of AddTypeAssum(assum,new_assum) which is done via tcDeclList1() */
    }
  else setMapDomRestrBy(&A_IDEXTS(assum),ve); /* shadow declarations of extern variables */
}

PRIVATE ExprPtr tcVal1(expr,assum,state,recp) /* A_VAL, A_VALREC, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
IN Boolean recp;
{
  ExprPtr args,new_assum,local_assum,e,ve_vz,ve,vz,tyvars;
#if IMPERATIVE_TYPES==STD_ML_TYVAR
  ExprPtr exps;
#endif
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  ExprPtr savestrengths;
#endif
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  ExprPtr oldFx;
#endif
  ExprPtr ll,bind,pat,pattype,pattypes,lastpattypes,aux;
  ExprPtr memoref,impmemoref;
  ExprPtr subs;
  ENTER;

  flushTyVarSubCache(state);
  subs = S_TYVARSUBS(state);
  (void) KEEP(subs);
  args = ARGS(expr);
  local_assum = DupTypeAssum(assum);
  tyvars = valTyVars(ARGS(CAR(args)),local_assum,state);
#ifdef DEBUG_EXPLICIT
  hfprintf(hstderr,"Entering tcVal1; explicit type variables scoped here: ");
  EP(tyvars);
#endif
  setMapMerge(&A_TYVARS(local_assum),tyvars);
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  savestrengths = getStrengths(tyvars,state);
  setStrengths(tyvars,state,INF_STRENGTH>>1);
#endif
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  oldFx = KEEP(theFx);
  resetFx();
#endif
  /* all explicit type variables scoped at the val declaration
     are assumed generic */
  args = CDR(args); /* args is a non-empty list */
  ll = args;

  bind = CAR(ll);
  pat = CAR(ARGS(bind));
  pattype = TC(pat,local_assum,state); /* = type::ve::vz */
  ve_vz = CDR(pattype);
  ve = CAR(ve_vz);
  vz = CDR(ve_vz);
  if (ve==NewMap())
    {
nobind:
      if (A_FLAGS(local_assum) & A_NOWARN_NOBIND)
	;
      else
	type_warn(ZONE(pat),(CHAR *)message(58));
    }
#if IMPERATIVE_TYPES==STD_ML_TYVAR
  exps = (expansivep(CAR(CDR(ARGS(bind)))))?ve:NewMap();
#endif
  aux = Cons(CAR(pattype),NIL);
  pattypes = lastpattypes = aux;
  while (ll=CDR(ll), ll!=NIL)
    {
      bind = CAR(ll);
      pat = CAR(ARGS(bind));
      pattype = TC(pat,local_assum,state); /* = type::ve::vz */
      ve_vz = CDR(pattype); /* ve::vz */
      if (CAR(ve_vz)==NewMap())
	goto nobind;
      checkRep(ve,CAR(ve_vz),CDR(ve_vz));
      ve = MapMerge(ve,CAR(ve_vz));
      vz = MapMerge(vz,CDR(ve_vz));
#if IMPERATIVE_TYPES==STD_ML_TYVAR
      if (expansivep(CAR(CDR(ARGS(bind)))))
	exps = SetUnion(exps,CAR(ve_vz));
#endif
      aux = Cons(CAR(pattype),NIL);
      SETCDR(lastpattypes,aux);
      lastpattypes = aux;
    }
  if (recp)
    {
      setMapMerge(&A_IDS(local_assum),ve);
#ifdef OBSOLETE
      aux = A_IDEXTS(local_assum);
      (void) KEEP(aux);
#endif
      setMapDomRestrBy(&A_IDEXTS(local_assum),ve);
      /* new variables shadow externals; if externs, swap them temporarily out of A_IDEXTS */
    }
  for (ll=args; ll!=NIL; ll=CDR(ll), pattypes=CDR(pattypes))
    {
      bind = CAR(ll);
      e = CAR(CDR(ARGS(bind)));
      (void) coerceTypes(ZONE(e),TC(e,local_assum,state),CAR(pattypes),
			 local_assum,state,false);
    }
#ifdef OBSOLETE
  if (recp)
    {
      SETQ(A_IDEXTS(local_assum),aux); /* reinstall all A_IDEXTS */
    }
#endif
  aux = A_NONGENS(assum);
  (void) KEEP(aux);
  SETQ(A_NONGENS(assum),A_NONGENS(local_assum)); /* inform checkIdExtsTypes of new non-generics */
  checkIdExts(ve,vz,assum,state);
  memoref = newMemoRef();
  /* all type variables that are created during elaboration of expressions
     must be unquantified */
  new_assum = NewTypeAssum();
  SETQ(A_IDS(new_assum),ve);
  {
    MapPtr imported;

    imported = MapDomRestrTo(A_IDS(local_assum), A_IDEXTS(local_assum));
    imported = MapDomRestrTo(MapDomRestrBy(imported, A_IDS(assum)), A_IDEXTS(assum));
    /* imported contains only the values imported, but which were singular before
       the val declaration: make this change of status visible to new_assum: */
    setMapMerge(&A_IDEXTS(new_assum),imported);
    setMapMerge(&A_IDS(new_assum),imported);
  }
  SETQ(A_IDLOCS(new_assum),vz);
#ifdef OLD
  aux = unquant_x(aux,local_assum,state,memoref);
#endif
  flushTyVarSubCache(state);
  {
    ExprPtr modifs;

    modifs = MapDomRestrTo(aux,
			   MapDomRestrBy(S_TYVARSUBS(state),
					 subs));
    aux = MapMerge(aux,unquant_x(modifs,local_assum,state,memoref));
  }
  /* exclude all type variables
     appearing in old nongens from being quantified */
  aux = SetUnion(aux,MapDomRestrTo(A_NONGENS(local_assum),S_TYVARSUBS(state)));
  /* we exclude from quantification:
     - the old type variables A_NONGENS(assum) (those existing before tcVal)
     - all type variables that were created afterwards but became bound
     moreover, we also exclude (unquant) imperative type variables associated
     with expansive expressions
     */
#if IMPERATIVE_TYPES==STD_ML_TYVAR
  impmemoref = newMemoRef();
  aux = SetUnion(aux,unquant_y(exps,local_assum,state,impmemoref));
#endif
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  impmemoref = newMemoRef();
  /* unquantify all free strength 0 variables */
  aux = SetUnion(aux,unquant_y(ve,local_assum,state,impmemoref));
#endif
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  /* unquantify all variables that are free in the current effect */
  trimObsFx(assum,ve,local_assum,state,memoref);
  /* normally, the second argument to trimObsFx is a type,
     not a map from ids to types; but such a map can be viewed
     as a record type, which has the good properties (the same
     free type variables) */
  aux = SetUnion(aux,unquant_x(theFx,local_assum,state,memoref));
#endif
  aux = SetUnion(aux,unquant_packers(local_assum,state,memoref));
  /* type variables still free in packed types must not be quantified */
  SETQ(A_NONGENS(new_assum),aux);
  /* check that all type variables in tyvars have been quantified,
     i.e. A_NONGENS(new_assum) & tyvars = {} */
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  checkStrengths(ZONE(expr),tyvars,state);
  setMapMerge(&S_STRENGTH(state),savestrengths);
#endif
  aux = checkExplicits(ZONE(expr),tyvars,new_assum,state);
  /*
  if (aux!=NIL)
    type_error(ZONE(expr),(CHAR *)varMsg(message(59),aux));
  */
  updateAssum(assum,local_assum);
  SETQ(A_NONGENS(assum),A_NONGENS(new_assum)); /* record new non-generic type variables */
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  aux = SetUnion(oldFx,theFx);
  SETQ(theFx,aux);
#endif
#ifdef DEBUG_EXPLICIT
  hfprintf(hstderr,"Leaving tcVal1.\n");
#endif
  KEEPONLY(new_assum);
  return new_assum;
}

PRIVATE ExprPtr tcVal(expr,assum,state) /* A_VAL, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  return tcVal1(expr,assum,state,false);
}

PRIVATE ExprPtr tcValRec(expr,assum,state) /* A_VALREC, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  return tcVal1(expr,assum,state,true);
}

#define tcBind tcIllegal
#define tcFunBind tcIllegal

PRIVATE ExprPtr tcPatTuple1(list,assum,state,vep,vezp)
IN CONST ExprPtr list;
INOUT ExprPtr assum;
INOUT ExprPtr state;
OUT MapPtr *vep,*vezp;
{
  ExprPtr pat,type,t,tt;
  ExprPtr *exprs,*ep;
  COUNTER i,n;

  type = NIL;
  *vep = NewMap();
  *vezp = NewMap();
  if (list!=NIL)
    {
      n = ListLength(list);
      exprs = (ExprPtr *)alloctmp(n*sizeof(ExprPtr));
      for (i=0, ep=exprs; i<n; i++, list=CDR(list))
	{
	  pat = CAR(list);
	  if (TYPE(pat)==A_PAT_TUPLE_ELLIPSIS)
	    {
	      type = freshInstfromFlags(0,falseValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
					,INF_STRENGTH,INF_STRENGTH
#endif
					);
	      break;
	    }
	  *ep++ = TC(pat,assum,state);
	}
      for (; ep!=exprs; )
	{
	  t = *--ep;
	  tt = CDR(t);
	  checkRep(CAR(tt),*vep,*vezp);
	  *vep = MapMerge(CAR(tt),*vep);
	  *vezp = MapMerge(CDR(tt),*vezp);
	  type = KONS(CAR(t),type);
	}
    }
  return type;
}

PRIVATE ExprPtr tcPatTuple(expr,assum,state)
	/* A_PAT_TUPLE, return type::ve::vz */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr tlist,ve,vz,type;
  ENTER;

  tlist = tcPatTuple1(ARGS(expr),assum,state,&ve,&vz);
  addAnnotation(expr,tlist,A_FNLEVEL(assum),state); /* kept for compilation */
  /* the list tlist is itself the cartesian product type */
  type = Kons(tlist,KONS(ve,vz));
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcDecl(expr,assum,state)
	/* A_QUANTITY, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,e,new_assum;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  new_assum = TC(e,assum,state);
  while (args=CDR(args), args!=NIL)
    {
      e = CAR(args);
      AddTypeAssum(new_assum,TC(e,assum,state));
    }
  KEEPONLY(new_assum);
  return new_assum;
}

#define tcQuantity tcDecl

PRIVATE ExprPtr tcNewQuan(expr,assum,state)
	/* A_NEW_QUAN, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,quan,unit,new_assum,type,quanname,aux,oldtyfcn;
  Boolean found;
  ENTER;

  new_assum = NewTypeAssum();
  args = ARGS(expr);
  quan = CAR(ARGS(CAR(args)));
  found = false;
  if (A_KIND(assum)==TA_PRECOMPILE)
    {
      oldtyfcn = MapGet(quan,A_TYEXTS(assum),&found);
      if (found)
	{
	  /* imported datatype or dimension */
	  if (CDR(oldtyfcn)->kind!=(CHAR)E_NUMTYPE)
	    {
	      type_soft_error(ZONE(CAR(args)),(CHAR *)varMsg(message(824),quan));
	      goto fresh;
	    }
	  quanname = applFun(ChooseInDom(REFVAL(CDR(oldtyfcn))),state);
	}
      else
	{
	  (void) MapGet(quan,A_TYFCNS(assum),&found);
	  if (found)
	    type_soft_error(ZONE(CAR(args)),(CHAR *)varMsg(message(780),quan));
	  goto fresh;
	}
    }
  else
    {
fresh:
      found = false;
      quanname = freshTyName(quan,assum,state);
    }
  args = CDR(args);
  unit = CAR(ARGS(CAR(args)));
  type = KtypeAppl(quanname,NIL);
  type = KnumType(MapSingleton(KEEP(type),one));
  (void) KEEP(type);
  aux = MapSingleton(quan,KONS(KONS(NIL,NewMap()),type));
  SETQ(A_TYFCNS(new_assum),aux);
  if (found)
    /* signal that external dimension has been defined; put it in the regular state. */
    setMapRemove(quan,&A_TYEXTS(assum));
  /* {quan => ((),{%quan() => 1.0})}, with tyvarseq=() */
  aux = MapSingleton(unit,KONS(KONS(one,NewMap()),type));
  SETQ(A_UNITS(new_assum),aux);
  setMapPut(quanname,unit,&S_TYCONS(state));
  KEEPONLY(new_assum);
  return new_assum;
}

PRIVATE ExprPtr tcUnit(expr,assum,state)
	/* A_UNIT, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,units,e,udesc,new_assum;
  ENTER;

  units = NewMap();
  for (args=ARGS(expr); args!=NIL; args=CDR(args))
    {
      e = CAR(args);
      udesc = TC(e,assum,state);
      units = MapPut(CAR(udesc),CDR(udesc),units);
    }
  new_assum = NewTypeAssum();
  SETQ(A_UNITS(new_assum),units);
  KEEPONLY(new_assum);
  return new_assum;
}

PRIVATE ExprPtr tcNewUnit(expr,assum,state)
	/* A_NEWUNIT, return unit::scale::type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,scale,unit,type,e;
  ENTER;

  args = ARGS(expr);
  unit = CAR(ARGS(CAR(args)));
  args = CDR(args);
  scale = CAR(ARGS(CAR(args)));
  args = CDR(args);
  e = CAR(args);
  type = TC(e,assum,state); /* and scales number e in place */
  scale = z_div(CAR(ARGS(e)),scale);
  (void) KEEP(scale);
  scale = KONS(scale,((CDR(ARGS(e))==NIL)?NewMap():CAR(CDR(ARGS(e)))));
  expr = Kons(unit,KONS(scale,type));
  KEEPONLY(expr);
  return expr;
}

PRIVATE ExprPtr tcType(expr,assum,state)
	/* A_TYPE, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,e,tdesc,new_assum;
  ENTER;

  new_assum = NewTypeAssum();
  for (args=ARGS(expr); args!=NIL; args=CDR(args))
    {
      e = CAR(args);
      tdesc = TC(e,assum,state);
      setMapPut(CAR(tdesc),KONS(getTyHdr(CAR(CDR(tdesc)),NIL,state),CDR(CDR(tdesc))),
		&A_TYFCNS(new_assum));
    }
  KEEPONLY(new_assum);
  return new_assum;
}

PRIVATE ExprPtr tcTypeBind(expr,assum,state)
	/* A_TYPE_BIND, return tyname::tyvarseq::type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,name,type,e,tyvarseq;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  tyvarseq = TC(e,assum,state); /* a A_TYPE_VAR_SEQ elaborates to a list of symbols */
  args = CDR(args);
  name = CAR(ARGS(CAR(args)));
  args = CDR(args);
  e = CAR(args);
  type = TC(e,assum,state);
  expr = Kons(name,KONS(tyvarseq,type));
  KEEPONLY(expr);
  return expr;
}

PRIVATE ExprPtr letrecDatatype(expr,assum,state,spec)
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
OUT MapPtr *spec;
{
  ExprPtr args,dtb,a,tyvarseq,e,id,tyname,new_assum,type,tyhdr,oldtyfcn=NOEXPR,s1,s2;
  Boolean found;

  args = CDR(ARGS(expr));
  new_assum = DupTypeAssum(assum);
  *spec = NewMap();
  for (; args!=NIL; args=CDR(args))
    {
      dtb = CAR(args);
      a = ARGS(dtb);
      e = CAR(a);
      tyvarseq = TC(e,assum,state);
      id = CAR(ARGS(CAR(CDR(a))));
      found = false;
      if (A_KIND(assum)==TA_PRECOMPILE)
	{
	  oldtyfcn = MapGet(id,A_TYEXTS(assum),&found);
	  if (found)
	    {
	      /* imported datatype or dimension */
	      if (CDR(oldtyfcn)->kind!=(CHAR)E_TYPE_APPL)
		{
		  type_soft_error(ZONE(CAR(CDR(a))),(CHAR *)varMsg(message(822),id));
		  goto fresh;
		}
	      tyname = applFun(CDR(oldtyfcn),state);
	      *spec = SetAdd(tyname,*spec);
	    }
	  else
	    {
	      (void) MapGet(id,A_TYFCNS(assum),&found);
	      if (found)
		type_soft_error(ZONE(CAR(CDR(a))),(CHAR *)varMsg(message(782),id));
	      goto fresh;
	    }
	}
      else
	{
fresh:
	  found = false;
	  tyname = freshTyName(id,assum,state);
	}
      type = KtypeAppl(tyname,tyvarseq);
      (void) KEEP(type);
      tyhdr = getTyHdr(tyvarseq,NIL,state);
      if (!found ||
	  (ListLength(CAR(tyhdr))==ListLength(CAR(CAR(oldtyfcn))) &&
	   (getTyHdrSubsts(tyhdr,CAR(oldtyfcn),&s1,&s2,state),
	    s2==NewMap())))
	{
	  setMapPut(id,KONS(tyhdr,type),&A_TYFCNS(new_assum));
	  if (found)
	    /* signal that extern datatype has been redefined;
	       put it in regular state. */
	    setMapRemove(id,&A_TYEXTS(assum));
	}
      else if (found)
	{
	  typeError(ZONE(e),(CHAR *)message(783),CDR(oldtyfcn),state);
	}
    }
  /* no KEEPONLY or QUIT ! It would enable the GC to free up *spec... */
  return new_assum;
}

PRIVATE void patchConsInfo P((ExprPtr e, ExprPtr state));

PRIVATE ExprPtr tcDataType(expr,assum,state)
	/* A_DATATYPE, A_DATATYPE_DESC, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr new_assum,e,consmap,r,spec;
  ENTER;

  e = CAR(ARGS(expr));
  new_assum = letrecDatatype(e,assum,state,&spec);
  r = TC(e,new_assum,state);
  consmap = CAR(r); /* maps type ids to cons (maps constructor -> type) */
  new_assum = CDR(r);
  maximizeEquality(consmap,A_TYFCNS(new_assum),state,NewMap(),spec,
		   (A_FLAGS(assum) & A_NOEQ_DATATYPES)!=0,ZONE(expr));
  patchConsInfo(CAR(ARGS(e)),state);
  checkIdExts(A_IDS(new_assum),A_IDLOCS(new_assum),assum,state);
  KEEPONLY(new_assum);
  return new_assum;
}

PRIVATE ExprPtr tcDataTypeBind(expr,assum,state)
	/* A_DATATYPE_BIND, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,e,tyhdr,name,cons,tyname,type,new_assum,con_t,con,t;
  ENTER;

  args = ARGS(expr);
  /*e = CAR(args);
    tyvarseq = TC(e,assum,state);*/
  args = CDR(args);
  name = CAR(ARGS(CAR(args)));
  args = CDR(args);
  type = MapGetSure(name,A_TYFCNS(assum));
  tyhdr = CAR(type);
  type = CDR(type);
  tyname = applFun(type,state);
  tyname = fxSym(tyname);
  new_assum = NewTypeAssum();
  cons = NewMap();
  /*ids = NewMap();*/
  for (;args!=NIL; args=CDR(args))
    {
      e = CAR(args);
      con_t = TC(e,assum,state); /* conname::type or conname */
      if (con_t->kind!=(CHAR)E_CONS)
	{
	  con = con_t;
	  t = type;
	  /*type = newTypeInst(t,assum,state
	    #if IMPERATIVE_TYPES!=EFFECT_TYVAR
	    ,false
	    #endif
	    );
	    no need to instantiate, it will be generalized
	    later anyway*/
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	  /*type = Weaken(type,assum,state,1L);
	    no need to weaken, the type could
	    actually be strengthened to infinity,
	    because a constant has all subtypes*/
#endif
	}
      else
	{
	  con = CAR(con_t);
	  t = funType(CDR(con_t),type);
	  /*type = t;
	    type = newTypeInst(t,assum,state
	    #if IMPERATIVE_TYPES!=EFFECT_TYVAR
	    ,false
	    #endif
	    );
	    no need to instantiate, it will be generalized
	    later anyway*/
	}
      cons = MapPut(con,t,cons);
      /*ids = MapPut(con,type,ids); useless, now, since ids==cons */
      setMapPut(con,e,&A_IDLOCS(new_assum));
    }
  SETQ(A_IDS(new_assum),cons); /* IDEXTS remains empty. */
  SETQ(A_CONS(new_assum),cons);
  e = MapSingleton(name,KONS(tyhdr,type));
  SETQ(A_TYFCNS(new_assum),e);
  setMapPut(tyname,KONS(tyhdr,cons),&S_TYCONS(state));
  KEEPONLY(new_assum);
  return new_assum;
}

PRIVATE ExprPtr tcAbsType(expr,assum,state)
	/* A_ABSTYPE, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,declist,e,eassum,local_assum,new_assum,noneqs,name,tyname;
  ExprPtr consmap,r,oldstycons,aux,spec;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  eassum = letrecDatatype(e,assum,state,&spec);
  args = CDR(args);
  declist = CAR(args);
  args = CDR(args);
  oldstycons = KEEP(S_TYCONS(state));
  local_assum = beginAssum(assum);
  r = TC(e,eassum,state);
  consmap = CAR(r);
  eassum = CDR(r);
  SETQ(S_TYCONS(state),oldstycons); /* forget internal details */
  maximizeEquality(consmap,A_TYFCNS(eassum),state,getSureeq(args),spec,
		   (A_FLAGS(assum) & A_NOEQ_DATATYPES)!=0,ZONE(expr));
  patchConsInfo(CAR(ARGS(e)),state);
  AddTypeAssum(local_assum,eassum);
  noneqs = InitAbsEq(A_TYFCNS(eassum),state);
  for (; args!=NIL; args=CDR(args))
    {
      name = CAR(args);
      tyname = TC(name,local_assum,state);
      aux = tdgAnd(funEqTdg(tyname,state),S_EQCONDS(state));
      if (aux==falseValue)
	type_error(ZONE(name),(CHAR *)nameMsg(message(60),tyname));
      SETQ(S_EQCONDS(state),aux);
      noneqs = SetRemove(tyname,noneqs);
    }
  new_assum = TC(declist,local_assum,state);
  SETQ(A_IDS(eassum),NewMap());
  SETQ(A_IDEXTS(eassum),NewMap());
  SETQ(A_CONS(eassum),NewMap()); /* forget all constructors */
  setMapMerge(&S_NONEQS(state),noneqs);
  AddTypeAssum(eassum,new_assum);
  endAssum(assum,local_assum);
  checkIdExts(A_IDS(eassum),A_IDLOCS(eassum),assum,state);
  {
    MapPtr imported,aux;

    /*
      Any imported identifier in new_assum should be added as imported
      to assum2.
    */
    imported = MapDomRestrTo (A_IDS(new_assum), A_IDEXTS(new_assum));
    aux = MapOverwrite (imported, A_IDS(eassum));
    SETQ(A_IDS(eassum), aux);
    aux = MapOverwrite (imported, A_IDEXTS(eassum));
    SETQ(A_IDEXTS(eassum), aux);
    imported = MapDomRestrTo (A_IDLOCS(new_assum), imported);
    aux = MapOverwrite (imported, A_IDLOCS(eassum));
    SETQ(A_IDLOCS(eassum), aux);
    /*
      Any imported type... similarly: no, this is useless.
    imported = A_TYEXTS(new_assum);
    setMapMerge (&A_TYEXTS(assum2), imported);
    setMapMerge (&A_TYFCNS(assum2), imported);
     */
  }
  KEEPONLY(eassum);
  return eassum;
}

PRIVATE ExprPtr tcTypeVarSeq(expr,assum,state)
	/* A_TYPE_VAR_SEQ, return list of type variables
	(note: '#a is listed as {'#a => 1.0}) */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ENTER;
  expr = tcTuple1(ARGS(expr),assum,state);
  KEEPONLY(expr);
  return expr;
}

PRIVATE ExprPtr tcConstructor(expr,assum,state)
	/* A_CONSTRUCTOR, return conname::type or conname */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,e,sym,type;
  ENTER;

  args = ARGS(expr);
  sym = CAR(ARGS(CAR(args)));
  args = CDR(args);
  if (args==NIL)
    e = sym;
  else
    {
      e = CAR(args);
      type = TC(e,assum,state);
      e = Kons(sym,type);
    }
  KEEPONLY(e);
  return e;
}

PRIVATE ExprPtr tcException(expr,assum,state)
	/* A_EXCEPTION, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,e,new_assum,type;
  ENTER;

  new_assum = NewTypeAssum();
  SETQ(A_NONGENS(new_assum),A_NONGENS(assum));
  for (args=ARGS(expr); args!=NIL; args=CDR(args))
    {
      e = CAR(args);
      type = TC(e,assum,state); /* id::type */
      setMapPut(CAR(type),CDR(type),&A_IDS(new_assum));
      /* IDEXTS remains empty so as to define exceptions as regular. */
      setMapPut(CAR(type),e,&A_IDLOCS(new_assum));
    }
  checkIdExts(A_IDS(new_assum),A_IDLOCS(new_assum),assum,state);
  KEEPONLY(new_assum);
  return new_assum;
}

PRIVATE ExprPtr tcNewExc(expr,assum,state)
	/* A_NEWEXC, [A_EXT_EXC,] return (sym,type) */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,type,e,sym;
  ENTER;

  /* in case we are in a TA_PRECOMPILE assum,
     patch the code to be an A_EXT_EXC instead of an A_NEW_EXC,
     so that compile.c knows that it should not compile a
     generative exception, but an exception known at compile-time
     (otherwise we cannot share exceptions across modules...) */
  if (A_KIND(assum)==TA_PRECOMPILE)
    TYPE(expr) = A_EXT_EXC;
  args = ARGS(expr);
  sym = CAR(ARGS(CAR(args)));
  args = CDR(args);
  if (args==NIL)
    type = exnType;
  else
    {
      e = CAR(args);
      type = TC(e,assum,state);
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
      if (!impTypep(type,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
		    ,INF_STRENGTH
#endif
		    ))
	{
	  typeError(ZONE(e),(CHAR *)message(61),type,state);
	}
#endif
      type = typeSubst(ZONE(e),type,A_TYVARS(assum),assum,state);
      type = funType(type,exnType);
    }
#if DEBUGGING!=0
  addAnnotation(expr,type,A_FNLEVEL(assum),state);
#endif
  e = Kons(sym,type);
  KEEPONLY(e);
  return e;
}

PRIVATE ExprPtr tcExcBind(expr,assum,state)
	/* A_EXC_BIND, return (sym,type) */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,sym,sym2,type,f,type2;
  Boolean found;
  ENTER;

  args = ARGS(expr);
  sym = CAR(ARGS(CAR(args)));
  args = CDR(args);
  sym2 = CAR(ARGS(CAR(args)));
  type = correctTypeAppl(MapGet(sym2,A_IDS(assum),&found),state);
  if (!(found && (
		  (/*type!=NIL && */type->kind==(CHAR)E_TYPE_APPL &&
		   APPLFUN(type)==exnSym) ||
		  (/*type!=NIL && */type->kind==(CHAR)E_TYPE_APPL
		   && (f = APPLFUN(type), fxSym(f)==funSym)
		   && (type2 = correctTypeAppl(CAR(CDR(APPLARG(type))),state),
		       (/*type2!=NIL && */type2->kind==(CHAR)E_TYPE_APPL &&
			APPLFUN(type2)==exnSym))))))
    type_error(ZONE(CAR(args)),(CHAR *)varMsg(message(62),sym2));
  expr = Kons(sym,type);
  KEEPONLY(expr);
  return expr;
}

PRIVATE ExprPtr tcMapletType(expr,assum,state)
	/* A_MAPLET_TYPE, return label::type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,label,e,type;
  ENTER;

  args = ARGS(expr);
  label = CAR(args);
  label = TC(label,assum,state);
  args = CDR(args);
  e = CAR(args);
  type = TC(e,assum,state);
  e = Kons(label,type);
  KEEPONLY(e);
  return e;
}

PRIVATE ExprPtr tcRecordType(expr,assum,state)
	/* A_RECORD_TYPE, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,e,ltype,type;
  ENTER;

  type = NewMap();
  for (args=ARGS(expr); args!=NIL; args=CDR(args))
    {
      e = CAR(args);
      ltype = TC(e,assum,state);
      if (TYPE(e)==A_TYPE_ELLIPSIS)
	type = MapPut(REST_TAG,ltype,type);
      else
	type = MapPut(CAR(ltype),CDR(ltype),type);
    }
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcTypeApply(expr,assum,state)
	/* A_TYPE_APPLY, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,tname,tyfcn,subst;
  Boolean found;
  ENTER;

  args = ARGS(expr);
  tname = CAR(ARGS(CAR(args)));
  tyfcn = MapGet(tname,A_TYFCNS(assum),&found);
  if (!found || !ISCONS(tyfcn))
    type_error(ZONE(CAR(args)),(CHAR *)varMsg(message(63),tname));
  args = CDR(args);
  subst = getTypeSubst(ZONE(expr),CAR(tyfcn),args,assum,state);
  expr = typeSubst(ZONE(expr),CDR(tyfcn),subst,assum,state);
  KEEPONLY(expr);
  return expr;
}

PRIVATE ExprPtr tcTypePower(expr,assum,state)
	/* A_TYPE_POWER, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,e,type,num;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  type = TC(e,assum,state);
  args = CDR(args);
  num = CAR(ARGS(CAR(args)));
  type = typePower(ZONE(e),type,num,state);
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcTypeStar(expr,assum,state)
	/* A_TYPE_STAR, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,e1,e2,t1,t2;
  ENTER;

  args = ARGS(expr);
  e1 = CAR(args);
  t1 = TC(e1,assum,state);
  args = CDR(args);
  e2 = CAR(args);
  t2 = TC(e2,assum,state);
  t1 = typeMul(ZONE(e1),ZONE(e2),t1,t2,state);
  KEEPONLY(t1);
  return t1;
}

PRIVATE ExprPtr tcTupleType1(args,assum,state)
IN CONST ExprPtr args;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr e,type;
  ExprPtr *exprs,*ep;
  COUNTER i,n;

  type = NIL;
  if (args!=NIL)
    {
      n = ListLength(args);
      exprs = (ExprPtr *)alloctmp(n*sizeof(ExprPtr));
      for (i=0, ep=exprs; i<n; i++, args=CDR(args))
	{
	  e = CAR(args);
	  if (TYPE(e)==A_TYPE_ELLIPSIS)
	    {
	      type = TC(e,assum,state);
	      break;
	    }
	  *ep++ = TC(e,assum,state);
	}
      for (; ep!=exprs; )
	type = KONS(*--ep,type);
    }
  return type;
}

PRIVATE ExprPtr tcTupleType(expr,assum,state) /* A_TUPLE_TYPE, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr type;
  ENTER;

  type = tcTupleType1(ARGS(expr),assum,state);
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcFunctionType(expr,assum,state)
	/* A_FUNCTION_TYPE, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,e1,e2,t1,t2;
  ENTER;

  args = ARGS(expr);
  e1 = CAR(args);
  t1 = TC(e1,assum,state);
  args = CDR(args);
  e2 = CAR(args);
  t2 = TC(e2,assum,state);
  t1 = funType(t1,t2);
  KEEPONLY(t1);
  return t1;
}

PRIVATE ExprPtr tcMapType(expr,assum,state)
	/* A_MAP_TYPE, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,e1,e2,t1,t2,eqtdg;
  ENTER;

  args = ARGS(expr);
  e1 = CAR(args);
  t1 = TC(e1,assum,state);
  eqtdg = tdgAnd(eqTypeTdg(t1,state,false),S_EQCONDS(state));
  if (eqtdg==falseValue)
    typeError(ZONE(e1),(CHAR *)message(64),t1,state);
  SETQ(S_EQCONDS(state),eqtdg);
  args = CDR(args);
  e2 = CAR(args);
  t2 = TC(e2,assum,state);
  t1 = mapType(t1,t2);
  KEEPONLY(t1);
  return t1;
}

PRIVATE ExprPtr tcPatAny(expr,assum,state)
	/* A_PAT_ANY, return type::ve::vz */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr type;
  ENTER;

#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  type = Kons(freshInstfromFlags(0,falseValue,assum,state,INF_STRENGTH,INF_STRENGTH),
	      KONS(NewMap(),NewMap()));
#else
  type = Kons(freshInstfromFlags(0,falseValue,assum,state),
	      KONS(NewMap(),NewMap()));
#endif
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcPatRecordEllipsis(expr,assum,state)
	/* A_PAT_RECORD_ELLIPSIS, return REST_TAG::type::ve::vz */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr type;
  ENTER;

#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  type = Kons(REST_TAG,KONS(freshInstfromFlags(0,falseValue,assum,state,
					       INF_STRENGTH,INF_STRENGTH),
			    KONS(NewMap(),NewMap())));
#else
  type = Kons(REST_TAG,KONS(freshInstfromFlags(0,falseValue,assum,state),
			    KONS(NewMap(),NewMap())));
#endif
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcIllegal(expr,assum,state)
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  crash_error(ZONE(expr),(CHAR *)message(65));
  return NIL;
}

#define tcPatTupleEllipsis tcIllegal

/*PRIVATE ExprPtr tcPatTupleEllipsis(expr,assum,state)*/
	/* A_PAT_TUPLE_ELLIPSIS, return type::ve::vz */
	/* notice this is the same code as A_PAT_ANY; there can be
	no confusion because the pattern (...) is not allowed */
/*IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
	ExprPtr type;
	ENTER;

	type = Kons(MapSingleton(REST_TAG,freshInstfromFlags(0,falseValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
							,INF_STRENGTH,INF_STRENGTH
#endif
							)),
					KONS(NewMap(),NewMap()));
	KEEPONLY(type);
	return type;
}
*/

PRIVATE ExprPtr tcPatRecordMaplet(expr,assum,state)
	/* A_PAT_RECORD_MAPLET, return label::type::ve::vz */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,e,label,t_ve_vz;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  label = TC(e,assum,state);
  args = CDR(args);
  e = CAR(args);
  t_ve_vz = TC(e,assum,state);
  t_ve_vz = Kons(label,t_ve_vz);
  KEEPONLY(t_ve_vz);
  return t_ve_vz;
}

PRIVATE ExprPtr tcPatRecord1(list,assum,state,vep,vezp)
IN CONST ExprPtr list;
INOUT ExprPtr assum;
INOUT ExprPtr state;
OUT MapPtr *vep,*vezp;
{
  ExprPtr pat,tmap,label,t,type;

  *vep = NewMap();
  *vezp = NewMap();
  tmap = NewMap();
  for (; list!=NIL; list=CDR(list))
    {
      pat = CAR(list);
      type = TC(pat,assum,state);
      label = CAR(type);
      type = CDR(type);
      t = CAR(type);
      type = CDR(type);
      checkRep(CAR(type),*vep,*vezp);
      *vep = MapMerge(CAR(type),*vep);
      *vezp = MapMerge(CDR(type),*vezp);
      tmap = MapPut(label,t,tmap);
    }
  return tmap;
}

PRIVATE ExprPtr tcPatRecord(expr,assum,state)
	/* A_PAT_RECORD, return type::ve::vz */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr tmap,ve,vz;
  ENTER;

  tmap = tcPatRecord1(ARGS(expr),assum,state,&ve,&vz);
  addAnnotation(expr,tmap,A_FNLEVEL(assum),state); /* kept for compilation */
  tmap = Kons(tmap,KONS(ve,vz));
  KEEPONLY(tmap);
  return tmap;
}

PRIVATE ExprPtr tcPatAs(expr,assum,state)
	/* A_PAT_AS, return type::ve::vz */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,e,t1,t2,tvevz1,tvevz2,aux,aux1,aux2;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  tvevz1 = TC(e,assum,state);
  t1 = CAR(tvevz1);
  tvevz1 = CDR(tvevz1);
  args = CDR(args);
  e = CAR(args);
  tvevz2 = TC(e,assum,state);
  t2 = CAR(tvevz2);
  tvevz2 = CDR(tvevz2);
  checkRep(CAR(tvevz2),CAR(tvevz1),CDR(tvevz1));
  aux1 = MapMerge(CAR(tvevz2),CAR(tvevz1));
  aux2 = MapMerge(CDR(tvevz2),CDR(tvevz1));
  aux = Cons(aux1,aux2);
  (void) coerceTypes(ZONE(expr),t2,t1,assum,state,false);
  MAKECONS(t1,aux,tvevz2);
  KEEPONLY(tvevz2);
  return tvevz2;
}

PRIVATE ExprPtr tcPatCast(expr,assum,state) /* A_PAT_CAST, return type::ve::vz */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,type,type2,tvevz,e;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  tvevz = TC(e,assum,state);
  type = CAR(tvevz);
  e = CAR(CDR(args));
  type2 = typeSubst(ZONE(e),TC(e,assum,state),A_TYVARS(assum),assum,state);
  (void) coerceTypes(ZONE(expr),type,type2,assum,state,false);
  KEEPONLY(tvevz); /* must not return type2, which could have NIL holes
		      in latent effects */
  return tvevz;
}

PRIVATE ExprPtr tcPatAnyList(expr,assum,state)
	/* A_PAT_NIL, return type::ve::vz */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr tvevz;
  ENTER; /* type = 'a list */

#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  tvevz = Kons(listType(freshInstfromFlags(0,falseValue,assum,state,
					   INF_STRENGTH,INF_STRENGTH)),
	       KONS(NewMap(),NewMap()));
#else
  tvevz = Kons(listType(freshInstfromFlags(0,falseValue,assum,state)),
	       KONS(NewMap(),NewMap()));
#endif
  KEEPONLY(tvevz);
  return tvevz;
}

#define tcPatNil tcPatAnyList

PRIVATE ExprPtr tcPatApply(expr,assum,state)
	/* A_PAT_APPLY, return type::ve:vz */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,f,arg,ftype,argtype;
  ENTER;

  args = ARGS(expr);
  f = CAR(args);
  ftype = TC(f,assum,state); /* type::ve::vz */
  ftype = CAR(ftype);
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  ftype = Weaken(ftype,assum,state,1L);
#endif
  arg = CAR(CDR(args));
  argtype = TC(arg,assum,state);
  ftype = tcApply1(ZONE(f),ZONE(arg),ftype,CAR(argtype),assum,state);
  addAnnotation(expr,ftype,A_FNLEVEL(assum),state); /* kept for compilation */
  ftype = Kons(ftype,CDR(argtype));
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  cvtPatFx(); /* constructor patterns have read, not init effects */
#endif
  KEEPONLY(ftype);
  return ftype;
}

PRIVATE ExprPtr tcPatAnyMap(expr,assum,state)
	/* A_PAT_ANYMAP, return type::ve::vz */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr tvevz;
  ENTER; /* type = ''a -m> 'b */

#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  tvevz = Kons(mapType(freshInstfromFlags(F_EQ,trueValue,assum,state,
					  INF_STRENGTH,INF_STRENGTH),
		       freshInstfromFlags(0,falseValue,assum,state,
					  INF_STRENGTH,INF_STRENGTH)),
	       KONS(NewMap(),NewMap()));
#else
  tvevz = Kons(mapType(freshInstfromFlags(F_EQ,trueValue,assum,state),
		       freshInstfromFlags(0,falseValue,assum,state)),
	       KONS(NewMap(),NewMap()));
#endif
  KEEPONLY(tvevz);
  return tvevz;
}

#define tcPatEmpty tcPatAnyMap

PRIVATE ExprPtr tcPatMaplet(expr,assum,state)
	/* A_PAT_MAPLET, return type1::type2::ve::vz */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,domtype,rngtype,e,domt_ve_vz,rngt_ve_vz,ve,vz;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  domt_ve_vz = TC(e,assum,state);
  domtype = CAR(domt_ve_vz);
  coerceFlagTypes(ZONE(CAR(args)),0/*F_EQ*/,trueValue,
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
		  INF_STRENGTH,INF_STRENGTH,
#endif
		  domtype,assum,state);
  e = CAR(CDR(args));
  rngt_ve_vz = TC(e,assum,state);
  rngtype = CAR(rngt_ve_vz);
  checkRep(CAR(CDR(domt_ve_vz)),CAR(CDR(rngt_ve_vz)),CDR(CDR(rngt_ve_vz)));
  ve = MapMerge(CAR(CDR(domt_ve_vz)),CAR(CDR(rngt_ve_vz)));
  vz = MapMerge(CDR(CDR(domt_ve_vz)),CDR(CDR(rngt_ve_vz)));
  expr = Kons(domtype,KONS(rngtype,KONS(ve,vz)));
  KEEPONLY(expr);
  return expr;
}

PRIVATE ExprPtr tcPatMapAdd(expr,assum,state)
	/* A_PAT_MAPADD, return type::ve::vz */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,e,t1,t2,domtype,rngtype,type,ve,vz;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  t1 = TC(e,assum,state);
  domtype = CAR(t1);
  t1 = CDR(t1);
  rngtype = CAR(t1);
  t1 = CDR(t1);
  type = mapType(domtype,rngtype);
  args = CDR(args);
  e = CAR(args);
  t2 = TC(e,assum,state);
  (void) coerceTypes(ZONE(e),CAR(t2),type,assum,state,false);
  t2 = CDR(t2);
  checkRep(CAR(t1),CAR(t2),CDR(t2));
  ve = MapMerge(CAR(t1),CAR(t2));
  vz = MapMerge(CDR(t1),CDR(t2));
  addAnnotation(expr,type,A_FNLEVEL(assum),state); /* kept for compilation */
  type = Kons(type,KONS(ve,vz));
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcPrimeDataBind(expr,assum,state)
	/* A_PRIME_DATABIND, return consmap::assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,e,with_assum,local_assum,new_assum,eassum,
  consmap,id,cons,tyname,tykons,aux;
#if DEBUGGING!=0
  ExprPtr subdecls;
#endif
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  with_assum = TC(e,assum,state);
  local_assum = DupTypeAssum(assum);
  AddTypeAssum(local_assum,with_assum);
  new_assum = NewTypeAssum();
  consmap = NewMap();
  tykons = NewMap();
#if DEBUGGING!=0
  subdecls = NewMap();
#endif
  while (args = CDR(args), args!=NIL)
    {
      e = CAR(args);
      eassum = TC(e,local_assum,state);
      id = ChooseInDom(A_TYFCNS(eassum));
      cons = A_CONS(eassum);
      consmap = MapPut(id,cons,consmap);
      AddTypeAssum(new_assum,eassum);
      tyname = applFun(CDR(MapGetSure(id,A_TYFCNS(local_assum))),state);
      tyname = fxSym(tyname);
      tykons = MapPut(tyname,cons,tykons);
#if DEBUGGING!=0
      subdecls = MapPut(tyname,e,subdecls);
      addAnnotation(e,cons,A_FNLEVEL(assum),state);
      /* cheat : we don't put the type of constructors as annotation,
	 but the map from constructors to types (the so-called constructor map).
	 This is used in wrapDebugName(). */
#endif
    }
  updateAssum(assum,local_assum);
  args = Cons(tykons,Cons(NewMap(),NIL)); /* second argument will
					     be patched after calling
					     maximizeEquality(), to hold
					     the set of tynames in tykons
					     that admit equality by design.
					     (This is for abstypes: see
					     Test/abstype1.ml for a type
					     that admits equality by design
					     but won't admit equality after
					     the abstype context is closed.) */
#if DEBUGGING!=0
  args = Cons(subdecls,args);
#endif
  aux = ZBuild(ZONE(e),A_CONS_INFO,args);
#ifdef OBSOLETE
  formerly, we did:
  e = Cons(KEEP(aux),ARGS(expr)); /* patch in place */
  SETQ(ARGS(expr),e);
  but this last SETQ cannot be recorded in any write barrier,
  so we do the following, which stores the new objects
    inside the first cons cells of the argument list
#endif
  (void) KEEP(aux);
  args = ARGS(expr);
  e = Cons(CAR(args),CDR(args));
  SETCDR(args,e);
  SETCAR(args,aux); /* this is a bit messy, that's right */
  expr = Kons(consmap,new_assum);
  KEEPONLY(expr);
  return expr;
}

MAPCOLLECTFUNC(do_patchconsinfo)
{
  if (funEqTdg(x,MAPDATA(ExprPtr))!=falseValue)
    return MapSingleton(x,UNIT);
  return NewMap();
}

PRIVATE void patchConsInfo(expr,state)
INOUT ExprPtr expr; /* should be an A_CONS_INFO */
IN CONST ExprPtr state;
{
  ExprPtr args, tyeqs;
  ENTER;

  args = ARGS(expr);
#if DEBUGGING!=0
  args = CDR(args);
#endif
  tyeqs = MapCollect(CAR(args),do_patchconsinfo,(void *)state);
  args = CDR(args);
  SETCAR(args,tyeqs);
  QUIT;
}

PRIVATE ExprPtr tcUnitName(expr,assum,state)
	/* A_UNIT_NAME, return couple (scale,type) */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr unitname,unit;
  Boolean found;

  getLongStruct(ZONE(expr),CDR(ARGS(expr)),&assum,state);
  unitname = CAR(ARGS(expr));
  unit = MapGet(unitname,A_UNITS(assum),&found);
  if (!found)
    type_error(ZONE(expr),(CHAR *)varMsg(message(73),unitname));
  return unit;
}

PRIVATE ExprPtr tcTypeVar(expr,assum,state) /* A_TYPE_VAR, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr type;

  type = CAR(ARGS(expr));
  if (varFlags(type) & F_NUM)
    {
      type = KnumType(MapSingleton(type,one));
      (void) KEEP(type);
    }
  return type;
}

PRIVATE ExprPtr tcRealVar(expr,assum,state)
	/* A_REAL_VAR, return type::ve::vz */
	/* A_REAL_VAR is always a varintro */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr tyvar,ve,vz,var,aux,tvevz;
  ENTER;

  var = CAR(ARGS(expr));
  if (var==equalSym)
    type_warn(ZONE(expr),(CHAR *)message(66));
  tyvar = freshInstfromFlags(0,falseValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			     ,INF_STRENGTH,INF_STRENGTH
#endif
			     );
  vz = MapSingleton(var,expr); /* to record the zone of expr */
  ve = MapSingleton(var,tyvar);
  aux = Cons(ve,vz);
  MAKECONS(tyvar,aux,tvevz);
  KEEPONLY(tvevz);
#if DEBUGGING!=0
  addAnnotation(expr,tyvar,A_FNLEVEL(assum),state);
#endif
  return tvevz;
}

PRIVATE ExprPtr tcPatVar(expr,assum,state)
	/* A_PAT_VAR, return type::ve::vz */
	/* A_PAT_VAR may be patched to become a A_REAL_VAR when a varintro */
INOUT ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr type,ve,vz,var,aux,tvevz;
  ENTER;

  var = CAR(ARGS(expr));
  getLongStruct(ZONE(expr),CDR(ARGS(expr)),&assum,state);
  type = varType(var,assum,state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
		 ,true
#endif
		 );
  /* flag noimpp is true, so that, for example,
     pattern ref has type 'a -> 'a ref, not
     '_a -> '_a ref */
  if (type!=NOEXPR)
    {
      vz = NewMap();
      ve = NewMap();
    }
  else
    {
      TYPE(expr) = A_REAL_VAR;
      type = freshInstfromFlags(0,falseValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
				,INF_STRENGTH,INF_STRENGTH
#endif
				);
      vz = MapSingleton(var,expr); /* to record the zone of expr */
      ve = MapSingleton(var,type);
#if DEBUGGING!=0
      addAnnotation(expr,type,A_FNLEVEL(assum),state);
#endif
    }
  aux = Cons(ve,vz);
  MAKECONS(type,aux,tvevz);
  KEEPONLY(tvevz);
  return tvevz;
}

PRIVATE ExprPtr tcPatConst(expr,assum,state)
	/* A_PAT_CONST, return type::ve::vz */
INOUT ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr type,e,aux,tvevz;
  ENTER;

  e = CAR(ARGS(expr));
  type = TC(e,assum,state);
  aux = Cons(NewMap(),NewMap());
  MAKECONS(type,aux,tvevz);
  KEEPONLY(tvevz);
  return tvevz;
}

PRIVATE ExprPtr tcSelectFn(expr,assum,state)
	/* A_SELECT_FN, return type */
INOUT ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr label,type,tyvar;
  COUNTER i;
  ENTER;

  label = CAR(ARGS(expr));
  if (label->kind==(CHAR)E_INT)
    /* #1 : 'a * ... -> 'a,
       #2 : 'b * 'a * ... -> 'a
       ...
       */
    {
      i = IVAL(label);
      if (i<=0)
	type_error(ZONE(expr),(CHAR *)message(67));
      tyvar = freshInstfromFlags(0,falseValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
				 ,INF_STRENGTH,INF_STRENGTH
#endif
				 );
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
      type = KONS(tyvar,freshInstfromFlags(0,falseValue,assum,state,
					   INF_STRENGTH,INF_STRENGTH));
#else
      type = KONS(tyvar,freshInstfromFlags(0,falseValue,assum,state));
#endif
      while (--i!=0)
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	type = KONS(freshInstfromFlags(0,falseValue,assum,state,
				       INF_STRENGTH,INF_STRENGTH),type);
#else
      type = KONS(freshInstfromFlags(0,falseValue,assum,state),type);
#endif
    }
  else /* #lab : |[lab : 'a,...]| -> 'a */
    {
      tyvar = freshInstfromFlags(0,falseValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
				 ,INF_STRENGTH,INF_STRENGTH
#endif
				 );
      type = MapPair(label,tyvar,
		     REST_TAG,freshInstfromFlags(0,falseValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
						 ,INF_STRENGTH,INF_STRENGTH
#endif
						 ));
    }
  type = funTypeFx(type,tyvar,freshFxVar(NewMap(),effectSym,assum,state));
  addAnnotation(expr,type,A_FNLEVEL(assum),state); /* kept for compilation */
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcMu(expr,assum,state)
	/* A_MU, return type */
INOUT ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,e1,e2,rec,mod,all;
  ENTER;

  args = ARGS(expr);
  e1 = CAR(args);
  rec = TC(e1,assum,state);
  e2 = CAR(CDR(args));
  mod = TC(e2,assum,state); /* mod is a record type without ellipsis */
  all = overrideMapType(rec,mod,assum,state);
  if (all==NIL)
    typeError(ZONE(e1),(CHAR *)message(473),rec,state);
  addAnnotation(e1,rec,A_FNLEVEL(assum),state); /* kept for compilation */
  addAnnotation(e2,mod,A_FNLEVEL(assum),state); /* kept for compilation */
  addAnnotation(expr,all,A_FNLEVEL(assum),state); /* kept for compilation */
  KEEPONLY(all);
  return all;
}

PRIVATE ExprPtr tcSubMap(expr,assum,state)
	/* A_SUB_MAP, return (ve,vz) */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,pat,map,pattype,maptype;
  ENTER;

  args = ARGS(expr);
  pat = CAR(args);
  map = CAR(CDR(args));
  pattype = TC(pat,assum,state); /* = type::ve::vz */
  maptype = TC(map,assum,state); /* = type */
  (void) coerceTypes(ZONE(pat),maptype,
		     mapType(freshInstfromFlags(F_EQ,trueValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
						,INF_STRENGTH,INF_STRENGTH
#endif
						),
			     freshInstfromFlags(0,falseValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
						,INF_STRENGTH,INF_STRENGTH
#endif
						)),
		     assum,state,false);
  (void) coerceTypes(ZONE(pat),CAR(pattype),maptype,assum,state,false);
  pattype = CDR(pattype);
  KEEPONLY(pattype);
  return pattype;
}

PRIVATE ExprPtr tcDelay(expr,assum,state)
	/* A_DELAY, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr e,type;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  ExprPtr oldmaxstrengths;
#endif
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  ExprPtr oldFx;
#endif
  ENTER;

#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  oldmaxstrengths = S_MAXSTRENGTH(state);
  (void) KEEP(oldmaxstrengths);
  setMapMerge(&S_MAXSTRENGTH(state),S_STRENGTH(state));
  /* to prevent strengthening to go over the limit
     imposed by the strengths of outer type variables;
     this is actually redundant, since maxstrengths
     could be inferred (slowly) from S_TYVARSUBS and
     the strengths of outer type variables (which are
     those which were created before entering the delay) */
#endif
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  oldFx = KEEP(theFx);
  resetFx();
#endif
  e = CAR(ARGS(expr));
  A_FNLEVEL(assum)++;
  type = TC(e,assum,state);
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  type = Weaken(type,assum,state,-1L);
  setMapDomRestrTo(&S_MAXSTRENGTH(state),oldmaxstrengths);
#endif
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  type = promiseTypeFx(type,freshFxVar(theFx,effectSym,assum,state));
  SETQ(theFx,oldFx);
#else
  type = promiseType(type);
#endif
  A_FNLEVEL(assum)--;
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcPack1(expr,assum,state,patp)
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
IN Boolean patp;
	/* A_PACK (resp. A_PAT_PACK) return type (resp. type::ve::vz) */
{ /* pack : '_a -> dynamic; or '1a -> dynamic in STD_ML_NJ_TYVAR */
  /* does typing of pack <expr> as though we had written
     let val _ = <expr> in <dynamic> end, but (if <type> is the
     type of <expr> after quantification):
     - variables free in <type> must never be quantified
     by subsequent packs or vals (this is the purpose
     of the S_PACKERS field)
     - the scope of explicit type variables is never put
     on a dynamic
     - the return type is dynamic (dynType)
     */
  ExprPtr local_assum;
  ExprPtr e,type,aux,tvevz;
  ExprPtr memoref,impmemoref;
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  ExprPtr oldFx;
#endif
  ExprPtr subs;
  ENTER;

#if IMPERATIVE_TYPES==EFFECT_TYVAR
  oldFx = KEEP(theFx);
  resetFx();
#endif
  flushTyVarSubCache(state);
  subs = S_TYVARSUBS(state);
  (void) KEEP(subs);
  e = CAR(ARGS(expr));
  local_assum = beginAssum(assum);
  tvevz = TC(e,local_assum,state);
  if (patp)
    {
      type = CAR(tvevz);
#if IMPERATIVE_TYPES==EFFECT_TYVAR
      addTopFx(type,local_assum,state);
      /* this way, all possible effects (that we cannot compute
	 at compile time) are subsumed by 'top' effects.
	 The rule is that 'top(tau)' subsumes any effect having
	 at most the variables free in tau. This will match
	 anything because only observable effects are useful. */
#endif
    }
  else
    type = tvevz;

  memoref = newMemoRef();
  /* all type variables that are created during elaboration of expressions
     must be unquantified */
  flushTyVarSubCache(state);
  {
    ExprPtr modifs;

    aux = A_NONGENS(assum);
    modifs = MapDomRestrTo(A_NONGENS(assum),
			   MapDomRestrBy(S_TYVARSUBS(state),
					 subs));
    aux = MapMerge(aux,unquant_x(modifs,local_assum,state,memoref));
  }
  /* exclude all type variables
     appearing in old nongens from being quantified */
  aux = SetUnion(aux,MapDomRestrTo(A_NONGENS(local_assum),S_TYVARSUBS(state)));
  /* we exclude from quantification:
     - the old type variables A_NONGENS(assum) (those existing before tcVal)
     - all type variables that were created afterwards but became bound
     moreover, we also exclude (unquant) imperative type variables associated
     with expansive expressions
     */
#if IMPERATIVE_TYPES==STD_ML_TYVAR
  impmemoref = newMemoRef();
  if (expansivep(e))
    aux = SetUnion(aux,tyVars(type,local_assum,state,F_IMP,impmemoref));
#endif
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  impmemoref = newMemoRef();
  aux = SetUnion(aux,tyVars(type,local_assum,state,F_IMP,0L,impmemoref));
#endif
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  /* unquantify all variables that are free in the current effect */
  trimObsFx(assum,type,local_assum,state,memoref);
  aux = SetUnion(aux,unquant_x(theFx,local_assum,state,memoref));
#endif
  aux = SetUnion(aux,unquant_packers(local_assum,state,memoref));
  SETQ(A_NONGENS(local_assum),aux);
  endAssum(assum,local_assum);
  if (patp && (memoref = newMemoRef(),
	       tyVars(type,assum,state,0x80
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
		      ,INF_STRENGTH
#endif
		      ,memoref)!=NewMap()))
    type_error(ZONE(expr),(CHAR *)message(618));
  /* check that all variables have been quantified in patterns;
     this is important in EFFECT_TYVAR because the type is patched
     to include maximal effects (addTopFx); if unquantified type variables
     remained, they could be replaced by unpatched types */
  addPacker(expr,type,A_FNLEVEL(assum),state);
#if DEBUGGING!=0
  if (!patp)
    addAnnotation(expr,dynType,A_FNLEVEL(assum),state);
#endif
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  aux = SetUnion(oldFx,theFx);
  SETQ(theFx,aux);
#endif
  tvevz = patp?Cons(dynType,CDR(tvevz)):dynType;
  KEEPONLY(tvevz);
  return tvevz;
}

PRIVATE ExprPtr tcPack(expr,assum,state)
	/* A_PACK, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  return tcPack1(expr,assum,state,false);
}

PRIVATE ExprPtr tcPatPack(expr,assum,state) /* A_PAT_PACK, return type::ve::vz */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  return tcPack1(expr,assum,state,true);
}

PRIVATE ExprPtr tcInteger(expr,assum,state) /* A_INTEGER, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  return intType;
}

PRIVATE ExprPtr tcTypeEllipsis(expr,assum,state) /* A_TYPE_ELLIPSIS, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr e,type;
  ENTER;

  e = CAR(ARGS(expr));
  type = TC(e,assum,state);
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr tcExtDatatype1(expr,assum,state,eqp) /* A_EXT_DATATYPE, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
IN Boolean eqp;
{
  ExprPtr args,tydesc,a,tyvarseq,e,id,tyname,tyhdr,new_assum,type,consmap,fakecons;
  Boolean found;
  ENTER;

  args = ARGS(expr);
  new_assum = NewTypeAssum();
  consmap = NewMap(); /* a fake consmap, to be used by maximizeEquality() */
  fakecons = MapSingleton(trueValue,funType(eqp?NIL:funType(NIL,NIL), NIL));
  /* trueValue and type are dummies; the type is so that it respects (if eqp) or
     not (if not eqp) equality. */
  for (; args!=NIL; args=CDR(args))
    {
      tydesc = CAR(args);
      a = ARGS(tydesc);
      e = CAR(a);
      tyvarseq = TC(e,assum,state);
      id = CAR(ARGS(CAR(CDR(a))));
      (void) MapGet(id,A_TYFCNS(assum),&found);
      if (found)
	type_soft_error(ZONE(CAR(CDR(a))),(CHAR *)varMsg(message(790),id));
      tyname = freshTyName(id,assum,state);
      type = KtypeAppl(tyname,tyvarseq);
      (void) KEEP(type);
      tyhdr = getTyHdr(tyvarseq,NIL,state);
      type = KONS(tyhdr,type);
      setMapPut(id,type,&A_TYFCNS(new_assum)); /* to make the type known */
      setMapPut(tyname,KONS(tyhdr,NewMap()),&S_TYCONS(state)); /* associate the type name
								  with no constructor */
      consmap = MapPut(id,fakecons,consmap);
    }
  SETQ(A_TYEXTS(new_assum),A_TYFCNS(new_assum)); /* to record the constraint for
						    when the real declaration comes up. */
  maximizeEquality(consmap,A_TYFCNS(new_assum),state,NewMap(),NewMap(),
		   false,ZONE(expr));
  KEEPONLY(new_assum);
  return new_assum;
}

PRIVATE void checkExtern(e,assum)
IN CONST ExprPtr e;
IN CONST ExprPtr assum;
{
  if (A_KIND(assum)!=TA_PRECOMPILE)
    type_error(ZONE(e),(CHAR *)message(786));
}

PRIVATE ExprPtr tcExtDatatype(expr,assum,state) /* A_EXT_DATATYPE, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  checkExtern(expr,assum);
  return tcExtDatatype1(expr,assum,state,false);
}

PRIVATE ExprPtr tcExtEqtype(expr,assum,state) /* A_EXT_DATATYPE, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  checkExtern(expr,assum);
  return tcExtDatatype1(expr,assum,state,true);
}

MAPSWEEPFUNC(do_check_extquan)
{
  struct checkimport *data;
  Boolean found;

  data = MAPDATA(struct checkimport *);
  (void) MapGet(x,A_TYFCNS(data->assum),&found);
  if (found)
    {
      type_soft_error(ZONE(data->zone),(CHAR *)varMsg(message(790),x));
      data->err = true;
    }
  return false;
}

PRIVATE ExprPtr tcExtQuantity(expr,assum,state) /* A_EXT_QUANTITY, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr new_assum;
  struct checkimport *data;
  ENTER;

  checkExtern(expr,assum);
  new_assum = tcDecl(expr,assum,state);
  SETQ(A_TYEXTS(new_assum),A_TYFCNS(new_assum)); /* imported */
  data = (struct checkimport *)alloctmp(sizeof(struct checkimport));
  data->zone = expr;
  data->assum = assum;
  data->new_assum = new_assum;
  data->state = state;
  data->err = false;
  (void) MapSweep(A_TYEXTS(new_assum),do_check_extquan,data);
  /* Note that declaring extern dimension quan(unit) also declares unit
     as an external unit, the default unit of the dimension. (So no rescaling
     is necessary when compiling 1`unit, this is compiled as 1.) */
  KEEPONLY(new_assum);
  return new_assum;
}

PRIVATE ExprPtr tcValSpec P((IN CONST ExprPtr expr,INOUT ExprPtr assum,INOUT ExprPtr state));

MAPSWEEPFUNC(do_extvalerr)
{
  struct extvalerr *data;
  ExprPtr zonex;

  data = MAPDATA(struct extvalerr *);
  zonex = MapGetSure(x,A_IDLOCS(data->assum));
  type_soft_error(ZONE(zonex),(CHAR *)varMsg(message(789),x));
  return false;
}

PRIVATE ExprPtr tcExtVal(expr,assum,state)
     /* A_EXT_VAL, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr new_assum;
  ExprPtr errvars;
  struct extvalerr *data;

  checkExtern(expr,assum);
  new_assum = tcValSpec(expr,assum,state);
  SETQ(A_IDEXTS(new_assum),A_IDS(new_assum));
  SETQ(A_IDS(new_assum),NewMap());
  errvars = MapDomRestrTo(A_IDS(assum),A_IDEXTS(new_assum));
  data = (struct extvalerr *)alloctmp(sizeof(struct extvalerr));
  data->assum = new_assum;
  data->state = state;
  (void) MapSweep(errvars,do_extvalerr,data);
  return new_assum;
}

#define tcExtExc tcNewExc

PRIVATE ExprPtr tcMemoFn(expr,assum,state) /* A_MEMO_FN, return type */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr type;
  ENTER;

  type = tcMatch(expr,assum,state);
  type = correctTypeAppl(type,state);
  coerceFlagTypes(ZONE(expr),0/*F_EQ*/
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
		  | F_IMP
#endif
		  ,trueValue,
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
		  0, 0, /*INF_STRENGTH,INF_STRENGTH,*/
#endif
		  CAR(APPLARG(type)),assum,state);
  addAnnotation(expr,type,A_FNLEVEL(assum),state); /* so as to detect
						      whether function is unary
						      or n-ary later on. */
  KEEPONLY(type);
  return type;
}

#define tcConsInfo tcIllegal

PRIVATE ExprPtr tcEqtypeInfo(expr,assum,state)
	/* A_EQTYPE_INFO, return type name */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,e,id,idp,methods,type,wanted,tyname;
  ExprPtr oldeqflags,tyfcn;
  ENTER;

  args = ARGS(expr);
  idp = CAR(args);
  id = CAR(ARGS(idp));
  args = CDR(args);
  e = CAR(args);
  methods = TC(e,assum,state);
  tyfcn = MapGetSure(id,A_TYFCNS(assum));
  type = CDR(tyfcn);
  wanted = NewMap();
  oldeqflags = KEEP(S_EQFLAGS(state));
  setMapMerge(&S_EQFLAGS(state),CDR(CAR(tyfcn)));
  (void) coerceTypes(ZONE(e),methods,wanted,assum,state,false);
  SETQ(S_EQFLAGS(state),oldeqflags);
  tyname = applFun(type,state);
  tyname = fxSym(tyname);
  SETCAR(ARGS(idp),tyname); /* patch in place, so that we get back the type name */
  KEEPONLY(tyname);
  return tyname;
}

MAPCOLLECTFUNC(do_nondummytycons)
{
  if (y->kind==(CHAR)E_CONS) /* then this is a datatype description */
    {
      if (CDR(y)==NewMap()) /* then this is a dummy datatype description */
	return NewMap();
    }
  return MapletKons(x,y);
}

PRIVATE MapPtr nonDummyTyCons(state)
IN CONST ExprPtr state;
{
  return MapCollect(S_TYCONS(state),do_nondummytycons,(CHAR *)NULL);
}

PRIVATE void untrimState(state,local_state)
INOUT ExprPtr state;
IN CONST ExprPtr local_state;
{
  ExprPtr aux;
  ENTER;

  aux = MapMerge(S_NONEQS(state),S_NONEQS(local_state));
  SETQ(S_NONEQS(state),aux);
  aux = MapMerge(S_TYCONS(state),nonDummyTyCons(local_state));
  SETQ(S_TYCONS(state),aux);
  QUIT;
}

MAPSWEEPFUNC(do_checkimports)
{
  struct checkimport *data;
  ExprPtr local_state;
  ExprPtr type,wanted;
  Boolean found;
  ENTER;

  data = MAPDATA(struct checkimport *);
  type = MapGet(x,A_IDS(data->assum),&found);
  if (!found)
    type = MapGet(x,A_IDEXTS(data->assum),&found);
  if (!found)
    {
      if (A_KIND(data->assum)==TA_REGULAR)
	{
	  type_soft_error(ZONE(data->zone),(CHAR *)varMsg(message(806),x));
	  data->err = true;
	}
    }
  else
    {
      local_state = DupTypeState(data->state);
      wanted = MapGetSure(x,A_IDEXTS(data->new_assum));
      type = newTypeInst(type,data->assum,local_state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			 ,false
#endif
			 );
      (void) coerceTypesImport(ZONE(data->zone),type,wanted,
			       data->assum,local_state,x,false);
      /* check the type, but in the local_state, i.e. do not instantiate it
	 for outside the module. */
    }
  QUIT;
  return false;
}

MAPSWEEPFUNC(do_checkdatatypes)
{
  struct checkimport *data;
  ExprPtr tyfcn,new_tyfcn;
  Boolean found;
  ENTER;

  data = MAPDATA(struct checkimport *);
  new_tyfcn = MapGetSure(x,A_TYFCNS(data->new_assum));
  tyfcn = MapGet(x,A_TYFCNS(data->assum),&found);
  if (!found)
    {
      if (A_KIND(data->assum)==TA_REGULAR)
	type_soft_error(ZONE(data->zone),(CHAR *)varMsg(message(498),x));
    }
  else /* then we can link the datatype or dimension exactly only when
	  they were declared imported and exported from the same module,
	  i.e. when tyfcn and new_tyfcn are exactly equal (we could check
	  that they have the same type names, arities and equality
	  conditions, but this can only happen when they are declared
	  in the same .mlx). */
    {
      if (tyfcn!=new_tyfcn)
	{
	  type_soft_error(ZONE(data->zone),(CHAR *)varMsg(message(821),x));
	  data->err = true;
	}
    }
  QUIT;
  return false;
}

PRIVATE ExprPtr tcBasis(expr,assum,state)
     /* A_BASIS, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  Basis *basis;
  ExprPtr imports;
  ENTER;

  basis = (Basis *)EXTDATA(CAR(ARGS(expr)));
  untrimState(state,basis->basisState);
  /* compute the set of imported variables: */
  imports = A_IDEXTS(basis->basisAssum);
  /* If assum is TA_REGULAR, then check that all imports exist and have the right types;
     if it is TA_PRECOMPILE, check that all imports that exist have the right types: */
  {
    struct checkimport *data =
      (struct checkimport *)alloctmp(sizeof(struct checkimport));

    data->zone = expr;
    data->assum = assum;
    data->state = state;
    data->new_assum = basis->basisAssum;
    data->err = false;
    (void) MapSweep(imports,do_checkimports,data);
    if (data->err)
      type_error(ZONE(expr),(CHAR *)message(817));
  }
  /* Check that extern datatypes declared as admitting equality really do,
     and that they have the right type names and arities. */
  imports = A_TYEXTS(basis->basisAssum);
  {
    struct checkimport *data =
      (struct checkimport *)alloctmp(sizeof(struct checkimport));

    data->zone = expr;
    data->assum = assum;
    data->state = state;
    data->new_assum = basis->basisAssum;
    data->err = false;
    (void) MapSweep(imports,do_checkdatatypes,data);
    if (data->err)
      type_error(ZONE(expr),(CHAR *)message(817));
  }
  KEEPONLY(basis->basisAssum);
  return basis->basisAssum;
}

PRIVATE ExprPtr tcEnterFile(expr,assum,state)
     /* A_ENTER_FILE, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{ /* this is for error reporting reasons, so that errors occurring in
     an included file indeed get reported as errors from that file. */
  ExprPtr arg;

  if (yyfilename!=NULL)
    MARKFREE(yyfilename);
  arg = CAR(ARGS(expr));
  yyfilename = (CHAR *)MALLOC(SLEN(arg)+1);
  safememcpy(yyfilename,STEXT(arg),SLEN(arg));
  yyfilename[SLEN(arg)] = 0;
  setCurrentFile();
  return NewTypeAssum();
}

PRIVATE ExprPtr tcExitFile(expr,assum,state)
     /* A_EXIT_FILE, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{ /* this is for error reporting reasons, so that errors occurring in
     an included file indeed get reported as errors from that file. */
  if (yyfilename!=NULL)
    MARKFREE(yyfilename);
  yyfilename = NULL;
  setCurrentFile();
  return NewTypeAssum();
}

PRIVATE ExprPtr tcPragma(expr,assum,state)
     /* A_PRAGMA, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  CHAR *pragma;

  pragma = STEXT(CAR(ARGS(expr)));
  switch (pragma[3])
    {
    case 'E': /* equality for datatypes pragma */
      if (pragma[4]=='-')
	A_FLAGS(assum) |= A_NOEQ_DATATYPES;
      else A_FLAGS(assum) &= ~A_NOEQ_DATATYPES;
      break;
    case 'V':
      if (pragma[4]=='-')
	A_FLAGS(assum) |= A_NOWARN_NOBIND;
      else A_FLAGS(assum) &= ~A_NOWARN_NOBIND;
      break;
    }
  return NewTypeAssum();
}

	/***** modules *****/

#if MODULE_SYSTEM & STD_ML_MODULE
PRIVATE ExprPtr tcSig(expr,assum,state)
	/* A_SIG, return signature */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,e,new_assum,assum1;
  ENTER;

  args = ARGS(expr);
  assum = DupTypeAssum(assum); /* to be sure it won't be modified */
  new_assum = NewTypeAssum();
  SETQ(A_NONGENS(assum),NewMap()); /* outer variables become rigid locally */
  for (args = ARGS(expr); args!=NIL; args=CDR(args))
    {
      e = CAR(args);
      assum1 = TC(e,assum,state);
      AddTypeAssum(new_assum,assum1);
    }
  KEEPONLY(new_assum);
  return new_assum;
}

PRIVATE ExprPtr tcLocalSpec(expr,assum,state)
	/* A_LOCAL_SPEC, return assum */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,sig1,sig2;
  ExprPtr e;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  sig1 = TC(e,assum,state);
  AddTypeAssum(sig1,assum);
  e = CAR(CDR(args));
  sig2 = TC(e,sig1,state);
  KEEPONLY(sig2);
  return sig2;
}
#else
#define tcSig tcIllegal
#define tcLocalSpec tcIllegal
#endif

PRIVATE ExprPtr tcValSpec(expr,assum,state)
	/* A_VAL_SPEC, A_EXT_SPEC, return assum */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr new_assum,args,e,id_type;
  ENTER;

  new_assum = NewTypeAssum();
  for (args=ARGS(expr); args!=NIL; args=CDR(args))
    {
      e = CAR(args);
      id_type = TC(e,assum,state);
      setMapPut(CAR(id_type),CDR(id_type),&A_IDS(new_assum));
      setMapPut(CAR(id_type),e,&A_IDLOCS(new_assum));
    }
  KEEPONLY(new_assum);
  return new_assum;
}

PRIVATE ExprPtr tcValSpecBind(expr,assum,state)
	/* A_VAL_SPEC_BIND, return (id,type) */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,id,e,type;

  args = ARGS(expr);
  id = CAR(ARGS(CAR(args)));
  e = CAR(CDR(args));
  type = TC(e,assum,state);
#if DEBUGGING!=0
  addAnnotation(CAR(args),type,A_FNLEVEL(assum),state);
#endif
  type = Cons(id,type);
  return KEEP(type);
}

PRIVATE ExprPtr tcTypeSpec1(expr,assum,state,tst,eqp)
	/* A_TYPE_SPEC, return assum */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
IN typeSpecTag tst;
IN Boolean eqp;
{
  ExprPtr new_assum,args,e,id_type,tyfcn,tyname[2],nm,aux,aux2;
  COUNTER i,n;
  ENTER;

  new_assum = NewTypeAssum();
  for (args=ARGS(expr); args!=NIL; args=CDR(args))
    {
      e = CAR(args);
      id_type = TC(e,assum,state);
      tyfcn = CDR(CDR(id_type));
      tyname[0] = CAR(CDR(id_type));
      n = 1;
      switch (tst)
	{
	case tst_Type:
	  break;
	case tst_NumType:
	  aux = KnumType(MapSingleton(CDR(tyfcn),one));
	  tyfcn = KONS(CAR(tyfcn),KEEP(aux));
	  break;
	case tst_TupleType:
	  n = 2;
	  tyname[1] = getCdrTyName(tyname[0],state);
	  tyname[0] = getCarTyName(tyname[0],state);
	  aux = KtypeAppl(tyname[0],CDR(APPLARG(tyfcn)));
	  (void) KEEP(aux);
	  aux2 = KtypeAppl(tyname[1],CDR(APPLARG(tyfcn)));
	  (void) KEEP(aux2);
	  tyfcn = KONS(CAR(tyfcn),KONS(aux,aux2));
	  break;
	case tst_RecordType:
	  tyfcn = KONS(CAR(tyfcn),MapSingleton(REST_TAG,CDR(tyfcn)));
	  break;
	}
      setMapPut(CAR(id_type),tyfcn,&A_TYFCNS(new_assum));
      for (i=0; i<n; i++)
	{
	  nm = tyname[i];
	  setSetAdd(nm,&A_NONGENS(new_assum)); /* to make the tynames flexible */
	  if (!eqp)
	    {
	      setMapPut(nm,nameToTdg(nm),&S_NONEQS(state));
	    }
	  nm = Kref(E_REF,nm); /* don't know if it's a ref/array or not */
	  (void) KEEP(nm);
	  setMapPut(nm,nameToTdg(nm),&S_NONEQS(state));
	}
    }
  KEEPONLY(new_assum);
  return new_assum;
}

PRIVATE ExprPtr tcTypeSpec(expr,assum,state)
	/* A_TYPE_SPEC, return assum */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  return tcTypeSpec1(expr,assum,state,tst_Type,false);
}

#if MODULE_SYSTEM & STD_ML_MODULE
PRIVATE ExprPtr tcEqTypeSpec(expr,assum,state)
	/* A_EQTYPE_SPEC, return assum */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  return tcTypeSpec1(expr,assum,state,tst_Type,true);
}

PRIVATE ExprPtr tcNumTypeSpec(expr,assum,state)
	/* A_NUMTYPE_SPEC, return assum */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  return tcTypeSpec1(expr,assum,state,tst_NumType,true);
}
#else
#define tcEqTypeSpec tcIllegal
#define tcNumTypeSpec tcIllegal
#endif

PRIVATE ExprPtr tcTypeSpecBind(expr,assum,state)
	/* A_TYPE_SPEC_BIND, return id::tyname::tyfcn */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,name,type,e,tyvarseq,tyname;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  tyvarseq = TC(e,assum,state); /* a A_TYPE_VAR_SEQ elaborates to a list of symbols */
  args = CDR(args);
  name = CAR(ARGS(CAR(args)));
  tyname = freshTyName(name,assum,state);
  type = KtypeAppl(tyname,tyvarseq);
  (void) KEEP(type);
  expr = Kons(name,KONS(tyname,KONS(getTyHdr(tyvarseq,NIL,state),type)));
  KEEPONLY(expr);
  return expr;
}

#if MODULE_SYSTEM & STD_ML_MODULE
MAPCOLLECTFUNC(do_gettynames)
{
  ExprPtr state;

  state = MAPDATA(ExprPtr);
  y = applFun(CDR(y),state);
  return Singleton(fxSym(y));
}

PRIVATE ExprPtr tcDataTypeDesc(expr,assum,state)
	/* A_DATATYPE_DESC, return assumptions */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr new_assum,tynames;
  ENTER;

  new_assum = tcDataType(expr,assum,state);
  tynames = MapCollect(A_TYFCNS(new_assum),do_gettynames,state);
  setSetUnion(&A_NONGENS(new_assum),tynames); /* to make them flexible */
  KEEPONLY(new_assum);
  return new_assum;
}
#else
#define tcDataTypeDesc tcIllegal
#endif

#define tcExceptionDesc tcValSpec
#define tcNewExcDesc tcNewExc

#if MODULE_SYSTEM & STD_ML_MODULE
PRIVATE ExprPtr tcStructDesc(expr,assum,state)
	/* A_STRUCT_DESC, return assum */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr new_assum,args,e,id_sig,strname;
  ENTER;

  new_assum = NewTypeAssum();
  for (args=ARGS(expr); args!=NIL; args=CDR(args))
    {
      e = CAR(args);
      id_sig = TC(e,assum,state);
      strname = CDR(CDR(id_sig));
      setSetAdd(strname,&A_NONGENS(new_assum)); /* flexible */
      setMapPut(CAR(id_sig),KONS(strname,CAR(CDR(id_sig))),&A_STRUCTS(new_assum));
    }
  KEEPONLY(new_assum);
  return new_assum;
}

PRIVATE ExprPtr tcStructBindDesc(expr,assum,state)
	/* A_STRUCT_BIND_DESC, return id::sig::strname,
		where sig is not copied */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr e,args,id_sig,id,sig,strname;
  ENTER;

  args = ARGS(expr);
  id = CAR(ARGS(CAR(args)));
  e = CAR(CDR(args));
  sig = TC(e,assum,state);
  strname = freshTyName(id,assum,state);
  setMapPut(strname,sig,&S_TYCONS(state));
  id_sig = Cons(id,Cons(sig,strname));
  KEEPONLY(id_sig);
  return id_sig;
}

PRIVATE ExprPtr tcSharingDesc(expr,assum,state)
	/* A_SHARING_DESC, return assum */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr args,e;
  ENTER;

  for (args=ARGS(expr); args!=NIL; args=CDR(args))
    {
      e = CAR(args);
      (void) TC(e,assum,state);
    }
  QUIT;
  return NewTypeAssum();
}

PRIVATE ExprPtr StructsToShare(args,assum,state)
IN CONST ExprPtr args;
IN CONST ExprPtr assum;
IN CONST ExprPtr state;
{
  ExprPtr share,var,args2,namesig,assum2;
  Boolean found;
  ENTER;

  share = NewMap();
  for (; args!=NIL; args=CDR(args))
    {
      var = CAR(args);
      args2 = ARGS(var);
      assum2 = assum;
      getLongStruct(ZONE(var),CDR(args2),&assum2,state);
      namesig = MapGet(CAR(args2),A_STRUCTS(assum2),&found);
      share = SetAdd(CAR(namesig),share);
    }
  KEEPONLY(share);
  return share;
}

PRIVATE ExprPtr tcStructShareDesc(expr,assum,state)
	/* A_STRUCT_SHARE_DESC, return NIL */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  /*!!!*/
  return NIL;
}

#else
#define tcStructDesc tcIllegal
#define tcStructBindDesc tcIllegal
#define tcSharingDesc tcIllegal
#define tcStructShareDesc tcIllegal
#endif

PRIVATE tcAct tcAction[] = {
   tcString, tcNumber, tcVariable,
   tcUnitPower, tcUnitStar, tcSetUnit,
   tcLabel, tcRecordMaplet, tcRecord, tcFixity,
   tcTuple, tcProgn, tcList, tcListOf, tcListMap,
   tcInSet, tcInList,
   tcMaplet, tcMap, tcMapOf, tcMapMap,
   tcMapUnder, tcMapUnderOf, tcMapUnderMap,
   tcProgOf, tcProgMap,
   tcAllOf, tcAllMap, tcExistsOf, tcExistsMap,
   tcSomeOf, tcSomeMap,
   tcLet, tcApply,
   tcCast, tcAnd, tcOr,
   tcHandle, tcRaise,
   tcFn, tcLambda,
   tcIf, tcWhile, tcWhileList, tcWhileMap, tcWhileMapUnder,
   tcWhileAll, tcWhileExists, tcWhileSome,
   tcDeclList, tcLocal,
   tcVal, tcValRec, tcBind,
   tcPatTuple,
   tcQuantity, tcNewQuan, tcUnit, tcNewUnit,
   tcType, tcTypeBind, tcDataType, tcDataTypeBind, tcAbsType,
   tcTypeVarSeq, tcConstructor,
   tcException, tcNewExc, tcExcBind,
   tcMapletType, tcRecordType, tcTypeApply,
   tcTypePower, tcTypeStar, tcTupleType,
   tcFunctionType, tcMapType,
   tcPatAny, tcPatRecordEllipsis, tcPatTupleEllipsis,
   tcPatRecordMaplet, tcPatRecord, tcPatAs, tcPatCast,
   tcPatNil, tcPatAnyList, tcPatApply, tcPatEmpty, tcPatAnyMap,
   tcPatMaplet, tcPatMapAdd, tcPrimeDataBind,
   tcUnitName, tcTypeVar, tcRealVar, tcPatVar, tcPatConst,
   tcSelectFn, tcMatch, tcMu, tcSubMap, tcDelay, tcNonSense,
   tcPack, tcPatPack, tcInteger,
   tcTypeEllipsis, tcExtDatatype, tcExtEqtype, tcExtQuantity, tcExtVal,
   tcMemoFn, tcConsInfo, tcEqtypeInfo, tcExtExc, tcBasis,
   tcEnterFile, tcExitFile, tcPragma, tcFunBind,

   tcSig, tcLocalSpec, tcValSpec, tcValSpecBind, tcTypeSpec,
   tcEqTypeSpec, tcNumTypeSpec, tcTypeSpecBind, tcDataTypeDesc,
   tcExceptionDesc, tcNewExcDesc, tcStructDesc, tcStructBindDesc,
   tcSharingDesc, tcStructShareDesc
 };

DEF_PRIVATE_ACTION_1ARG(typecheck)
{
  struct typecheck *data;

  data = (struct typecheck *)ENV_UNDERLYING(env);
  return (*tcAction[TYPE(data->expr)])(data->expr,data->assum,data->state);
}
ACTION_END

PRIVATE ExprPtr typecheck1(expr,assum,state)
INOUT ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  struct typecheck *data = (struct typecheck *)alloctmp(sizeof(struct typecheck));

  data->expr = expr;
  data->assum = assum;
  data->state = state;
  return CallCC(ACTION_NAME(typecheck),(ExprPtr)data,true);
}

PUBLIC ExprPtr tc(expr,assum,state)
INOUT ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr type;
  ENTER;

  if (STKFULL(type))
    return typecheck1(expr,assum,state);
  type = (*tcAction[TYPE(expr)])(expr,assum,state);
  KEEPONLY(type);
  return type;
}
